/*****************************************************************************
 * igmp.c:
 *****************************************************************************
 * Copyright (C) 2006-2012 the FTT-SE team.
 *
 * Author: Ricardo Marau <marau at fe.up.pt>
 *
 * FTT-SE is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  FTT-SE is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with FTT-SE.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

#include "Ports.h"
#include "igmp.h"
#include "ethernet.layer.h"


typedef struct ip_header {
    unsigned char vers_hlen;
    unsigned char servType;
    unsigned short lenght;
    unsigned short ident;
    unsigned short flags_frOffset;
    unsigned char ttl;
    unsigned char protocol;
    unsigned short headChecksum;
    IP_ADDR source;
    IP_ADDR dest;
}  __attribute__ ((packed)) IP_HEADER;

typedef struct igmp_header {
    unsigned char type;
    unsigned char mrt;
    unsigned short igmp_checksum;
    IP_ADDR gda;
 }  __attribute__ ((packed)) IGMP_HEADER;

static struct ip_addr myIpAddr = {{0,0,0,0}};

/* Send an IP packet */
void IGMP_set_ip(void *pkt, unsigned short len, IP_ADDR dest, unsigned short protocol_id)
{
	static unsigned short ip_ident = 0;
	IP_HEADER *iphd;
	unsigned short check = 0, oldCheck = 0;
	int i;
	unsigned short *pt;

	iphd = (IP_HEADER *)(pkt+14);
	iphd->vers_hlen = (4 << 4) + 5;
	iphd->servType = 8;
	iphd->lenght = htons(len + sizeof(IP_HEADER));
	iphd->ident = htons(ip_ident++);
	iphd->flags_frOffset = 0;
	iphd->ttl = 10;
	iphd->protocol = protocol_id; //IP_UDP_TYPE; //IGMP=2
	iphd->headChecksum = 0;
	iphd->source = myIpAddr;
	iphd->dest = dest;

	/* Compute the checksum */
	pt = (unsigned short *)iphd;
	check = oldCheck = *pt;
	pt++;
	for (i = 1; i < 10; i++) {
		check += *pt;
		if (oldCheck > check) check++;
		oldCheck = check;
		pt++;
	}
	check = ~check;
	iphd->headChecksum = check;
}

#define IGMG_report (0x16)
#define IGMG_leave (0x17)

unsigned short IGMP_assign_igmp_cmd(void *pkt, IP_ADDR *gda)
{
	IGMP_HEADER *igmp_hd;
	unsigned short check = 0, oldCheck = 0;
	int i;
	unsigned short *pt;

	
	igmp_hd = (IGMP_HEADER *)(pkt+14+sizeof(IP_HEADER));
	igmp_hd->type = IGMG_report;
	igmp_hd->mrt = 0;

	igmp_hd->igmp_checksum = 0;
//	igmp_hd->type = IGMG_report;
	igmp_hd->gda = *gda;


	/* Compute the checksum */
	pt = (unsigned short *)igmp_hd;
	check = oldCheck = *pt;
	pt++;
	for (i = 1; i < 4; i++) {
		check += *pt;
		if (oldCheck > check) check++;
		oldCheck = check;
		pt++;
	}
	check = ~check;
	igmp_hd->igmp_checksum = check;

	return sizeof(IGMP_HEADER);
}


unsigned short IGMP_deassign_igmp_cmd(void *pkt, IP_ADDR *gda)
{
	IGMP_HEADER *igmp_hd;
	unsigned short check = 0, oldCheck = 0;
	int i;
	unsigned short *pt;

	
	igmp_hd = (IGMP_HEADER *)(pkt+14+sizeof(IP_HEADER));
	igmp_hd->type = IGMG_report;
	igmp_hd->mrt = 0;

	igmp_hd->igmp_checksum = 0;
	igmp_hd->type = IGMG_leave;
	igmp_hd->gda = *gda;


	/* Compute the checksum */
	pt = (unsigned short *)igmp_hd;
	check = oldCheck = *pt;
	pt++;
	for (i = 1; i < 4; i++) {
		check += *pt;
		if (oldCheck > check) check++;
		oldCheck = check;
		pt++;
	}

	check = ~check;
	igmp_hd->igmp_checksum = check;

	return sizeof(IGMP_HEADER);
}



unsigned char ip_compAddr(IP_ADDR ip1, IP_ADDR ip2)
{
	int i;
    
	for (i=0; i < 4; i++)
		if (ip1.ad[i] != ip2.ad[i]) return 0;

	return 1;
}

/* traslate an IP address from text string to 4 bytes */
unsigned char ip_str2addr(char *a, IP_ADDR *ip)
{
	int ad[6];
	int i,j;
	char c;

	i = 0;
	for(j = 0; j < 4; j++) {
		ad[j] = 0;
		while((a[i] != '.') && (a[i] != 0)) {
			c = a[i++];
			if (c <= '9') c = c - '0';
			else c = c - 'A' + 10;
			ad[j] = ad[j] * 10 + c;
		}
		i++;
	}

	for (i=0; i<4; i++)
		ip->ad[i] = ad[i];

	return 1;
}

unsigned short IGMP_sizeof(void)
{
	return (sizeof(IGMP_HEADER));
}
unsigned short IGMP_frameSize(void)
{
	return (sizeof(IGMP_HEADER)+sizeof(IP_HEADER)+12+2);
}


unsigned char IGMP_isMulticast(unsigned char *mac)
{
	if ((mac[0]==0x01)&&(mac[1]==0x00)&&(mac[2]==0x5E))
		return 1;
	else
		return 0;
}

void IGMP_setMulticastBase(unsigned char *mac)
{
	mac[0]=0x01;
	mac[1]=0x00;
	mac[2]=0x5E;
	mac[3]=0x01; //The most significant bit here must be 0! Don't know why but the whole byte can't be 0.
	mac[4]=0x00;
	mac[5]=0x00;
	return ;
}

