/*****************************************************************************
 * M_ECreg.h:
 *****************************************************************************
 * Copyright (C) 2006-2012 the FTT-SE team.
 *
 * Author: Ricardo Marau <marau at fe.up.pt>
 *
 * FTT-SE is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  FTT-SE is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with FTT-SE.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

#ifndef ECSHEDULEREG_H
#define ECSHEDULEREG_H

#include "ftt-global.h"

/** @brief Status of the EC-Schedule registers */
enum EC_SHED_STATUS
{
	UPDATED             = 0,   ///< EC-schedule built and ready to be used)
	OUT_OF_DATE         = 1,   ///< EC-schedule empty (during boot) 
	BUSY_SCHED          = 2,   ///< EC-schedule being used by the scheduler
	BUSY_DISP           = 3,   ///< EC-schedule being used by the dispatcher 
	MISSED_DEADLINE     = 4,   ///< Hard deadline missed while scheduling
	DISPATCHED          = 5    ///< EC-schedule has been dispatched
};

/** @name Scheduler Errors
 * This errors are to be used in the ec_schedules plans - \ref ECType - err_code field
 * @{
 */
enum EC_SHED_ERRORS
{
	SCH_ERRCODE_OK             = 0,   ///< No error
	SCH_ERRCODE_SM_MISS_DEADLINE         = 1,   ///< Deadline miss in synchronous message 
	SCH_ERRCODE_AM_HARD_MISS_DEADLINE          = 2,   ///< Deadline miss in hard asynchronous message
	SCH_ERRCODE_AM_SOFT_MISS_DEADLINE           = 3,   ///< Deadline miss in soft asynchronous message 
	SCH_ERRCODE_AM_UNCON_MISS_DEADLINE     = 4,   ///< Deadline miss in unconstrained asynchronous message
};

/* Scheduler Errors (err_code field) */
/* Note that a missed deadline in hard messages causes the scheduler to stop */
/* Hard messages are all the synchronous and the hard asynchrobnous */


#ifdef SW_UNICAST

typedef struct{
	unsigned long begin;
	unsigned long end;
	unsigned long used;
}DOWN_LOAD_BLOCK;

/** @brief SW Port Link Load struct */
typedef struct 
{
	unsigned long up_load; /* from sw to station */
	unsigned long down_load; /* from station to sw */
	DOWN_LOAD_BLOCK down_blocks[EC_MAX_VARS];
	unsigned short down_blocks_no;
}LINK_LOAD;
#endif

typedef struct
{
	unsigned char status;                  ///< Status of the EC-Schedule : UPDATED, OUT_OF_DATE,...
	unsigned char err_code;                ///< Details about (if relevant) the cause of the error

	struct{
#ifdef SW_UNICAST		
		LINK_LOAD ports_load[SW_MAX_PORTS];	/* array similar to the one in sw_table */
#else
		unsigned long updown_load;
#endif
		unsigned short n_mesgs;
		struct{
			unsigned short id;
			unsigned char source_nodeID;
			unsigned short n_fragmentation;
#ifdef SLOTED_EC			
			unsigned char sloted_size;
#endif
		}msgs[EC_MAX_SM_VARS];
	}sm;

	struct{
		unsigned long updown_load;
		unsigned short n_mesgs;
		struct{
			unsigned short id;
			unsigned char source_nodeID;
			unsigned short n_fragmentation;
		}msgs[EC_MAX_AM_VARS];
	}am;
} ECType;

/** @brief Pointer to EC-schedule registers in use */
extern ECType *sched_ec;

/** @brief Pointer to EC-schedule registers to be dispatched */
extern ECType *disp_ec;

/** @fn void ECreg_Init(void) 
 *  @brief This function initializes the ftt-master ECreg plans.
 *
 *  Initializes ECSchedule plans
 *  @return -
 */
extern void ECreg_Init(void);

/** @fn void ECreg_SwapPlans(void) 
 *  @brief This function swaps the plan to be filled in the nesx Schedule
 *
 *  @return -
 */
extern void ECreg_SwapPlans(void);


extern void ECreg_prepare2build(void);


/** @fn signed char EC_printsched(void) 
 *  @brief This function prints the EC-schedule contents
 *
 *  @return - 	( 0) - OK\n
 *  		(-1) - Schedule Empty
 */
extern void EC_printsched(void);
extern void EC_printdisp(void);

/** @brief Sequence number of the trigger message */
extern unsigned long EC_Get_and_Inc_seqno(void);


extern signed char EC_push_Signaling_notification( unsigned short nodeID, unsigned char seq_no );

extern signed char EC_pull_Signaling_notification( unsigned short *nodeID, unsigned char *seq_no );


#endif
