/*****************************************************************************
 * M_SW_add_table.c:
 *****************************************************************************
 * Copyright (C) 2006-2012 the FTT-SE team.
 *
 * Author: Ricardo Marau <marau at fe.up.pt>
 *
 * FTT-SE is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  FTT-SE is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with FTT-SE.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/


#include "Ports.h"

#include "ethernet.layer.h"	//for the MAC associated functions
#include "ftt-global.h"		//for the SW_MAX_PORTS define
#include "M_appPnP.h"


#define M_SW_CONNECTION_LEASE_TIME  6 /*Ecs*/

static unsigned char M_SW_add_table_init=0;

typedef struct{
	unsigned char mac_address[6];
	unsigned char lease_time;	//in ECs
	                            // if 0 then it's forever
	unsigned char last_seq_no;
	unsigned char seq_no_reset_flag;
	
	unsigned char in_use;
}MAC_TABLE;

static MAC_TABLE arp_table[SW_MAX_PORTS]; //the index refers to the node ID (there can't be more than 255 nodes)
static unsigned char arp_table_size;


signed char M_SW_Init(void)
{
	unsigned char temp_id;

	for (temp_id=0; temp_id<SW_MAX_PORTS; temp_id++)
		arp_table[temp_id].in_use = 0;

	arp_table_size = 0;

	/* Declare the module as initialized */
	M_SW_add_table_init = 1;
#ifdef PRINT_MODULES_INIT
	PRINT_MSG(" -M_SW_Init Ok");
#endif
	return 0;
}

signed char M_SW_add_table_init_check(void)
{
	if (M_SW_add_table_init == 0)
		return -1;
	return 0;
}

signed int M_SW_subscribe_node( unsigned char *input_MAC, unsigned char lease_time )
{
	signed int item_found = -1;
	unsigned char temp_id;

	/* Check if there is room */
	if (arp_table_size == 0xFF){
		ERROR_MSG("No room left for additional nodes. The hotel is full booked ");
		return -1;
	}

	/* Check for duplicity while searching */
	for (temp_id=0; temp_id<SW_MAX_PORTS; temp_id++)
	{

		/* Is it free? */
		if ( (item_found == -1) && (arp_table[temp_id].in_use == 0) )
			item_found = temp_id;

		if ( arp_table[temp_id].in_use )
			if (ETH_L_CompareMAC( arp_table[temp_id].mac_address, input_MAC)){
//				ERROR_MSG("This MAC is already booked...");
				return -2; //temp_id;
			}
	}

	if ( item_found == -1 ){
		ERROR_MSG("No free room found after positive availability check");
		return -3;
	}

	/* Set the aging time and the MAC for this ID */
	if (lease_time)
		arp_table[item_found].lease_time = M_SW_CONNECTION_LEASE_TIME+5; /* +5 just for the kick-off */
    
	ETH_L_CopyMAC_to_from( arp_table[item_found].mac_address, input_MAC );
	arp_table[item_found].in_use = 1;

	arp_table[item_found].seq_no_reset_flag= 1 ; /* The seq_no is being reseted */
	
	arp_table_size++;
	return item_found;
}

signed int M_SW_unsubscribe_node( unsigned char index_id )
{
	if (arp_table[index_id].in_use){
		arp_table_size --;
		arp_table[index_id].in_use = 0;
	}
	return 0;
}

signed char M_SW_downall_lease_time(void)
{
	unsigned short index_id;
	/* Go through all the table to the MACs table and downgrade the lease times of all */

	for ( index_id=0; index_id<SW_MAX_PORTS; index_id++ )
		if ( arp_table[index_id].in_use )
			if (arp_table[index_id].lease_time)
				if ( --(arp_table[index_id].lease_time) == 0){
					/* Remove the item */
					PRINT_MSG("\n\n->Disconnecting node %d", index_id);
					arp_table[index_id].in_use = 0;
					M_appPnP_Append_request( -1, index_id ); /* Sending a node lost notification upwards */

				}

	return 0;
}

signed char M_SW_heartbeat( unsigned char index_id, unsigned char seq_no, unsigned char *last_seq_no )
{
	unsigned char temp_seq_no;
	if ( arp_table[index_id].in_use == 0 ){
		ERROR_MSG("The node was not in use");
		return -1;
	}

	if (arp_table[index_id].lease_time)
		arp_table[index_id].lease_time = M_SW_CONNECTION_LEASE_TIME;
	
	if ( arp_table[index_id].seq_no_reset_flag ) {
		arp_table[index_id].seq_no_reset_flag = 0;
		arp_table[index_id].last_seq_no = seq_no;
		return 0;
	}
	
	temp_seq_no = arp_table[index_id].last_seq_no;
	arp_table[index_id].last_seq_no = seq_no;
	
	if ( temp_seq_no != ((seq_no-1)&0xFF) ){
		if ( last_seq_no ) *last_seq_no = temp_seq_no;
		return -2;
	}
	
	return 0;
}

unsigned char M_SW_is_idx_in_use( unsigned char index_id )
{
	return arp_table[index_id].in_use;
}

unsigned char M_SW_get_mac( unsigned char index_id, unsigned char *mac )
{
	ETH_L_CopyMAC_to_from( mac, arp_table[index_id].mac_address );
	return arp_table[index_id].in_use;
}



void M_SW_Print(void)
{
	int i;
	char mac_str[30];

	PRINT_MSG("Master arp table (%u mesgs stored)",arp_table_size);
	for(i = 0 ; i < arp_table_size ; i++){
		if ( arp_table[i].in_use ){
			sprintf(mac_str, "%x:%x:%x:%x:%x:%x",
					arp_table[i].mac_address[0],
					arp_table[i].mac_address[1],
					arp_table[i].mac_address[2],
					arp_table[i].mac_address[3],
					arp_table[i].mac_address[4],
					arp_table[i].mac_address[5]);
		
			PRINT_MSG("[%2d] MAC:%s",
					i, mac_str);
		}
	}

	return ;
}


