/*****************************************************************************
 * M_db.c:
 *****************************************************************************
 * Copyright (C) 2006-2012 the FTT-SE team.
 *
 * Author: Ricardo Marau <marau at fe.up.pt>
 *
 * FTT-SE is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  FTT-SE is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with FTT-SE.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

#include "Ports.h"
#include "M_db.h"
#include "M_Ready_queue.h"
#include "M_SW_add_table.h"
#include "math_tools.h"

#include "M_master_ch.h"

typedef enum
{
	FORCE_NEW_ITEM=0,
	MODIFY_ITEM,
	PRESERVE_ITEM,
	ITEM_IS_IN_THE_MAIN_LIST,
	ITEM_IS_PIGGYBACKED, //means this item is piggybacked on another with the same ID using generic_p
}MDB_RT_mesgtype_STATUS;

typedef struct mdb_rt_mesg_type{
	/* requirements - the ID is already distinguishing the type! */
	MESSAGE_REQUESITS req;

	/* fragmentation calculus parameters */
	MESSAGE_FRAGMENTS frag;

	/* dynamic parameters */
	MESSAGE_DYNAMIC_DATA dynamic;
	
	/* status, used when:
	 * - moving the table plans 
	 * - piggybacking the item in other with the same ID */
	MDB_RT_mesgtype_STATUS item_status;

	struct mdb_rt_mesg_type *generic_p;
	
	struct mdb_rt_mesg_type *link;
}MDB_RT_mesgtype;

static void MDB_RT_mesgtype_print(MDB_RT_mesgtype *m, unsigned char f)
{
	MESSAGE_REQUESITS_print( &(m->req), 0);
//	MESSAGE_FRAGMENTS_print( &(m->frag), 0);
	MESSAGE_DYNAMIC_DATA_print( &(m->dynamic), 0);
	if (f) PRINTF("\n");
}


/* START - List manipulation functions */
static void MDB_list_add_this_point( MDB_RT_mesgtype **queue_p, MDB_RT_mesgtype *p)
{
	/* insert in the list top */
	p->link = *queue_p;
	*queue_p = p;
}

static signed char MDB_list_remove_from_top( unsigned char num, MDB_RT_mesgtype **queue_p, MDB_RT_mesgtype **ret_p)
{
	unsigned char i;
	MDB_RT_mesgtype **temp_p;
	if (ret_p!=NULL)
		*ret_p = *queue_p;

	temp_p = queue_p;
	for (i=0; i<num; i++){
		if (*temp_p == NULL) return -1;
		temp_p = &((*temp_p)->link);
	}
	*queue_p = *temp_p;
	*temp_p=NULL;

	return 0;
}

static signed char MDB_list_move_Item_to(MDB_RT_mesgtype **from, MDB_RT_mesgtype **to)
{
	MDB_RT_mesgtype *temp_p;
	if (MDB_list_remove_from_top( 1, from, &temp_p) < 0)
		return -1;
	MDB_list_add_this_point( to, temp_p);
	return 0;
}

static MDB_RT_mesgtype **MDB_list_track_pointer(MDB_RT_mesgtype **queue_p, MDB_RT_mesgtype *pointer)
{
	/* Search for it in the list and save the pointer that points to it */
	MDB_RT_mesgtype **temp_p=queue_p;

	if (temp_p == NULL) 
		return NULL;

	while (*temp_p != NULL){
		if (*temp_p == pointer){
			/* Found */
			return temp_p;
		}
		temp_p = &((*temp_p)->link);
	}
	return NULL;
}

/* Do not use the list pointer, but instead, use a temporary list pointer (properly initialized) */
static MDB_RT_mesgtype *MDB_list_get_item_and_move_ahead(MDB_RT_mesgtype **queue_p)
{	
	MDB_RT_mesgtype *ret_p = *queue_p;
	if (*queue_p != NULL) 
		*queue_p = ((*queue_p)->link);
	return ret_p;
}


//static unsigned short MDB_list_get_size(MDB_RT_mesgtype **queue_p)
//{
//	unsigned short temp_size=0;
//	while (*queue_p != NULL){
//		temp_size++;
//		queue_p = &((*queue_p)->link);
//	}
//	return temp_size;
//}

//static void MDB_list_print(MDB_RT_mesgtype **queue_p)
//{
//	while (*queue_p != NULL){
//		PRINTF("->%p",*queue_p);
//		queue_p = &((*queue_p)->link);
//	}
//	PRINTF("->%p\n",*queue_p);
//}

/* STOP - List manipulation functions */

#define MDB_MAX_POOL_SIZE ( (MST_MAX_SM_VARS+MST_MAX_AM_VARS)*2 )

static MDB_RT_mesgtype MDB_RT_pool[MDB_MAX_POOL_SIZE];
static MDB_RT_mesgtype *MDB_free_list = NULL;

static inline void MDB_RT_pool_Init(void)
{
	unsigned int i;

	for( i=0; i < MDB_MAX_POOL_SIZE; i++)
		MDB_list_add_this_point( &MDB_free_list, &(MDB_RT_pool[i]) );
}


typedef struct{
	MDB_RT_mesgtype *SRT_mesgtable;
	MDB_RT_mesgtype *ART_mesgtable;
	rtl_pthread_mutex_t mutex_locker;  //locker
	unsigned char locker_flag;	// 0/1
}MDB_tabtype;

static signed char MDB_tabtype_Init(MDB_tabtype *table)
{
	table->SRT_mesgtable = NULL;
	table->ART_mesgtable = NULL;
	rtl_pthread_mutex_init(&(table->mutex_locker), 0);
	table->locker_flag = 0;
	return 0;
}
static signed char MDB_tabtype_Close(MDB_tabtype *table)
{
	/* Should I empty the SRT and ART list? */
	rtl_pthread_mutex_destroy(&(table->mutex_locker));
	table->locker_flag = 0;
	return 0;
}
static signed char MDB_tabtype_EnterLock(MDB_tabtype *table)
{
	rtl_pthread_mutex_lock(&(table->mutex_locker));
	table->locker_flag = 1;
	return 0;
}
static signed char MDB_tabtype_LeaveLock(MDB_tabtype *table)
{
	table->locker_flag = 0;
	rtl_pthread_mutex_unlock(&(table->mutex_locker));
	return 0;
}
static signed char MDB_tabtype_IsLocked(MDB_tabtype *table)
{
	return table->locker_flag;
}

static void MDB_tabtype_print(MDB_tabtype *table, unsigned char f)
{
	MDB_RT_mesgtype *temp_p, *i_p;

	PRINT_MSG("  Locker = %d", table->locker_flag );

	PRINT_MSG("  SM_list\n");
	temp_p = table->SRT_mesgtable;
	while ( (i_p = MDB_list_get_item_and_move_ahead(&temp_p)) != NULL ){
		PRINTF("  -");
		MDB_RT_mesgtype_print( i_p, 0);
		while (i_p->generic_p != NULL){
			PRINTF("  -");
			MDB_RT_mesgtype_print( i_p->generic_p, 0);
			i_p = i_p->generic_p;
		}
		PRINTF("\n");
	}
	
	PRINT_MSG("  AM_list\n");
	temp_p = table->ART_mesgtable;
	while ( (i_p = MDB_list_get_item_and_move_ahead(&temp_p)) != NULL ){
		PRINTF("  -");
		MDB_RT_mesgtype_print( i_p, 0);
		while (i_p->generic_p != NULL){
			PRINTF("  -");
			MDB_RT_mesgtype_print( i_p->generic_p, 0);
			i_p = i_p->generic_p;
		}
		PRINTF("\n");
	}
}


static MDB_tabtype MDB_runtime_table;
static MDB_tabtype MDB_backtime_table;

static unsigned char MDB_Init_flag = 0;

signed char MDB_Init(unsigned char *mac_reg, unsigned long ec_period_reg )
{
	/* Initialize the pool of table items */
	MDB_RT_pool_Init();

	/* Initialize the table as empty */
	MDB_tabtype_Init(&MDB_runtime_table);
	MDB_tabtype_Init(&MDB_backtime_table);

#ifndef TESTES	
	/* Check some modules dependencies */
	if (M_SW_add_table_init_check()){
		ERROR_MSG("M_SW_add_table must be initialized before this one");
		return -1;
	}
#endif

	MDB_Init_flag = 1;
	
#ifdef PRINT_MODULES_INIT
	PRINT_MSG(" -MDB_Init Ok");
#endif
	return 0;
}

signed char MDB_Close(void)
{
	MDB_tabtype_Close(&MDB_runtime_table);
	MDB_tabtype_Close(&MDB_backtime_table);
	MDB_Init_flag = 0;
#ifdef PRINT_MODULES_INIT
	PRINT_MSG(" -MDB_Close Ok");
#endif
	return 0;
}

unsigned char MDB_Init_check(void)
{
	return MDB_Init_flag;
}

// START - Search and Track
static MDB_RT_mesgtype *MDB_RT_track_mesg_id(MDB_RT_mesgtype *t_RT, unsigned short vid){
	MDB_RT_mesgtype *ret_p;
	while ( (ret_p = MDB_list_get_item_and_move_ahead(&t_RT))!=NULL ){
		if( FTT_VAR_ID_CMP_ID(ret_p->req.m_id, vid) )
			return ret_p;
	}
	return NULL;
}

static MDB_RT_mesgtype *MDB_RT_track_mesg_id_and_tag(MDB_RT_mesgtype *t_RT, unsigned short vid){
	MDB_RT_mesgtype *ret_p;
	while ( (ret_p = MDB_list_get_item_and_move_ahead(&t_RT))!=NULL ){
		if( FTT_VAR_ID_CMP_ID(ret_p->req.m_id, vid) ){
			while (ret_p->req.m_id != vid){
				if (ret_p->generic_p == NULL ) return NULL;
				ret_p=ret_p->generic_p;
			}
			return ret_p;
		}
	}
	return NULL;
}

static unsigned char MDB_RT_check_existence(MDB_RT_mesgtype *temp, unsigned int id){
	return ( (MDB_RT_track_mesg_id(temp, id)!=NULL) );
}

static MDB_RT_mesgtype **MDB_RT_track_mesg_id_pp(MDB_RT_mesgtype **t_RT, unsigned short vid){

	if (t_RT == NULL){
	       ERROR_MSG("Bad input poiter");
       	       return NULL;
	}
	while ( *t_RT != NULL ){
		if ( FTT_VAR_ID_CMP_ID((*t_RT)->req.m_id, vid) )
			break;
		t_RT = &((*t_RT)->link);
	}
	return t_RT;
}


static inline MDB_RT_mesgtype *MDB_RT_get_Free_item(void)
{
	MDB_RT_mesgtype *ret_p;
	if ( MDB_list_remove_from_top( 1, &MDB_free_list, &ret_p) < 0)
		return NULL;
	
	memset(ret_p, 0, sizeof(MDB_RT_mesgtype)); /* Cleaning just in case for any array comparison */
	return ret_p;
}

static inline signed char MDB_RT_Empty_list(MDB_RT_mesgtype **temp)
{
	MDB_RT_mesgtype **to;
//	/* Put in the tail of of free items */
	to = &MDB_free_list;
//	while( *to != NULL )
//		to = &((*to)->link);

	while ( !(MDB_list_move_Item_to(temp, to)<0) );
	return 0;
}

static inline signed char MDB_RT_Move_list_to(MDB_RT_mesgtype **temp, MDB_RT_mesgtype **to)
{
	while ( !(MDB_list_move_Item_to(temp, to)<0) );
	return 0;
}

//STOP - Search and Track
////////////////////////////////////////////////////////////////////

static signed char MDB_SM_validation( MESSAGE_REQUESITS *reqs )
	/* Parameters validation */
{
	/* Overrite the Type tag */
//	reqs->m_id = ( reqs->m_id & MESGID_MASK) | SDATA_MESG_ID;
	FTT_VAR_ID_SET_ID( reqs->m_id, reqs->m_id );
	FTT_VAR_ID_SET_TYPE_SYNC( reqs->m_id );

	if (reqs->m_relative_id<0)
		reqs->m_relative_id = -1;
	else{
		FTT_VAR_ID_SET_TYPE_SYNC(reqs->m_relative_id);
	}

	/* Validation */
	if((reqs->m_period <= 0)||(reqs->m_deadline <= 0)) return(-1);
//	if(aux_var->m_init < 0) return(-2);
	
	if ( reqs->m_size > reqs->m_max_size ){
		ERROR_MSG("Insufficient size for max_size, m_size is bigger");
		return -2;
	}

//	PRINT_MSG("MDB_SM_validation trying id:%d", reqs->m_size);
	if(reqs->m_max_size > 2000000) return(-3);
	if(reqs->m_MTU > FTT_PKT_SDM_DATALEN_MAX) return(-4);

	return 0;
}
static signed char MDB_AM_validation( MESSAGE_REQUESITS *reqs )
{
	/* Overrite the Type tag */
	//reqs->m_id = (reqs->m_id & MESGID_MASK) | ADATA_MESG_ID;
	FTT_VAR_ID_SET_ID( reqs->m_id, reqs->m_id );
	FTT_VAR_ID_SET_TYPE_ASYNC( reqs->m_id );

	/* Validate the message parameters */
	if((reqs->m_period <= 0)||(reqs->m_deadline <= 0)) return(-1);

	if ( reqs->m_size > reqs->m_max_size ){
		ERROR_MSG("Insufficient size for max_size, m_size is bigger");
		return -2;
	}
	
	if ( reqs->sw_producer_no > reqs->sw_producer_max_no ){
		ERROR_MSG("Insufficient size for sw_producer_max_no, sw_producer_no is bigger");
		return -2;
	}

//#warning NOT TESTED
//	if(reqs->m_size > FTT_PKT_ADM_DATALEN_MAX) return (-3);
//	reqs->m_MTU = FTT_PKT_ADM_DATALEN_MAX;

	return 0;
}


static inline unsigned char MDB_fragmentation_calculus(MESSAGE_REQUESITS *reqs, MESSAGE_FRAGMENTS *frag )
{
	int i;
	if ( FTT_VAR_ID_IS_ID_SYNC( reqs->m_id) ){
		/* SM */
		frag->packets_no = reqs->m_size / reqs->m_MTU;
		if ( (i=(reqs->m_size % reqs->m_MTU)) ){
			/* Last packet not full */
			frag->packets_no ++;
			if( i >= FTT_PKT_SDM_DATALEN_MIN)
				frag->real_size.Lastm_bsize = FTT_PKT_SDM_REALSIZE_B(i)*8;
			else
				frag->real_size.Lastm_bsize = FTT_PKT_SDM_REALSIZE_B(FTT_PKT_SDM_DATALEN_MIN)*8;
//				frag->real_size.Lastm_bsize = FRAME_REALSIZE_B(ETH_DATALEN_MIN)*8;
		}else{
			/* Last packet full */
			frag->real_size.Lastm_bsize = FTT_PKT_SDM_REALSIZE_B(reqs->m_MTU)*8;
		}
		frag->real_size.Onem_bsize = FTT_PKT_SDM_REALSIZE_B(reqs->m_MTU)*8;
		
		/* Due to the variable prop. time on the bus from station to station, and as well */
		/* software and hardware(NIC) induced jitter, an "jitter" factor must be append   */
		/* to all frames, that must be greater than (Network pro. time + Software jitter) */
		frag->real_size.Onem_bsize += FTT_IFS_bits;
		frag->real_size.Lastm_bsize += FTT_IFS_bits;
		frag->real_size.Onem_ssize = m_ceil(((double)(frag->real_size.Onem_bsize)*((double)BIT_TIME_US))/((double)SLOTV_US));
		frag->real_size.Lastm_ssize = m_ceil(((double)(frag->real_size.Lastm_bsize)*((double)BIT_TIME_US))/((double)SLOTV_US));
	}else{
		/* AM */
		frag->packets_no = reqs->m_size / reqs->m_MTU;
	if ( (i=(reqs->m_size % reqs->m_MTU)) ){
		/* Last packet not full */
		frag->packets_no ++;
		if( i >= FTT_PKT_ADM_DATALEN_MIN)
			frag->real_size.Lastm_bsize = FTT_PKT_ADM_REALSIZE_B(i)*8;
		else
			frag->real_size.Lastm_bsize = FTT_PKT_ADM_REALSIZE_B(FTT_PKT_ADM_DATALEN_MIN)*8;
//			frag->real_size.Lastm_bsize = FRAME_REALSIZE_B(ETH_DATALEN_MIN)*8;
	}else{
		/* Last packet full */
		frag->real_size.Lastm_bsize = FTT_PKT_ADM_REALSIZE_B(reqs->m_MTU)*8;
	}
	frag->real_size.Onem_bsize = FTT_PKT_ADM_REALSIZE_B(reqs->m_MTU)*8;

	/* Due to the variable prop. time on the bus from station to station, and as well */
	/* software and hardware(NIC) induced jitter, an "jitter" factor must be append   */
	/* to all frames, that must be greater than (Network pro. time + Software jitter) */
	frag->real_size.Onem_bsize += FTT_IFS_bits;
	frag->real_size.Lastm_bsize += FTT_IFS_bits;
	frag->real_size.Onem_ssize = m_ceil(((double)(frag->real_size.Onem_bsize)*((double)BIT_TIME_US))/((double)SLOTV_US));
	frag->real_size.Lastm_ssize = m_ceil(((double)(frag->real_size.Lastm_bsize)*((double)BIT_TIME_US))/((double)SLOTV_US));
	}
	return 0;
}

static inline signed char MDB_dynamic_param_set_default(MDB_RT_mesgtype *var )
{
	if ( !(var->req.m_relative_id < 0)){
		/* Then there is some relationship */
		ERROR_MSG("Not yet implemented...");
		return -1;
	}

	if (var->req.m_init>=0){
		/* Start from this point with this offset */
		var->dynamic.message_status = SM_WAITING;
		var->dynamic.activation_counter = var->req.m_init;
	}else{	
		/* Then no offset is required - in this case the message will
		 * start in OPERATION state. This is useful when changing
		 * variables. When replacing the SRDB plans a dynamic ajustment
		 * is made and here it's possible to mantain some dynamic parameters.
		 */
		var->dynamic.message_status = SM_OPERATIONAL;
		var->dynamic.activation_counter = 0;
	}

	var->dynamic.replenish_counter = 0;
	MESSAGE_DYNAMIC_DATA_queue_size_clean( &(var->dynamic) );
	
	
	return 0;
}
static inline signed char MDB_dynamic_param_set_changes( MDB_RT_mesgtype *var, MDB_RT_mesgtype *runtime_var )
{
	unsigned char abort_tx = 0;
	unsigned char i;
	/* This function is called when to instead of initializing a variable's dynamic requirements,
	 * update the fixed contents of the already running in runtime. That message will then be
	 * included in the backtime, and finally passed to runtime among with all others.
	 * 
	 * So here, instead of changing my item, I'll propagate the changes
	 * to the based item in the runtime table.
	 *
	 * Additionally, in certain circumstances, there's the need to abort this message from 
	 * transmission, if it's pending. Which circumstances?
	 *  - C changed
	 *  - Connection (Prod/Consumers) changed
	 *
	 * Remember that this function is being issued with the M_db_access locked, i.e there is
	 * no conflict with the Scheduler that might be dealing with the ready queue.
	 */

	/* First match the id's - just in case */
	if ( !FTT_VAR_ID_CMP_ID( var->req.m_id, runtime_var->req.m_id) ){
		ERROR_MSG("FATAL ERROR: CHECK THIS");
		return -1;
	}

	/* Check for different C */
	if ( var->req.m_size != runtime_var->req.m_size )
		abort_tx |= 0x01;
	
	/* No need Check for different T */

	/* Check if the var maximum size changed. This is important for the slaves
	 * It should generate an error since new sizes MUST go with a FORCE NEW 
	 * flag, an so this code is not called at all */

	/* Check for different Producer */
	if ( (var->req.sw_producer_idx)[0] != (runtime_var->req.sw_producer_idx)[0] )
		abort_tx |= 0x02;

	if ( var->req.sw_consumer_no != runtime_var->req.sw_consumer_no )
		abort_tx |= 0x04;
	
	/* Check for different Consumer */
	for (i=0; i<var->req.sw_consumer_no; i++)
		if ( var->req.sw_consumer_idx[i] != runtime_var->req.sw_consumer_idx[i] )
			abort_tx |= 0x08;

	PRINT_MSG("Not yet implemented ");

//	/* Let's now check if the variable (already running) is in the 
//	 * critical state SM_TRANSMITING and abort its transmission */
//	if ( (abort_tx) && (runtime_var->dynamic.message_status == SM_TRANSMITING) ){
//		PRINT_MSG("M_db: Aborting var 0x%x from SM_TRANSMITING state", var->req.m_id );
//		/* Remove the var from the ready queue*/
//		if ( FTT_VAR_ID_IS_ID_SYNC(var->req.m_id ) )
//			ready_queue_remove_ID(SRT_QUEUE, var->req.m_id);
//		else
//			switch (var->req.m_timeliness){
//				case TMLN_HARD: ready_queue_remove_ID(ART_H_QUEUE, var->req.m_id); break;
//				case TMLN_SOFT: ready_queue_remove_ID(ART_S_QUEUE, var->req.m_id); break;
//				case TMLN_UNCONST: ready_queue_remove_ID(ART_U_QUEUE, var->req.m_id); break;
//				default : ERROR_MSG("Wrong timeliness");
//			}
//	}
//	/* Let's only re-check the var status */
//	if (runtime_var->dynamic.message_status == SM_TRANSMITING){
//		ERROR_MSG("There's a BIG design/programming error");
//		/* This message should had be the heading the queue */
//	}

	return 0;

}

static inline signed char MDB_static_param_validate_modify( MDB_RT_mesgtype *new_var, MDB_RT_mesgtype *old_var )
{
	if (new_var->req.m_max_size != old_var->req.m_max_size)
		return -1;
	
	if (new_var->req.sw_producer_max_no != old_var->req.sw_producer_max_no) {
		return -2;
	}

	return 0;
}
signed char MDB_add_msg( MESSAGE_TYPE m_type, unsigned char tag, MESSAGE_REQUESITS *reqs, unsigned char overwrite_f)
/* returns:
 * 	< 0 on error
 * 	= 0 on no changes
 * 	= 1 on changed
 * 	= 2 on force changes
 */
{
	MDB_RT_mesgtype **list_back_header;
	MDB_RT_mesgtype **list_run_header;
	MDB_RT_mesgtype *my_item;
	MDB_RT_mesgtype *temp_p;
	int ret;

	//Check the backup lock flag
	if (!MDB_tabtype_IsLocked(&MDB_backtime_table)){
		ERROR_MSG("Baktime table was not locked for changes");
		return -1;
	}
	
	//PRINT_MSG("In MDB_add_msg adding var 0x%x tag:%d ovwrite:%d", reqs->m_id, tag, overwrite_f );

	//Switch on type and point to the propper list
	//Make the requisits validation
	switch (m_type){
		case SYNC_MESSAGE: /* Synchronous message  */
			list_back_header = &(MDB_backtime_table.SRT_mesgtable);
			list_run_header = &(MDB_runtime_table.SRT_mesgtable);
			if ( (ret = MDB_SM_validation(reqs)) < 0)
			{
				ERROR_MSG("MDB_SM_validation returned: %d", ret);
				return -2;
			}
			break;

		case ASYNC_MESSAGE:
			list_back_header = &(MDB_backtime_table.ART_mesgtable);
			list_run_header = &(MDB_runtime_table.ART_mesgtable);
			if ( (ret = MDB_AM_validation(reqs)) < 0)
			{
				ERROR_MSG("MDB_AM_validation returned: %d", ret);
				return -2;
			}
			break;

		default:
			ERROR_MSG("Wrong m_type");
			return -2;
			break;
	}
	

	//Check if ID already in this back list
	if ( MDB_RT_check_existence(*list_back_header, reqs->m_id) )
		return -3;

//	/* Check the relativeness */
//	if (aux_var->m_relative_id>=0)
//		if (!DB_RT_check_existence(t_ART, aux_var->m_relative_id))
//			return -5;

	//Get one item from the pool
	if ( (my_item = MDB_RT_get_Free_item()) == NULL ){
		ERROR_MSG("DB is probably full");
		return -4;
	}

	//Copy the requisits
	my_item->req = *reqs;

	//Complete the fragmentation parameters
	MDB_fragmentation_calculus(&(my_item->req), &(my_item->frag) );

	/* Insert it */
	MDB_list_add_this_point( list_back_header, my_item);

	my_item->generic_p = NULL;
	
	if (overwrite_f){
		FTT_VAR_ID_SET_TAG( my_item->req.m_id, tag); /* Stamp the tag */
		my_item->item_status = FORCE_NEW_ITEM;
		return 2;
	}

	/* Let's check the possibilty of keeping the old (if it exists) */

	/* Search runtime for replica 
		 *  -I am not locking since it's readonly access 
		 *  and I am not planning to go through the item dynamic part */
	if ( (my_item->generic_p = MDB_RT_track_mesg_id(*list_run_header, reqs->m_id)) != NULL ){
		/* Let's move to the most recent piggyback to get its reqs - the last one */
		temp_p = my_item->generic_p;
		while ( temp_p->generic_p != NULL)
			temp_p = temp_p->generic_p;
		/* temp_p points to the most recently updated item */
		
		/* We got a running item with the same ID - Let's compare */

		/* To speed up the comparison let's temporarily copy the tag stamp */
		FTT_VAR_ID_SET_TAG( my_item->req.m_id, FTT_VAR_ID_GET_TAG(temp_p->req.m_id) );
		if ( memcmp( &(temp_p->req), &(my_item->req), sizeof(MESSAGE_REQUESITS)) == 0 ){
			/* NO changes needed */
			/* Let's keep the tag as it was copied */
			my_item->item_status = PRESERVE_ITEM;
			return 0; // on no changes
		}else{
			FTT_VAR_ID_SET_TAG( my_item->req.m_id, tag); /* Stamp the tag */
			my_item->item_status = MODIFY_ITEM;

			/* Before exit let's just check if the system supports this change */
			if ( MDB_static_param_validate_modify( my_item, temp_p ) < 0)
				return -5;
			return 1; //on changes
		}
	}

	FTT_VAR_ID_SET_TAG( my_item->req.m_id, tag); /* Stamp the tag */
	my_item->item_status = FORCE_NEW_ITEM;
	return 2; //on force_changes!
	//return 1; //on changes!
}

unsigned char MDB_Change_locker_check(void)
{
	return MDB_tabtype_IsLocked(&MDB_backtime_table);
}

signed char MDB_Change_lock(void)
{
	MDB_tabtype_EnterLock(&MDB_backtime_table);

	/* Both the synch and asynch  lists should be already empty */
	if ((MDB_backtime_table.SRT_mesgtable!=NULL)||(MDB_backtime_table.SRT_mesgtable!=NULL))
		ERROR_MSG("Something is wrong. Programming problem!");
	return 0;
}

signed char MDB_Change_forget(void)
{
	if (!MDB_tabtype_IsLocked(&MDB_backtime_table)){
		ERROR_MSG("Baktime table was not locked for changes");
		return -1;
	}

	/* Let's free all the queue */
	MDB_RT_Empty_list(&(MDB_backtime_table.SRT_mesgtable));
	MDB_RT_Empty_list(&(MDB_backtime_table.ART_mesgtable));

	MDB_tabtype_LeaveLock(&MDB_backtime_table);
	return 0;
}

signed char MDB_Change_submit(unsigned char swap_plans)
{
	MDB_RT_mesgtype *temp_p, *i_p;
	MDB_RT_mesgtype **temp2_pp;

	if (!MDB_tabtype_IsLocked(&MDB_backtime_table)){
		ERROR_MSG("Baktime table was not locked for changes");
		return -1;
	}

	/* Lock the runtime */
	MDB_tabtype_EnterLock(&MDB_runtime_table);

#ifdef MASTER_PRINT_DB_UPDATES
PRINT_MSG("MDB_Change_submit details:");
PRINT_MSG("SM:");
#endif
	/* SM ***********************************************************/
	temp_p = MDB_backtime_table.SRT_mesgtable;
	
	/* For each item in the backup list */
	while ( (i_p = MDB_list_get_item_and_move_ahead(&temp_p)) != NULL ){
		switch(i_p->item_status){
			case FORCE_NEW_ITEM:
				/* Ignore the runtime item if there is one
				 * The runtime item will eventually be erased */
//				MDB_dynamic_param_set_default( i_p );
#ifdef MASTER_PRINT_DB_UPDATES
				PRINT_MSG("Forcing new------------------------------------%x", i_p->req.m_id);
#endif
				i_p->dynamic.activation_counter = (i_p->req.m_init>=0)? i_p->req.m_init : 0;
				i_p->dynamic.message_status = SM_OPERATIONAL;
				MESSAGE_DYNAMIC_DATA_queue_size_clean( &(i_p->dynamic) );

				i_p->item_status = ITEM_IS_IN_THE_MAIN_LIST;
				i_p->generic_p = NULL;
				break;
			case MODIFY_ITEM:
#ifdef MASTER_PRINT_DB_UPDATES
				PRINT_MSG("Modifying-------%x will be piggy to %x---------", i_p->req.m_id, i_p->generic_p->req.m_id);
#endif
				/* Remove the back item from the backtime table */
				MDB_list_remove_from_top( 1, MDB_list_track_pointer( &(MDB_backtime_table.SRT_mesgtable), i_p ), NULL);

				/* Bring the running item to back and piggyback there the one herein created */
				MDB_list_move_Item_to( MDB_list_track_pointer( &(MDB_runtime_table.SRT_mesgtable), i_p->generic_p ), &(MDB_backtime_table.SRT_mesgtable) );				
				i_p->generic_p->item_status = ITEM_IS_IN_THE_MAIN_LIST; //The runtime is enforced

				temp2_pp = &(i_p->generic_p->generic_p); //pointer to the generic_p in the running one
				/* This is the point to piggyback, but We have to piggyback in the tail (newest) */
				while ( (*temp2_pp) != NULL )
					temp2_pp = &( (*temp2_pp)->generic_p );

				/* Now insert here the backtime item - recently removed from the list */
				*temp2_pp = i_p;
				i_p->generic_p = NULL; //I lost the reference to the runtime, now temporarily in the backtime
				i_p->item_status = ITEM_IS_PIGGYBACKED;
				MESSAGE_DYNAMIC_DATA_queue_size_clean( &(i_p->dynamic) ); //This is the only dynamic data that will be used when piggybacked
				
				break;
			case PRESERVE_ITEM:
#ifdef MASTER_PRINT_DB_UPDATES
				PRINT_MSG("Preserving-------------------------------------%x", i_p->req.m_id);
#endif
				/* Keep the running item intact - bring it to back */
				
				/* Erase my back item */
				MDB_list_move_Item_to( MDB_list_track_pointer( &(MDB_backtime_table.SRT_mesgtable), i_p ), &MDB_free_list);

				/* Move the running one to back, so that it becomes running when swapping the plans */
				MDB_list_move_Item_to( MDB_list_track_pointer( &(MDB_runtime_table.SRT_mesgtable), i_p->generic_p ), &(MDB_backtime_table.SRT_mesgtable) );
				
				i_p->generic_p->item_status = ITEM_IS_IN_THE_MAIN_LIST;
				
				break;

			default:
				ERROR_MSG("Unhandled state - design or programming error");
				break;
		}
	}

#ifdef MASTER_PRINT_DB_UPDATES
PRINT_MSG("AM:");
#endif
	/* AM ***********************************************************/
	temp_p = MDB_backtime_table.ART_mesgtable;
	
	/* For each item in the backup list */
	while ( (i_p = MDB_list_get_item_and_move_ahead(&temp_p)) != NULL ){

		switch(i_p->item_status){
			case FORCE_NEW_ITEM:
#ifdef MASTER_PRINT_DB_UPDATES
				PRINT_MSG("Forcing new------------------------------------%x", i_p->req.m_id);
#endif
				/* Ignore the runtime item if there is one
				 * The runtime item will eventually be erased */
//				MDB_dynamic_param_set_default( i_p );
				i_p->dynamic.activation_counter = (i_p->req.m_init>=0)? i_p->req.m_init : 0;
				i_p->dynamic.message_status = SM_OPERATIONAL;
				MESSAGE_DYNAMIC_DATA_queue_size_clean( &(i_p->dynamic) );


				i_p->item_status = ITEM_IS_IN_THE_MAIN_LIST;
				i_p->generic_p = NULL;
				break;
			case MODIFY_ITEM:
#ifdef MASTER_PRINT_DB_UPDATES
				PRINT_MSG("Modifying-------%x will be piggy to %x---------", i_p->req.m_id, i_p->generic_p->req.m_id);
#endif
			
				/* Remove the back item from the backtime table */
				MDB_list_remove_from_top( 1, MDB_list_track_pointer( &(MDB_backtime_table.ART_mesgtable), i_p ), NULL);

				/* Bring the running item to back and piggyback there the one herein created */
				MDB_list_move_Item_to( MDB_list_track_pointer( &(MDB_runtime_table.ART_mesgtable), i_p->generic_p ), &(MDB_backtime_table.ART_mesgtable) );
				i_p->generic_p->item_status = ITEM_IS_IN_THE_MAIN_LIST; //The runtime is enforced

				temp2_pp = &(i_p->generic_p->generic_p); //pointer to the generic_p in the running one
				/* This is the point to piggyback, but We have to piggyback in the tail (newest) */
				while ( (*temp2_pp) != NULL )
					temp2_pp = &( (*temp2_pp)->generic_p );

				/* Now insert here the backtime item - recently removed from the list */
				*temp2_pp = i_p;
				i_p->generic_p = NULL; //I lost the reference to the runtime, now temporarily in the backtime
				i_p->item_status = ITEM_IS_PIGGYBACKED;

				MESSAGE_DYNAMIC_DATA_queue_size_clean( &(i_p->dynamic) ); //This is the only dynamic data that will be used when piggybacked

			
				break;
			case PRESERVE_ITEM:
#ifdef MASTER_PRINT_DB_UPDATES
				PRINT_MSG("Preserving-------------------------------------%x", i_p->req.m_id);
#endif
				/* Keep the running item intact - bring it to back */
				
				/* Erase my back item */
				MDB_list_move_Item_to( MDB_list_track_pointer( &(MDB_backtime_table.ART_mesgtable), i_p ), &MDB_free_list);

				/* Move the running one to back, so that it becomes running when swapping the plans */
				MDB_list_move_Item_to( MDB_list_track_pointer( &(MDB_runtime_table.ART_mesgtable), i_p->generic_p ), &(MDB_backtime_table.ART_mesgtable) );
				
				i_p->generic_p->item_status = ITEM_IS_IN_THE_MAIN_LIST;
				
				break;

			default:
				ERROR_MSG("Unhandled state - design or programming error");
				break;
		}
	}
	
	/******************************************************************/

	/* Depending on the update method */
	if (swap_plans){
		/* Move all the items that remain in the runtime to the free items */
		MDB_RT_Empty_list( &(MDB_runtime_table.SRT_mesgtable) );
		MDB_RT_Empty_list( &(MDB_runtime_table.ART_mesgtable) );

		/* SM swap the backtime and the runtime pointers */
		temp_p = MDB_runtime_table.SRT_mesgtable;
		MDB_runtime_table.SRT_mesgtable = MDB_backtime_table.SRT_mesgtable;
		MDB_backtime_table.SRT_mesgtable = temp_p;

		/* AM swap the backtime and the runtime pointers */
		temp_p = MDB_runtime_table.ART_mesgtable;
		MDB_runtime_table.ART_mesgtable = MDB_backtime_table.ART_mesgtable;
		MDB_backtime_table.ART_mesgtable = temp_p;
	}else{
		/* Merge plans backtime messages to runtime */
		MDB_RT_Move_list_to( &(MDB_backtime_table.SRT_mesgtable), &(MDB_runtime_table.SRT_mesgtable));
		MDB_RT_Move_list_to( &(MDB_backtime_table.ART_mesgtable), &(MDB_runtime_table.ART_mesgtable));
	
	}

	MDB_tabtype_LeaveLock(&MDB_runtime_table);
	MDB_tabtype_LeaveLock(&MDB_backtime_table);

	return 0;
}


/** Now the functions for the **/

signed char MDB_SRT_get_runtime_activations( READY_RETURN_ITEMS *array_return_p, unsigned short *num_ret )
{
	MDB_RT_mesgtype *temp_p, *i_p;
	unsigned short array_max = *num_ret;
	/* num_ret is used as input and as output */

	*num_ret=0;

	/* Check runtime flag */
	if (!MDB_tabtype_IsLocked(&MDB_runtime_table)){
		ERROR_MSG("Runtime table was not locked for changes");
		return -1;
	}
	/* For all the SM in the runtime backbone */
	temp_p = MDB_runtime_table.SRT_mesgtable;
	while ( (i_p = MDB_list_get_item_and_move_ahead(&temp_p)) != NULL ){

		switch (i_p->dynamic.message_status){
			case SM_OPERATIONAL:
			case SM_WAITING:
				if (i_p->dynamic.activation_counter<=0){
					i_p->dynamic.message_status = SM_SCHEDULED;
					/* Let's now seek the last Piggyback (newest tag), if any */
					while (i_p->generic_p != NULL)
						i_p = i_p->generic_p;

					/* Let's now activate this slot - append to return list */
					if (*num_ret == array_max) return -2;
					array_return_p[*num_ret].message_p = i_p;
					array_return_p[*num_ret].pakets_no = i_p->frag.packets_no;
					
					/* Important note! - Not adding the source_nodeID for now. Not implemented in the synch traffic */			
					array_return_p[*num_ret].source_nodeID = (i_p->req.sw_producer_idx)[0];
					
					(*num_ret)++;
				}
				break;
			case SM_SCHEDULED:
			case SM_TRANSMITING:
				/* Well, probably occured a missed deadline (if D<=T) or overrun*/
				//ERROR_MSG("Unexpected situation");
				break;
			default:
				ERROR_MSG("This state doesn't exist...");
		}
	}
	return 0;
}

signed char MDB_ART_get_runtime_activations(  READY_RETURN_ITEMS *array_return_p, unsigned short *num_ret )
{
	/* num_ret is used as input and as output */
	
	MDB_RT_mesgtype *temp_p, *i_p;
	unsigned short array_max = *num_ret;

	unsigned char ret_source_nodeID;
	unsigned short ret_queue_size;

	*num_ret=0;

	/* Check runtime flag */
	if (!MDB_tabtype_IsLocked(&MDB_runtime_table)){
		ERROR_MSG("Runtime table was not locked for changes");
		return -1;
	}

	/* For all the AM in the runtime backbone */
	temp_p = MDB_runtime_table.ART_mesgtable;
	while ( (i_p = MDB_list_get_item_and_move_ahead(&temp_p)) != NULL ){
		
		/* For each slot with this i_p */
		while (i_p != NULL){

			/* Check pending activations */
			while ( (ret_queue_size = MESSAGE_DYNAMIC_DATA_queue_size_check_and_decrement( &(i_p->dynamic), &ret_source_nodeID )) ){
				/* Handle a bunch of requests from a given source_nodeID */
				
				while (ret_queue_size--) {
					/* One by one */
					if (*num_ret == array_max){ /* Roll back */
						MESSAGE_DYNAMIC_DATA_queue_size_increment( &(i_p->dynamic), ret_source_nodeID, ret_queue_size+1 /* +1 because the while decrements */ );
						return -2;
					}
					
					/* Append to the return structure */
					array_return_p[*num_ret].message_p = i_p;
					array_return_p[*num_ret].pakets_no = i_p->frag.packets_no;
					switch (i_p->req.m_timeliness){
						case TMLN_HARD:    array_return_p[*num_ret].am_type = TMLN_HARD; break;
						case TMLN_SOFT:    array_return_p[*num_ret].am_type = TMLN_SOFT; break;
						case TMLN_UNCONST: array_return_p[*num_ret].am_type = TMLN_UNCONST; break;
						default: ERROR_MSG("Not a defined constraint");
					}
					array_return_p[*num_ret].source_nodeID = ret_source_nodeID;
					(*num_ret)++;
				}
			}
			
			
			i_p = i_p->generic_p;
		}
	}
	return 0;
}

unsigned char MDB_SRT_downgrade_runtime_activations( void )
{
	MDB_RT_mesgtype *temp_p, *i_p;
	unsigned char flag=0;

	/* Check runtime flag */
	if (!MDB_tabtype_IsLocked(&MDB_runtime_table)){
		ERROR_MSG("Runtime table was not locked for changes");
		return -1;
	}

	/* For all the messages in runtime */
	temp_p = MDB_runtime_table.SRT_mesgtable;
	while ( (i_p = MDB_list_get_item_and_move_ahead(&temp_p)) != NULL ){

		/* Decrease the dynamic counter */	
		i_p->dynamic.activation_counter --;
	
		if ( i_p->dynamic.activation_counter <= -(signed int)(i_p->req.m_period)){
			flag = 1;
			PRINT_MSG("A deadline miss was detected in message id: %x", i_p->req.m_id);
		}
	}

	return (flag);
}
unsigned char MDB_ART_downgrade_runtime_activations( void )
{
	MDB_RT_mesgtype *temp_p, *i_p;
	unsigned char flag=0;

	/* For all the messages in runtime */
	temp_p = MDB_runtime_table.ART_mesgtable;
	while ( (i_p = MDB_list_get_item_and_move_ahead(&temp_p)) != NULL ){

		/* Decrease the dynamic counter */	
		if (i_p->dynamic.activation_counter) 
			i_p->dynamic.activation_counter --;

//		PRINT_MSG("async activ. counter: %d",i_p->dynamic.activation_counter);
	}
	return (flag);
}

void MDB_runtime_print(void)
{
	PRINTF("\nM_DB runtime:");	
	
	MDB_tabtype_print(&MDB_runtime_table,1);
}
void MDB_backtime_print(void)
{
	PRINTF("\nM_DB backtime:");	
	MDB_tabtype_print(&MDB_backtime_table,1);
}

void MDB_Access_Enter_and_Lock(void)
{
	MDB_tabtype_EnterLock(&MDB_runtime_table);
}

void MDB_Access_UnLock(void)
{
	MDB_tabtype_LeaveLock(&MDB_runtime_table);
}

#ifdef NODE_ASYNC_REPLY
signed short MDB_ART_update_status (unsigned short id, unsigned char source_nodeID, unsigned short update_size)
{
	MDB_RT_mesgtype *temp_p;

	//if (update_size) PRINT_MSG("-> 0x%04x  from: %d with:%d", id, source_nodeID, update_size);
	
	/* Check runtime flag */
	if (!MDB_tabtype_IsLocked(&MDB_runtime_table)){
		/* Enter mutex region - access must be serialized with respect to the scheduler */
		/* This function will be called on every Message status reception */
		ERROR_MSG("Runtime table was not locked for changes");
		return -1;
	}

	if (update_size){
		if ( (temp_p = MDB_RT_track_mesg_id_and_tag( (MDB_runtime_table.ART_mesgtable), id)) == NULL )
			return -1;

		MESSAGE_DYNAMIC_DATA_queue_size_increment( &(temp_p->dynamic) , source_nodeID, update_size );
	}

	return(0);
}

#endif

MESSAGE_REQUESITS *MDB_Get_runtime_info_reqs( void *mesgtype_p )
{
	MDB_RT_mesgtype *in_p = mesgtype_p;
	/* Check runtime flag */
	if (!MDB_tabtype_IsLocked(&MDB_runtime_table)){
		ERROR_MSG("Runtime table was not locked for changes");
	}

	return &(in_p->req);
}

MESSAGE_FRAGMENTS *MDB_Get_runtime_info_frag( void *mesgtype_p )
{
	MDB_RT_mesgtype *in_p = mesgtype_p;
	/* Check runtime flag */
	if (!MDB_tabtype_IsLocked(&MDB_runtime_table)){
		ERROR_MSG("Runtime table was not locked for changes");
	}

	return &(in_p->frag);
}

MESSAGE_DYNAMIC_DATA *MDB_Get_runtime_info_dynamic( void *mesgtype_p )
{
	MDB_RT_mesgtype *in_p = mesgtype_p;
	/* Check runtime flag */
	if (!MDB_tabtype_IsLocked(&MDB_runtime_table)){
		ERROR_MSG("Runtime table was not locked for changes");
	}

	return &(in_p->dynamic);
}

void MDB_Update_activation( void *mesgtype_p )
{
	MDB_RT_mesgtype *in_p = mesgtype_p;
	MDB_RT_mesgtype **temp_pp;
	MDB_RT_mesgtype *temp_p;

	/* Check runtime flag */
	if (!MDB_tabtype_IsLocked(&MDB_runtime_table)){
		ERROR_MSG("Runtime table was not locked for changes");
	}
	
	/* Another instant of this var left the ready queue */

	/* If item is NOT piggybacked 
	 *   - update counter and status */
	if ( in_p->item_status == ITEM_IS_IN_THE_MAIN_LIST ){
		in_p->dynamic.activation_counter += in_p->req.m_period;
		in_p->dynamic.message_status = SM_OPERATIONAL;
	}

	/* If it's piggypacked - means that the older have executed all they had to execute
	 * (time to say good bye to the old and embrasse the new one by refreshing the counter)
	 *   - search in the backbone list for this tagless id */
	if ( in_p->item_status == ITEM_IS_PIGGYBACKED ){
		//PRINT_MSG("\t-is piggybacked %x", in_p->req.m_id);

		/* seek in the backbone list - tagless */
		if (FTT_VAR_ID_IS_ID_SYNC( in_p->req.m_id ))
			temp_pp = MDB_RT_track_mesg_id_pp( &(MDB_runtime_table.SRT_mesgtable), in_p->req.m_id );
		else
			temp_pp = MDB_RT_track_mesg_id_pp( &(MDB_runtime_table.ART_mesgtable), in_p->req.m_id );

		if ( temp_pp == NULL ){
			ERROR_MSG("Table is inconsistent - seeking an item that should be there");
			return;
		}
		if ( *temp_pp == NULL ){
			ERROR_MSG("Table is inconsistent - seeking an item that should be there");
			return;
		}

		/* temp_pp is pointing to the pointer of the tagless id slot in the backbone */
		temp_p = *temp_pp;

		/* While not reached the once piggybacked item */
		while ( temp_p != in_p ){
			if ( temp_p->generic_p == NULL ){
				ERROR_MSG("Table is inconsistent - seeking an item that should be there");
				return;
			}

			/* It might be that this tag still holds messages from another source_nodeID */
			if (temp_p->dynamic.queue_total_size) {
				ERROR_MSG("Shouldn't occur!!!! skipping tag removal in the Master");
				continue;
			}
			
			/* Copy the necessary dynamic data to the new slot */
			temp_p->generic_p->dynamic.activation_counter = temp_p->dynamic.activation_counter;
			temp_p->generic_p->dynamic.replenish_counter = temp_p->dynamic.replenish_counter;

			/* Desrtoy the old - remember we have two lists: the backbone(link) and the paralel(generic_p) */
			
			/* In the backlist: 
			 *  - Add the temp_p->generic_p in front of temp_p 
			 *  - remove temp_p to free 
			 * In the paralel:
			 * - nothing is necessary 'cause the old item was sent to free 
			 * - I could NULL it's generic */
			MDB_list_add_this_point( &(temp_p->link), temp_p->generic_p );
			temp_p->generic_p->item_status = ITEM_IS_IN_THE_MAIN_LIST;
			MDB_list_move_Item_to( temp_pp, &MDB_free_list);
	
			if ( M_master_ch_Publish_del_old_tag( 0, (FTT_VAR_ID_IS_ID_SYNC( in_p->req.m_id )!=0)?SYNC_MESSAGE:ASYNC_MESSAGE , temp_p->req.m_id ) < 0 ){
				ERROR_MSG("Couldn't publish to delete tag: 0x%04x", temp_p->req.m_id);
			}

			temp_p = temp_p->generic_p;
			/* temp_pp 
			 *  - is pointing to a pointer that points to the previously piggybacked, i.e, the same 
			 * temp_p was updated */
			//PRINT_MSG("Was piggy---")
			//MDB_runtime_print();
		}
		
		/* Update the activation counter */
		in_p->dynamic.activation_counter += in_p->req.m_period;
		in_p->dynamic.message_status = SM_OPERATIONAL;
		
	}
}

static inline unsigned char MDB_Is_this_node_related_to_var(unsigned char nodeID, MESSAGE_REQUESITS *req )
{
	unsigned short i;

	/* Check if nodeID is producer or explicit consumer */
//	if ( req->sw_producer_idx == nodeID )
//		return 1;
	for (i=0; i<(req->sw_producer_no); i++){
		if ( req->sw_producer_idx[i] == nodeID )
			return 1;
	}
	

	for (i=0; i<(req->sw_consumer_no); i++){
		if ( req->sw_consumer_idx[i] == nodeID )
			return 2;
	}
	return 0;
}

signed char MDB_Dump_back_runtime_reqs( signed short nodeID, MDB_DUMP_BACK_ITEM *array_p, unsigned short *no )
{
	unsigned short max=*no;
	MDB_RT_mesgtype *temp_p, *i_p;
	
	*no=0;

	/* Check runtime flag */
	if (!MDB_tabtype_IsLocked(&MDB_runtime_table)){
		ERROR_MSG("Runtime table was not locked for changes");
		return -1;
	}

	temp_p = MDB_runtime_table.SRT_mesgtable;
	while ( (i_p = MDB_list_get_item_and_move_ahead(&temp_p)) != NULL ){
		if (nodeID >= 0){
			/* check if this node as anything to do with this var */
			if( !MDB_Is_this_node_related_to_var( (unsigned char)nodeID, &(i_p->req) ) ){
				continue; //Do not dump to the list
			}
		}
		if ( (*no) == max ) return -2;
		array_p[*no].reqs = &(i_p->req);
		array_p[*no].id.type = SYNC_MESSAGE;
		array_p[*no].id.id = FTT_VAR_ID_GET_ID(i_p->req.m_id);
		array_p[*no].tag = FTT_VAR_ID_GET_TAG(i_p->req.m_id);
		(*no)++;
	}
	
	temp_p = MDB_runtime_table.ART_mesgtable;
	while ( (i_p = MDB_list_get_item_and_move_ahead(&temp_p)) != NULL ){
		if (nodeID >= 0){
			/* check if this node as anything to do with this var */
			if( !MDB_Is_this_node_related_to_var( (unsigned char)nodeID, &(i_p->req) ) ){
				continue; //Do not dump to the list
			}
		}
		if ( (*no) == max ) return -2;
		array_p[*no].reqs = &(i_p->req);
		array_p[*no].id.type = ASYNC_MESSAGE;
		array_p[*no].id.id = FTT_VAR_ID_GET_ID(i_p->req.m_id);
		array_p[*no].tag = FTT_VAR_ID_GET_TAG(i_p->req.m_id);
		(*no)++;
	}

	return 0;
}

