/*****************************************************************************
 * M_ftt-core.layer.c:
 *****************************************************************************
 * Copyright (C) 2006-2012 the FTT-SE team.
 *
 * Author: Ricardo Marau <marau at fe.up.pt>
 *
 * FTT-SE is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  FTT-SE is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with FTT-SE.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

#define M_FTT_CORE_LAYER_C

#include "Ports.h"

#include "ethernet.layer.h"
#include "S_ftt-core.layer_embed.h"

#include "ftt-core.layer_types.h"

#include "M_Dispatcher.h"
#include "M_ECreg.h"
#include "M_SW_add_table.h"
#include "M_Scheduler.h"
#include "M_db.h"
#include "M_Ready_queue.h"
#include "M_eth_filter.h"
#include "M_master_ch.h"
#include "M_SRDB.h"
#include "M_appPnP.h"

#include "math_tools.h"

static signed int g_master_nodeID;

static unsigned long int g_master_ec_time; //us

signed char M_FTT_CORE_L_init(unsigned char *mac, unsigned long ec_time)
{
	int ret;
	
#ifdef MASTER_POLITE
	PRINT_MSG("Initializing the Core layer - Master");
#endif
	/* Check if Ethernet was initialized */
	if ( ETH_L_Init_check() < 0 ){
		ERROR_MSG("Init Fatal Error: ETH_L_Init_check failed");
		return -1;
	}

	g_master_ec_time = ec_time;

	M_SW_Init();
	g_master_nodeID = M_SW_subscribe_node( mac, 0  );

	if ((ret=M_appPnP_Init())<0){
		ERROR_MSG("M_appPnP_Init returned: %d",ret);
		return -2;
	}
	
	MDB_Init( mac, ec_time );

	M_SRDB_Init();

	Ready_queues_Init();
	ECreg_Init();

	SchedulerInit( ec_time );

	M_eth_filter_Init();
	
	/* Launch the slaves core */
	if ( (ret=S_FTT_CORE_L_init(mac,ec_time)) < 0){
		ERROR_MSG("S_FTT_CORE_L_init returned: %d",ret);
		return -3;
	}
	/* Let's emulate the Slave module NodeID discovery and set it */
	S_FTT_CORE_L_Set_NodeID((unsigned char)g_master_nodeID);
	
	if ( (ret=M_master_ch_Init(g_master_nodeID)) < 0){
		ERROR_MSG("M_master_ch_Init returned: %d",ret);
		return -4;
	}

	//PRINT_MSG ("Master subscribed nodeID %d", g_master_nodeID);
	S_nodeID_write( g_master_nodeID );
#ifdef RTL_PRO
	M_DispatcherInit(mac,ec_time);
#else
	M_DispatcherInit(mac,ec_time*L26_SLOWING); //L26_SLOWING times slower, but the scheduling is the same
#endif
	WakeSheduler();

#ifdef PRINT_MODULES_INIT
	PRINT_MSG(" -Starting FTT-SE core system...");
#endif
	M_DispatcherStartSystem();

//	MDB_runtime_print();
//	M_SRDB_Hardcoded_Table_Print();
	return 0;
}

signed char M_FTT_CORE_L_close(void) {
#ifdef MASTER_POLITE
	PRINT_MSG("Closing the Core layer - Master");
#endif
	M_DispatcherClose();

	M_master_ch_Close();
	S_FTT_CORE_L_close();
	M_eth_filter_Close();
	SchedulerClose();
	M_SRDB_Close();
	MDB_Close();
	M_appPnP_Close();
	return 0;
}

signed char M_FTT_CORE_L_get_appPnP( signed short *app_id, unsigned char *nodeID)
{
	return M_appPnP_Get_request_blocking( app_id, nodeID);
}

signed char M_FTT_CORE_L_notify_appPnP_OK( unsigned char app_id, unsigned char nodeID, unsigned short id_tx, unsigned short id_rx)
{
	return M_appPnP_notify_Slave( 1, app_id, nodeID, id_tx, id_rx);
}

signed char M_FTT_CORE_L_notify_appPnP_KO( unsigned char app_id, unsigned char nodeID)
{
	return M_appPnP_notify_Slave( 0, app_id, nodeID, 0, 0);
}

signed int M_FTT_CORE_L_get_master_nodeID(void)
{
	return g_master_nodeID;
}

signed int M_FTT_CORE_L_get_hardcoded_table( MESSAGES_TABLE **table )
{
	return M_SRDB_get_hardcoded_table( table );
}


signed char M_FTT_CORE_L_update_requirements(MESSAGES_TABLE *dyn_table)
{
	return M_SRDB_Update(dyn_table);
}

/* This function is used to get from this layer the load in Bps that one message might have */
double M_FTT_CORE_L_get_message_load( MESSAGE_TYPE type, unsigned int size, unsigned int period, unsigned int MTU ) //bps
{
	unsigned short packets_MTU;
	unsigned short last_packet_size;
	unsigned long int var_size;
	double period_sec;

	packets_MTU = size/MTU;
	last_packet_size = size%MTU;

//	PRINT_MSG("M_FTT_CORE_L_get_message_load(size:%d period:%d MTU:%d)",size, period, MTU);
	switch(type){
		case SYNC_MESSAGE:
			var_size = FTT_PKT_SDM_REALSIZE_B(MTU)*packets_MTU + FTT_PKT_SDM_REALSIZE_B(last_packet_size);
		case ASYNC_MESSAGE:
		default:
			var_size = FTT_PKT_ADM_REALSIZE_B(MTU)*packets_MTU + FTT_PKT_ADM_REALSIZE_B(last_packet_size);
	}

	var_size = var_size * 8;

//	PRINT_MSG("   -var_size:%d bits", var_size);

	/* Var period in seconds */
//	PRINT_MSG("   -g_master_ec_time:%d  ", g_master_ec_time);
//	PRINT_MSG("   -e-6:%g  ", 1.0e-6);
//	PRINT_MSG("   -g_master_ec_timee-6:%g  ", (double)g_master_ec_time*1e-6);
	period_sec = (double)(period) * g_master_ec_time/*us*/ * 1e-6 ;
//	PRINT_MSG("   -period_sec:%g  ", period_sec);
	return ((double)var_size)/( period_sec );
}

unsigned int M_FTT_CORE_L_get_message_bytesize( MESSAGE_TYPE type, unsigned int size, unsigned int MTU )
{
	unsigned short packets_MTU;
	unsigned short last_packet_size;
	unsigned long int var_size;

	packets_MTU = size/MTU;
	last_packet_size = size%MTU;

	switch(type){
		case SYNC_MESSAGE:
			var_size = FTT_PKT_SDM_REALSIZE_B(MTU)*packets_MTU + FTT_PKT_SDM_REALSIZE_B(last_packet_size);
		case ASYNC_MESSAGE:
		default:
			var_size = FTT_PKT_ADM_REALSIZE_B(MTU)*packets_MTU + FTT_PKT_ADM_REALSIZE_B(last_packet_size);
	}
	return (var_size);
}

double M_FTT_CORE_L_get_message_max_pkt_iit_load( MESSAGE_TYPE type, unsigned int size, unsigned int MTU ) //bps
{
	unsigned short packets_MTU;
	unsigned short last_packet_size;
	unsigned long int var_size;
	double period_sec;

	packets_MTU = size/MTU;
	last_packet_size = size%MTU;

//PRINT_MSG("M_FTT_CORE_L_get_message_max_pkt_iit_load ( %d, %d, %d )",type,size,MTU );

	switch(type){
		case SYNC_MESSAGE:
			if (packets_MTU != 0) var_size = FTT_PKT_SDM_REALSIZE_B(MTU);
			else var_size = FTT_PKT_SDM_REALSIZE_B(last_packet_size);
		case ASYNC_MESSAGE:
		default:
			if (packets_MTU != 0) var_size = FTT_PKT_ADM_REALSIZE_B(MTU);
			else var_size = FTT_PKT_ADM_REALSIZE_B(last_packet_size);
	}
//PRINT_MSG("  var_size: %d", var_size);
	var_size = var_size * 8;
//PRINT_MSG("  var_size: %d", var_size);

	period_sec = (double) ((double)g_master_ec_time/*us*/ * 1e-6) ; //every EC
//PRINT_MSG("  g_master_ec_time: %d", g_master_ec_time );
//	PRINT_MSG("  M_FTT_CORE_L_get_message_max_pkt_iit_load returning: %g   ",var_size);
	return ((double)var_size)/( period_sec );
}


unsigned int M_FTT_CORE_L_get_max_per_pkt_bytesize( MESSAGE_TYPE type, unsigned int size, unsigned int MTU )
{
	unsigned short packets_MTU;
	unsigned short last_packet_size;
	unsigned long int var_size;

	packets_MTU = size/MTU;
	last_packet_size = size%MTU;

	switch(type){
		case SYNC_MESSAGE:
			if (packets_MTU != 0) var_size = FTT_PKT_SDM_REALSIZE_B(MTU);
			else var_size = FTT_PKT_SDM_REALSIZE_B(last_packet_size);
		case ASYNC_MESSAGE:
		default:
			if (packets_MTU != 0) var_size = FTT_PKT_ADM_REALSIZE_B(MTU);
			else var_size = FTT_PKT_ADM_REALSIZE_B(last_packet_size);
	}

	return (var_size);
}

unsigned long int M_FTT_CORE_L_get_EC_time( void ) //usec
{
	return g_master_ec_time/*us*/;
}

unsigned int M_FTT_CORE_L_get_corresponding_period_ticks( MESSAGE_TYPE type, unsigned int size, unsigned int MTU, double BW_load /*bps*/ )
{
	unsigned short packets_MTU;
	unsigned short last_packet_size;
	unsigned long int var_size;
	double period_sec;

	packets_MTU = size/MTU;
	last_packet_size = size%MTU;
	
	switch(type){
		case SYNC_MESSAGE:
			var_size = FTT_PKT_SDM_REALSIZE_B(MTU)*packets_MTU + FTT_PKT_SDM_REALSIZE_B(last_packet_size);
		case ASYNC_MESSAGE:
		default:
			var_size = FTT_PKT_ADM_REALSIZE_B(MTU)*packets_MTU + FTT_PKT_ADM_REALSIZE_B(last_packet_size);
	}

	var_size = var_size*8;

//	PRINT_MSG("var size:%d", size);
//	PRINT_MSG("var_size:%d", var_size);
//	PRINT_MSG("the BW_load:%g", BW_load*1e-6);
	period_sec = ((double)var_size)/( BW_load );
//	PRINT_MSG("Period_sec:%g", period_sec);

//	PRINT_MSG("Should return:%g", (period_sec * 1e6/( g_master_ec_time/*us*/ )) );

	return m_ceil(period_sec * 1e6/( g_master_ec_time/*us*/ ));
}

unsigned int M_FTT_CORE_L_get_corresponding_size_bytes( MESSAGE_TYPE type, unsigned int period, unsigned int MTU, double BW_load /*bps*/ )
{
	unsigned short packets_MTU;
	unsigned short last_packet_size;
	unsigned long int var_size;
	unsigned long int var_size_B;
	double period_sec;

//PRINT_MSG("M_FTT_CORE_L_get_corresponding_size_bytes\nperiod:%d", period);
	period_sec = (double)(period) * g_master_ec_time/*us*/ * 1e-6 ;
//	PRINT_MSG("Period_sec:%g", period_sec);
//	PRINT_MSG("the BW_load:%g", BW_load*1e-6);


	/*get var size in bits */
	var_size = (unsigned int) (BW_load * period_sec);
	var_size_B = var_size/8;
//PRINT_MSG("var_size_B:%d", var_size_B);

	switch(type){
		case SYNC_MESSAGE:
			packets_MTU = var_size_B/FTT_PKT_SDM_REALSIZE_B(MTU);
//PRINT_MSG("packets_MTU_no: %d", packets_MTU);
			last_packet_size = var_size_B%(FTT_PKT_SDM_REALSIZE_B(MTU));
//PRINT_MSG("last_packet_size: %d", last_packet_size);
//PRINT_MSG("FTT_PKT_SDM_REALSIZE_B_reverse(last_packet_size): %g", FTT_PKT_SDM_REALSIZE_B_reverse(last_packet_size));
//PRINT_MSG("MTU*packets_MTU: %g", MTU*packets_MTU );

//PRINT_MSG("Should return:%g", ((double)MTU*(double)packets_MTU + (double)FTT_PKT_SDM_REALSIZE_B_reverse(last_packet_size)) );
			return m_floor((double)MTU*(double)packets_MTU + (double)FTT_PKT_SDM_REALSIZE_B_reverse(last_packet_size));
		case ASYNC_MESSAGE:
		default:
			packets_MTU = var_size_B/FTT_PKT_ADM_REALSIZE_B(MTU);
			last_packet_size = var_size_B%FTT_PKT_ADM_REALSIZE_B(MTU);
			return m_floor( MTU*packets_MTU + FTT_PKT_ADM_REALSIZE_B_reverse(last_packet_size) );
	}
}

double M_FTT_CORE_L_get_SM_slot_load( void ) /* bps */
{
	double bits_per_usec;

	bits_per_usec = ( (double)(EC_MAX_us_SP( g_master_ec_time )) / ((double)g_master_ec_time) ) * 1/BIT_TIME_US;
	return bits_per_usec * 1e6;
}

double M_FTT_CORE_L_get_AM_slot_load( void ) /* bps */
{
	double bits_per_usec;

	bits_per_usec = ( (double)(EC_MAX_us_AP( g_master_ec_time )) / ((double)g_master_ec_time) ) * 1/BIT_TIME_US;
	return bits_per_usec * 1e6;
}

RET_SCHED_POLICIES M_FTT_CORE_L_ret_sm_sched_policy(void)
{
	switch (Ready_queues_Get_policy(SRT_QUEUE)){
		case SCHED_EDF: 
			return RET_SCHEDD_EDF;
		case SCHED_RM: 
			return RET_SCHEDD_RM;
		case SCHED_DM: 
			return RET_SCHEDD_DM;
		case SCHEDD_FIFO: 
			return RET_SCHEDD_FIFO;
		default:
			ERROR_MSG("Bad policy returned by Ready_queues_Get_policy");
			return -1;
	}
	return -1;
}

RET_SCHED_POLICIES M_FTT_CORE_L_ret_am_sched_policy(void)
{
	switch (Ready_queues_Get_policy(ART_H_QUEUE)){
		case SCHED_EDF: 
			return RET_SCHEDD_EDF;
		case SCHED_RM: 
			return RET_SCHEDD_RM;
		case SCHED_DM: 
			return RET_SCHEDD_DM;
		case SCHEDD_FIFO: 
			return RET_SCHEDD_FIFO;
		default:
			ERROR_MSG("Bad policy returned by Ready_queues_Get_policy");
			return -1;
	}
	return -1;
}

/////DEBUG
void M_FTT_CORE_L_print_db(void)
{
//	PRINT_MSG("Runtime table");
	MDB_runtime_print();
//	PRINT_MSG("Backtime table");
//	MDB_backtime_print();
}


