/*****************************************************************************
 * S_Dispatch.c:
 *****************************************************************************
 * Copyright (C) 2006-2012 the FTT-SE team.
 *
 * Author: Ricardo Marau <marau at fe.up.pt>
 *
 * FTT-SE is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  FTT-SE is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with FTT-SE.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

#include "Ports.h"

#include "ftt-global.h"
#include "ethernet.layer.h"
#include "S_ECscheduled.h"
#include "S_db.h"
#include "S_nodeID.h"
#include "S_igmp_queue.h"
#include "S_appPnP.h"
#include "S_sig_backup.h"


//#define DISABLE_DISPATCHING
//#define DISPATCHING_DECODE
//#define BE_POLITE
//#define DISABLE_AM_SM_TX
//#define DISABLE_ASM_TX
#define DISABLE_TURNAROUND_WAIT
#define DISABLE_SIGNALING_SYNCH_DELAY /*removed because of some signaling messages were being lost... in L26*/		



/*!
\brief Process ID of scheduler task
*/
static rtl_pthread_t ftt_sdispatcher_th;
static FTT_SEMAPHORE sdisp_wake_semaph;

static unsigned int S_ec_period=0; //us

void *ftt_sdispatcher(void *t);

signed char S_Dispatch_Init( void )
{
	rtl_pthread_attr_t attr;
	struct rtl_sched_param sched_param;

	int ret;


	/* Init semaphore to wake the sDispatcher */
	if ( FTT_SEMAPHORE_init(&sdisp_wake_semaph, 0) < 0){	//Initialization locked
		ERROR_MSG("Couldn't initialize semaphore");
		return -1;
	}
	
	/* Create ftt_sDispatcher task */
	rtl_pthread_attr_init( &attr );
	sched_param.sched_priority = rtl_sched_get_priority_max(RTL_SCHED_FIFO)-1;
	rtl_pthread_attr_setschedparam( &attr, &sched_param );
#ifdef RTL_PRO
	rtl_pthread_attr_setfp_np(&attr, 1);
#endif
//	rtl_pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	if ( (ret = rtl_pthread_create( &ftt_sdispatcher_th, &attr, ftt_sdispatcher, (void *)0 ))<0 ){
		ERROR_MSG("Init Fatal Error: pthread_create returned %d", ret);
		return -2;
	}
#ifdef PRINT_THREAD_ID
#ifdef RTL_PRO
	ERROR_MSG("New thread id: %d", ftt_sdispatcher_th);
#else
	ERROR_MSG("New thread id: %lu", ftt_sdispatcher_th);
#endif
#endif

#ifdef PRINT_MODULES_INIT
	PRINT_MSG(" -S_Dispatch_Init Ok");
#endif
	return 0;
}

void S_Dispatch_Close(void)
{

//	PRINT_MSG("Closing S_Dispatch");
	/* cancel the ftt_dispatcher thread */
	rtl_pthread_cancel( ftt_sdispatcher_th );
	rtl_pthread_join( ftt_sdispatcher_th, NULL );

	/* Clean up the semaphore */
	FTT_SEMAPHORE_destroy(&sdisp_wake_semaph);
	
#ifdef PRINT_MODULES_INIT
	PRINT_MSG(" -S_Dispatch_Close Ok");
#endif
	
}

void S_Dispatch_Wake(void)
{
	/* Activate ftt_sdispatcher() */
	FTT_SEMAPHORE_post(&sdisp_wake_semaph);
}

void S_Dispatch_Set_ec_time( int ec_time_ms )
{
	S_ec_period = ec_time_ms*1000;
	return;
}

unsigned int S_Dispatch_Read_ec_time_us( void )
{
	return S_ec_period;
}

static unsigned char S_Dispatcher_running_flag;

unsigned char S_Dispatcher_is_running(void)
{
	return (S_Dispatcher_running_flag != 0);
}

extern void S_NRDB_enter(void);
extern void S_NRDB_leave(void);


void *ftt_sdispatcher(void *t)
/*******************************************************************************/
/*!
\brief This task is responsible by timely produce the station messages.

	This task is responsible by timely produce the station messages.
    When an Ethernet frame with a Trigger message arrives, the routine\n
	that handles the Ethernet frame processing fills a data structure\n
	with the IDs and release time of the messages to be produced by this station.\n
	When the Trig mesg is completely processed, it activates this task.

*****************************************************************************/
{
	int ret;

	unsigned short MesgID;		/* Message ID */
	unsigned short Mesg_frag_no;		/* Message Length (in slots) */
	unsigned char  Mesg_source_nodeID;
	signed char mesgstatus;		/* Status ofa given message */

	unsigned char *tmsg;
	FTT_PKT_GENERIC_HEADERS *ftt_tmsg_generic_pointer;
	FTT_PKT_SDM_HEADERS *ftt_tmsg_sdm_pointer;
	FTT_PKT_ADM_HEADERS *ftt_tmsg_adm_pointer;

	unsigned short mesg_efflen;    /* Real data length of message (in bytes) */
	unsigned char mesg_flags=0;
	unsigned char mesg_seqn;      	/* Message sequence number */
	unsigned short frag_seqn;	/* Message fragmentation seqno*/

	FTT_SIGNAL *callback_sig;

	unsigned char *mac=NULL;

#ifdef QNX_TAGGING
	QNX_TAGGING_TYPE tagging;
#endif

	while (1){
		S_Dispatcher_running_flag = 0;
		/* Wait for start */
		if ( FTT_SEMAPHORE_wait(&sdisp_wake_semaph) < 0 )
			return NULL;

		S_Dispatcher_running_flag = 1;
#ifdef DISABLE_DISPATCHING
		continue;
#endif
		
#if defined(BE_POLITE)||defined(DISPATCHING_DECODE)
		PRINT_MSG("S_Dispatcher: ");
#endif

		/* Lets go through the table */
		while ( S_ECScheduled_GetMessage(&MesgID, &Mesg_source_nodeID, &Mesg_frag_no) ){

#ifdef DISPATCHING_DECODE
		PRINT_MSG("  Var:%x frag%d src:%d", MesgID, Mesg_frag_no, Mesg_source_nodeID );
#endif

			if ( (Mesg_source_nodeID != S_nodeID_read()) ||
			     ((mesgstatus = NDB_test_produce_and_bind(MesgID, &mac)) < 0) ){
				/* Station not producer of this message*/
#ifdef DISPATCHING_DECODE
				PRINTF(" pass");
#endif
				continue ; /* and nothing else matters...Metallica*/
			}
#ifdef DISPATCHING_DECODE
			PRINTF(" tx_ing");
#endif

			/* Get one new tx buffer */
			if ( ETH_L_tx_reserve_buffer( &tmsg ) < 0 ){
				ERROR_MSG("Problems getting a new Tx buffer - Progamming error ");
			}

			/* 1 - Prepare the message */
			ftt_tmsg_generic_pointer = (FTT_PKT_GENERIC_HEADERS *)tmsg;
					
			/* The source and destination address */
			ETH_L_CopyMAC_to_from( ftt_tmsg_generic_pointer->eth_header.eth_dest, mac );
			ETH_L_CopyMAC_to_from( ftt_tmsg_generic_pointer->eth_header.eth_src, S_Get_My_mac() );

			/* Message type */
			ftt_tmsg_generic_pointer->eth_header.eth_type = htons ( ETH_FTT_TYPE );

			mesg_efflen = 0;
			callback_sig = NULL;

			/* Select message type */
			if( FTT_VAR_ID_IS_ID_SYNC(MesgID) ){
				//PRINT_MSG("SDidp: SDATA_MESG_ID");
				ftt_tmsg_sdm_pointer = (FTT_PKT_SDM_HEADERS *)(tmsg);

				/* Get data into net buffer */
				S_NRDB_enter();
				mesgstatus = NDB_get_data2tx(
						MesgID,
						Mesg_frag_no,
						ftt_tmsg_sdm_pointer->data,
						&mesg_efflen,
						&mesg_seqn,
						&frag_seqn,
#ifdef QNX_TAGGING
						&tagging,
#endif
						&callback_sig );
				S_NRDB_leave();
				
				//PRINT_MSG(" NDB_get_data2tx ret: %d", mesgstatus);
//if (frag_seqn == 1) PRINTF(".");
//PRINTF("S");
				mesg_efflen = FTT_PKT_SDM_LOGICALSIZE_B( mesg_efflen );

				if(mesgstatus < 0) {
					ERROR_MSG("Strange Error: ftt_get2tx_sdata failed with 0x%x frag:%d ", MesgID, Mesg_frag_no );
					ETH_L_tx_un_reserve_buffer( );
					continue;
				}
					
				mesg_flags = 0;
				if(mesgstatus)   /* Update the TXNEWDAT bit on the message identifier */
					mesg_flags |= TXNEWDAT_MASK;

				/* Setup message ID */
				ftt_tmsg_sdm_pointer->pkt_header.type = htons( FTT_SDATA_MSG );
				ftt_tmsg_sdm_pointer->sdm_header.id = htons((unsigned short)MesgID);
				ftt_tmsg_sdm_pointer->sdm_header.seq_no = mesg_seqn;
				ftt_tmsg_sdm_pointer->sdm_header.flag_s = mesg_flags;
				ftt_tmsg_sdm_pointer->sdm_header.frag_seqno = htons(frag_seqn);
				ftt_tmsg_sdm_pointer->sdm_header.source_nodeID = S_nodeID_read();
#ifdef QNX_TAGGING
				ftt_tmsg_sdm_pointer->sdm_header.tagging  = htoni(tagging);
#endif
				
			}
			else {
				//PRINT_MSG("SDidp: ADATA_MESG_ID");
				ftt_tmsg_adm_pointer = (FTT_PKT_ADM_HEADERS *)(tmsg);
				
				/* Get data into net buffer */
				S_NRDB_enter();
				mesgstatus=NDB_get_data2tx(
						MesgID,
						Mesg_frag_no,
						ftt_tmsg_adm_pointer->data,
						&mesg_efflen,
						&mesg_seqn,
						&frag_seqn,
#ifdef QNX_TAGGING
						&tagging,
#endif
						&callback_sig );
				S_NRDB_leave();
				//PRINT_MSG(" ftt_get2tx_adata ret: %d", mesgstatus);

				/* If no message was in the FIFO */
				if (mesg_efflen == 0)
					ftt_tmsg_adm_pointer->pkt_header.type = htons( FTT_IDLE );
				else
					ftt_tmsg_adm_pointer->pkt_header.type = htons( FTT_ADATA_MSG );

				mesg_efflen = FTT_PKT_ADM_LOGICALSIZE_B( mesg_efflen );

				if(mesgstatus < 0) {
					ERROR_MSG("Strange Error: ftt_get2tx_adata failed after ftt_test_producer");
					ETH_L_tx_un_reserve_buffer( );
					continue;
				}

				mesg_flags = 0;
				if(mesgstatus)   /* Update the TXNEWDAT bit on the message identifier */
					mesg_flags |= TXNEWDAT_MASK;

//				PRINT_MSG("sdispatcher: Async mesg_eff_len = %d", mesg_efflen);

				/* Setup message ID */
				ftt_tmsg_adm_pointer->adm_header.id  = htons((unsigned short)MesgID);
				ftt_tmsg_adm_pointer->adm_header.seq_no	= mesg_seqn;
				ftt_tmsg_adm_pointer->adm_header.flag_s  = mesg_flags;
				ftt_tmsg_adm_pointer->adm_header.frag_seqno  = htons(frag_seqn);
				ftt_tmsg_adm_pointer->adm_header.source_nodeID = S_nodeID_read();
#ifdef QNX_TAGGING
				ftt_tmsg_adm_pointer->adm_header.tagging  = htoni(tagging);
#endif	
			}

#ifndef DISABLE_TURNAROUND_WAIT
			/* 2 - Waits for time-to-txmit */
			/* 3 - Busy waits the arrival of correct instant */ 
			{
				struct timespec tm_rx_abs_time;
				
				/* Get the TM absolute time reference */
				tm_rx_abs_time = S_ECScheduled_ReadTimestamp();

				timespec_add_ns( &tm_rx_abs_time, SCAN_WINDOW * 1000);
				jitter = timespec_from_ns(0);
				rtl_clock_nanosleep( CLOCK_REALTIME, TIMER_ABSTIME|RTL_TIMER_ADVANCE, &tm_rx_abs_time, &jitter);
			}
#endif
			
#ifndef DISABLE_AM_SM_TX
			/* 4 - Send packet */
//			PRINT_MSG ("Dispatching a packet with size: %d", mesg_efflen );
			if ( (ret = ETH_L_tx_send_messagebuffer(NULL, mesg_efflen)) < 0 ){
				ERROR_MSG("Message not sent. ETH_L_tx_send_messagebuffer returned: %d", ret);
				/* Maybe I should un-reserv the buffer tmsg */
			}
#endif
			/* 5 - Activate callback (if defined) */
			if(callback_sig!=NULL) {
				FTT_signal( callback_sig );
			}

		} // while get message

		S_Dispatcher_running_flag = 0;

#ifdef FTT_SLAVE_COMPILE
	{
		ASM_MSG_QUEUE_INDEX	*temp_p;
		unsigned short counter;
		signed short nodeID;
		FTT_PKT_ASM_HEADERS *ftt_tmsg_asm_pointer;
		unsigned char app_id;
		static unsigned char signaling_seq_no=0;

		NDB_AM_Get_sent_RETURN_TYPE ndb_return_AM_queues[STA_MAX_AM_VARS];
		unsigned short ndb_return_buf_size = STA_MAX_AM_VARS;
	
	
		if ( ETH_L_tx_reserve_buffer( &tmsg ) < 0 ){
			ERROR_MSG("Problems getting a new Tx buffer - Progamming error ");
		}

		ftt_tmsg_asm_pointer = (FTT_PKT_ASM_HEADERS *)tmsg;
		
		switch (nodeID=S_nodeID_read()){
			case -1:
				ERROR_MSG(" master_mac still not initialized");
				break;
			case -2:
				/* Send the Boot request */
//				PRINT_MSG("Preparing to send a PnP req");
				ETH_L_CopyMAC_to_from( ftt_tmsg_asm_pointer->eth_header.eth_dest, S_Get_Master_mac() );
				ETH_L_CopyMAC_to_from( ftt_tmsg_asm_pointer->eth_header.eth_src, S_Get_My_mac() );

				ftt_tmsg_asm_pointer->eth_header.eth_type = htons ( ETH_FTT_TYPE );
				ftt_tmsg_asm_pointer->pkt_header.type = htons( FTT_PLUGnPLAY );
		
				//Missing the Seq_no
				ftt_tmsg_asm_pointer->asm_header.seq_no = signaling_seq_no;
				ftt_tmsg_asm_pointer->asm_header.nodeID = (unsigned char)nodeID;
				ftt_tmsg_asm_pointer->asm_header.nam = htons((unsigned short)0);
				mesg_efflen = FTT_PKT_ASM_LOGICALSIZE_B_from_no(0);
				break;
			default:
//				PRINT_MSG("Preparing to send an ASM");
				/* Send the Asynch buffers status */
				ETH_L_CopyMAC_to_from( ftt_tmsg_asm_pointer->eth_header.eth_dest, S_Get_Master_mac() );
				ETH_L_CopyMAC_to_from( ftt_tmsg_asm_pointer->eth_header.eth_src, S_Get_My_mac() );

				ftt_tmsg_asm_pointer->eth_header.eth_type = htons ( ETH_FTT_TYPE );
				ftt_tmsg_asm_pointer->pkt_header.type = htons( FTT_ASTATUS_MSG );

				temp_p = (ASM_MSG_QUEUE_INDEX *)ftt_tmsg_asm_pointer->data;

				S_NRDB_enter ();
				NDB_AM_Get_sent( ndb_return_AM_queues, &ndb_return_buf_size);
				S_NRDB_leave();

				counter = 0;
				/* check if there is any app request */
				if (S_appPnP_is_there_a_request(&app_id)){
					// We will notify that will as if it was an assync notification
					// The AMesgId field will hold the MASTER_BROADCAST_A_CH_ID 
					//   - (not used in this context - so used to notify this request)
					// The AMesgQueueLen holds the app_id
					temp_p[counter].AMesgId = htons(MASTER_BROADCAST_A_CH_ID);
					temp_p[counter].AMesgQueueLen = htons((unsigned short)app_id);
					counter++;
				}

				/* now, the real AM status */
				while ( ndb_return_buf_size-- ){
//					PRINT_MSG("Sending status for id: %x   num:%d",ndb_return_AM_queues[ndb_return_buf_size].tagged_id, ndb_return_AM_queues[ndb_return_buf_size].num );
					temp_p[counter].AMesgId = 
						htons(ndb_return_AM_queues[ndb_return_buf_size].tagged_id);
					temp_p[counter].AMesgQueueLen = 
						htons(ndb_return_AM_queues[ndb_return_buf_size].num);
					counter++;
				}

#ifdef SIG_RECOVER
				{
				unsigned char seq_no_retransmit;
				void *ret_mem_p;
				unsigned short ret_mem_size;
				
				while ( S_ECScheduled_GetSigRequest( NULL, &seq_no_retransmit) ){
					S_sig_backup_pull( seq_no_retransmit, &ret_mem_p, &ret_mem_size );

					if ( ret_mem_size%sizeof(ASM_MSG_QUEUE_INDEX) ) {
						ERROR_MSG("Programming error! Size must always come as a multiple of sizeof(ASM_MSG_QUEUE_INDEX) ");
					}
					PRINT_MSG("Recovering signaling %d. Caught %d IDXs", seq_no_retransmit, ret_mem_size/sizeof(ASM_MSG_QUEUE_INDEX) );
					
					if (ret_mem_p) {
						memcpy( (void *)(temp_p+counter), ret_mem_p, ret_mem_size );
						counter += ret_mem_size/sizeof(ASM_MSG_QUEUE_INDEX);
					}
					
				}

				S_sig_backup_push( signaling_seq_no, temp_p, counter*sizeof(ASM_MSG_QUEUE_INDEX) );
				}
#endif
			
				ftt_tmsg_asm_pointer->asm_header.seq_no = signaling_seq_no;
				ftt_tmsg_asm_pointer->asm_header.nodeID = (unsigned char)nodeID;
				ftt_tmsg_asm_pointer->asm_header.nam = htons((unsigned short)counter);
				mesg_efflen = FTT_PKT_ASM_LOGICALSIZE_B_from_no(counter);

				break;
		}
		signaling_seq_no ++;

#ifndef DISABLE_SIGNALING_SYNCH_DELAY
#ifndef ICNOVA
		{		
			struct timespec tm_rx_abs_time;

			/* Get the TM absolute time reference */
			tm_rx_abs_time = S_ECScheduled_ReadTimestamp();
		
			//timespec_add_ns( &tm_rx_abs_time, ( S_ec_period-20 ) * 1000); //como estava antes
			timespec_add_ns( &tm_rx_abs_time, ( S_ec_period-(S_ec_period*0.15) ) * 1000);
		
			jitter = timespec_from_ns(0);
			//PRINTF(">Waiting");
			rtl_clock_nanosleep( CLOCK_REALTIME, TIMER_ABSTIME|RTL_TIMER_ADVANCE, &tm_rx_abs_time, NULL /*jitter*/);
		}
#endif
#endif
		
#ifndef DISABLE_ASM_TX
		if ( (ret = ETH_L_tx_send_messagebuffer(NULL, mesg_efflen)) < 0 ){
			ERROR_MSG("Message not sent. ETH_L_tx_send_messagebuffer returned: %d", ret);
			/* Perhaps I should un-reserv the buffer tmsg */
		}
//		PRINTF(">Sent");
#endif
	}
#endif

		/* Flush the igmp_queue */
		S_igmp_queue_flush_queue();

	}//while
		
	return NULL;
}
