/*****************************************************************************
 * S_NRDB.h:
 *****************************************************************************
 * Copyright (C) 2006-2012 the FTT-SE team.
 *
 * Author: Ricardo Marau <marau at fe.up.pt>
 *
 * FTT-SE is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  FTT-SE is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with FTT-SE.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

/*! \page NRDB NRDB - Node Requirements DataBase
 This block is responsible for dealing with the system registration requests coming from the Master. The "Master_ch" receives those change requests and submits those in this block, which then updates the local "db".
 The local db also supports the update tagging, as described in \ref SRDB.

 Associated to each entry in the "db" there is a pointer to the "allocator" module. This module is the memory manager for the buffers that hold data content from the variable. Upon creating from the NRDB, the pointer is "null", being the application responsible for issuing a Bind request, which will provide a memory block to be handled in the "allocator". This memory block may be provided by the application (the application must provide the enough memory size for the variable) or allocated using "malloc". This option comes from a limitation in RTLinuxPro that limits the amount of memory being dynamically allocated. Hence, in this case, memory is served statically by the application. 
 */


#ifndef S_NRDB_H
#define S_NRDB_H

#include "ftt-core.layer_types.h"

/** Initializes the NRDB. \return 0 - OK */
extern signed char S_NRDB_Init(void);

/** Closes the NRDB. \return 0 - OK */
extern signed char S_NRDB_Close(void);

/** PreClosing call - must be issued right before close */
extern void S_NRDB_Pre_Close(void);

/** Add a message. This is issued from the Master_ch. \return 0 - OK */
extern signed char S_NRDB_add_msg( unsigned char m_type,
		unsigned short m_id,
		unsigned char m_tag,
		unsigned int m_size,
		unsigned int m_max_size,
		unsigned int m_period,
		unsigned short m_MTU,
		unsigned char *m_prod_mac,
		unsigned char m_prod_max_no,
		unsigned char overwrite);

/** Del a message - as opposed to #S_NRDB_add_msg . \return 0 - OK */
extern signed char S_NRDB_del_msg( unsigned char m_type, unsigned int m_id);

/** Cleans all instantiations of a tag. When updating some variable requirements both tags are registered in the transition phase.
 This call cleans an old tag, as a consequence of a Master command that determines the end of the lifecycle for that tag. \return 0 - OK */
extern signed char S_NRDB_clean_msg_tag( unsigned char m_type, unsigned int m_id, unsigned char m_tag);

/** Binds a variable to the application. Memory may be provided by the application of (if null) allocated locally in the "allocator"
 After bound, a callback signal is returned to be used by the application as an event listener for every RX or TX that occur in the variable.
 \return 0 - OK */
extern signed char S_NRDB_bind_mesg(
		MESSAGE_TYPE m_type, /**< [in] Message type */
		unsigned int m_id, /**< [in] Message FTT id */
		FTT_SIGNAL **ret_callback_signal, /**< [out] Return pointer to an FTT_SIGNAL */
		void *memory_point, /**< [in] Pointer to a memory block */
		unsigned int mem_size, /**< [in] The size of the given memory block */
		unsigned char slots_no /**< [in] The number of slots that the application wants to register for this variable. For synchronous messages this must be 3, while for Asynchronous messages this is variable */ );

/** Unbinds a variable. \return 0 - OK */
extern signed char S_NRDB_unbind_mesg(
		MESSAGE_TYPE m_type,
		unsigned short m_id);

/** Checks whether a variable is correctly bound.
 \return 0 - not bound */
extern signed char S_NRDB_isbind_mesg(
		MESSAGE_TYPE m_type,
		unsigned short m_id);

/** Blocks waiting for the creation of a variable locally. This call should be prior to the bind process. It is a synchronization mechanism for the thread to wait for the creation of the variable. \return 0 - OK */
extern signed int S_NRDB_wait4_mesg(
		MESSAGE_TYPE m_type,
		unsigned int m_id);

/** Blocks waiting for the removal (deletion) of a variable locally from the "db". \return 0 - OK */
extern signed int S_NRDB_wait4_not_mesg(
		MESSAGE_TYPE m_type,
		unsigned int m_id);

/** Requests some information related to the variable, such as the current tag, size and period. This information can vary asynchronously, issued from the Master and the applications should use this call to catch up and produce information according to the new requirements \return 0 - OK */
extern signed char S_NRDB_get_current(
		MESSAGE_TYPE m_type,
		unsigned int m_id,
		unsigned char *tag,
		unsigned int *cur_size,
		unsigned short *period );

/** Requests the memory size that is necessary to provide when binding a variable. \return 0 - OK */
extern signed char S_NRDB_get_memory_pool_size(
		MESSAGE_TYPE m_type,
		unsigned int m_id,
		unsigned char slots_no,
		unsigned short *ret_mem_size );

extern signed char S_NRDB_printmesg( void );

#endif
