/*****************************************************************************
 * S_allocator.c:
 *****************************************************************************
 * Copyright (C) 2006-2012 the FTT-SE team.
 *
 * Author: Ricardo Marau <marau at fe.up.pt>
 *
 * FTT-SE is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  FTT-SE is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with FTT-SE.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

#include "Ports.h"

#include "S_allocator.h"

//#define DEBUG_P



#define MAX_OBJS	100

#ifdef RTL_PRO
#define RTL_O_VIRT 0x00800000
#define CABS_POOL 1	//this is the used pool of the rtl appended module 
inline unsigned char *vmalloc( unsigned long int size )
{
	if ( size > 30000 ){
		ERROR_MSG("A request of this size can't be issued to the FTT internal mem pool. Mem must be provided from the application");
		return NULL;
	}
	return rtl_rt_malloc_pool( CABS_POOL, size );
}

inline void vfree( unsigned char *p )
{
	return rtl_rt_free( p );
}

#else
inline unsigned char *vmalloc( unsigned long int size )
{
	return malloc( size );
}

inline void vfree( unsigned char *p )
{
	return free( p );
}
#endif


typedef struct data_slot{
	unsigned char *mem_p;
	struct data_slot *next;
	unsigned short n_uso; // used only in CABS. It tracks the number of processes locked in a given requested CAB slot
#ifdef QNX_TAGGING
	unsigned int tagging;
#endif
}DATA_SLOT;

typedef enum{
	NOT_IN_USE=0,
	USE_AS_CAB,
	USE_AS_FIFO
}OBJECT_TYPE;

typedef struct{
	DATA_SLOT *free_p;
	DATA_SLOT *mrd;	//most recent data

	unsigned char memory_type; //0-internal FTT memory
	                           //1-external FTT memory
	unsigned char *g_mem;      //global mem - pointer to the memory pool of this object
	FTT_SEMAPHORE free_items_sem;  //Semaphore to track and block for available items

	rtl_pthread_mutex_t access_locker;  //locker
	unsigned int mem_slot_size;
	unsigned short mem_slot_num;
	OBJECT_TYPE in_use;

	ALLOC_ID next_free_point;

}OBJECT_POINT;

static OBJECT_POINT object_points[MAX_OBJS];

static ALLOC_ID	next_free_point;

static unsigned char Allocator_init_flag=0;


signed char Allocator_init(void)
{
	unsigned short i;

	next_free_point = 0;
	for (i=0; i<MAX_OBJS-1; i++ ){
		object_points[i].next_free_point = i+1;
		object_points[i].in_use = NOT_IN_USE;
	}
	object_points[i].next_free_point = -1;
	object_points[i].in_use = NOT_IN_USE;
#if (defined (RTL_PRO) && defined (USE_FTT_INTERNAL_POOL))
	if ( rtl_init_mem_pool(CABS_POOL, RTL_O_VIRT, 30000, 20)<0)
		return -1;
#endif
	Allocator_init_flag = 1;
#ifdef PRINT_MODULES_INIT
	PRINT_MSG(" -Allocator_init Ok");
#endif	
	return 0;
}

signed char Allocator_close(void)
{
#if (defined (RTL_PRO) && defined (USE_FTT_INTERNAL_POOL))
	rtl_cleanup_mem_pool(CABS_POOL);
#endif
#ifdef PRINT_MODULES_INIT
	PRINT_MSG(" -Allocator_close Ok");
#endif		
	return 0;
}
signed char AllocatorInit_check(void)
{
	if (Allocator_init_flag == 0)
		return -1;
	return 0;
}

static signed char checkcab(ALLOC_ID id)
{
    if ((id >= MAX_OBJS)||(id < 0))
		return -1;

    if (object_points[id].in_use == USE_AS_CAB)	//Is a cab?
		return 0;
    return -1;
}

unsigned int Allocator_get_memory_size_needed(unsigned int dim_mes, unsigned char num_mes)
{
	return ((dim_mes + sizeof( DATA_SLOT )) * num_mes);
}

ALLOC_ID cab_create(unsigned int dim_mes, unsigned char num_mes, void *memory_point, unsigned int mem_size )
{
	ALLOC_ID id;
	OBJECT_POINT *pid;
	DATA_SLOT *tmp=NULL;
	unsigned char *mem;
	int i;

	if ( (id=next_free_point) != MAX_OBJS ){
		pid = &(object_points[id]);
		next_free_point = pid->next_free_point;
	}else{
		return -1;
	}

	if (memory_point != NULL){
		mem = (unsigned char *)memory_point;
		pid->memory_type = 1;//external memory point
		if (mem_size < Allocator_get_memory_size_needed( dim_mes, num_mes) ){
			ERROR_MSG("Not enough memory: Passed:%d but needed:%d", mem_size, Allocator_get_memory_size_needed( dim_mes, num_mes));
			return -3;
		}
	}else
	{
#ifdef USE_FTT_INTERNAL_POOL
		mem = vmalloc( Allocator_get_memory_size_needed( dim_mes, num_mes) );
		pid->memory_type = 0;//internal
#else
		return -2;
#endif
//		PRINT_MSG(" = a NULL");
	}

	/* descriptor */
	pid->g_mem = mem;
	pid->mem_slot_size = dim_mes;
	pid->mem_slot_num = num_mes;

	/* memory init */
	
	/* The mrd points to one and only one item, I am starting it with
	 * the first memory slot, but it could be any other
	 * I am not terminating the mrd->next=NULL 'cause it's only one in this list */
	pid->mrd = (DATA_SLOT *)mem;
	i=(int)num_mes;

	/* all memory slots shall be linked */
	while(i--){
		tmp = (DATA_SLOT *)mem;
		mem += (dim_mes + sizeof( DATA_SLOT ));
		tmp->next = (DATA_SLOT *)mem;
		tmp->mem_p = (unsigned char *)(tmp+1);
		tmp->n_uso = 0;
	}
	tmp->next = NULL;

	/* The free points to the second item listed 
	 * The first onr is pointed by mrd */
	pid->free_p = pid->mrd->next;

//	/* Initiate the free items semaphore - queue size */
//	FTT_SEMAPHORE_init(&(pid->free_items_sem), num_mes); /* It's the same as if I initiated with 0-locked 
//							     and the posted (num_mes) times */
	
	rtl_pthread_mutex_init(&(pid->access_locker), 0);

	pid->in_use = USE_AS_CAB;

	return (id);
}

void cab_delete(ALLOC_ID id)
{
	OBJECT_POINT *pid;

	if ( checkcab(id) < 0 ) return;

	pid = &(object_points[id]);
#ifdef USE_FTT_INTERNAL_POOL
	if 	(pid->memory_type == 0) //is internal
		vfree (pid->g_mem);
#endif
	pid->in_use = NOT_IN_USE;
	pid->next_free_point = next_free_point;
	next_free_point = id;

//	/* Deallocate the semaphore */
//	FTT_SEMAPHORE_destroy(&(pid->free_items_sem));

	rtl_pthread_mutex_destroy(&(pid->access_locker));


	return;
}

unsigned char *cab_reserve(ALLOC_ID id)
{
	OBJECT_POINT *pid;
	unsigned char *buf;
	
	if ( checkcab(id) < 0 ) return NULL;
#ifdef DEBUG_P
	PRINT_MSG("CAB reserv");
#endif
	pid = &(object_points[id]);
	
	rtl_pthread_mutex_lock(&(pid->access_locker));

	/* return if empty */
	if (pid->free_p == NULL){
		PRINTF("ret null");

		rtl_pthread_mutex_unlock(&(pid->access_locker));

		return NULL;
	}

	buf = pid->free_p->mem_p;
	pid->free_p = pid->free_p->next;
#ifdef DEBUG_P
	PRINTF("ret %p", buf);
#endif

		rtl_pthread_mutex_unlock(&(pid->access_locker));

	return (buf);
}

signed char cab_reserve_abort(ALLOC_ID id, void *pun_mes)
{
	DATA_SLOT *pbuf;
	OBJECT_POINT *pid;
#ifdef DEBUG_P
PRINT_MSG("CAB unreserv %p", pun_mes);
#endif
	if ( checkcab(id) < 0 ) return -1;
	pid = &(object_points[id]);


	rtl_pthread_mutex_lock(&(pid->access_locker));


	pbuf = ((DATA_SLOT *)(pun_mes)) - 1;

	/* Let's put the slot back to where it came from, the free list */
	pbuf->next = pid->free_p;
	pid->free_p = pbuf;

	/* SEM */

	rtl_pthread_mutex_unlock(&(pid->access_locker));

	return 0;
}

#ifdef QNX_TAGGING
signed char cab_putmes(ALLOC_ID id, void *pbuf, unsigned int qnx_tag)
#else
signed char cab_putmes(ALLOC_ID id, void *pbuf)
#endif
{
	OBJECT_POINT *pid;
	
	if ( checkcab(id) < 0 ) return -1;
	
	pid = &(object_points[id]);
#ifdef DEBUG_P
	PRINT_MSG("CAB putmes %p", pbuf);
#endif

	rtl_pthread_mutex_lock(&(pid->access_locker));

	/* Well, now we have a fresh buffer that shall replace the old slot in mrd
	 * If the slot was reserved for reading it will momentaneously be unaccessable
	 * from either the free or mrd lists. The application shall returne it and then 
	 * it will be placed on the free. */

	if (pid->mrd->n_uso == 0) {
		/* No process was holding it, so
		 * back to free list */
		pid->mrd->next = pid->free_p;
		pid->free_p = pid->mrd;

		/* SEM */
	}
	
	/* Put the fresh slot on mrd */
	pid->mrd = ((DATA_SLOT *)pbuf) -1;
#ifdef QNX_TAGGING
	pid->mrd->tagging= qnx_tag;
#endif

	rtl_pthread_mutex_unlock(&(pid->access_locker));
	
	return 0;
}

unsigned char *cab_getmes(ALLOC_ID id)
{
	OBJECT_POINT *pid;
	unsigned char *ret;
	
	if ( checkcab(id) < 0 ) return NULL;
	pid = &(object_points[id]);

	rtl_pthread_mutex_lock(&(pid->access_locker));

	(pid->mrd->n_uso)++;
	ret = (unsigned char *)(pid->mrd->mem_p);


	rtl_pthread_mutex_unlock(&(pid->access_locker));



#ifdef DEBUG_P
	PRINT_MSG("CAB getmes %p", ret);
#endif

	return ( ret );
}

#ifdef QNX_TAGGING
signed char cab_unget(ALLOC_ID id, void *pun_mes, unsigned int *ret_qnx_tag)
#else
signed char cab_unget(ALLOC_ID id, void *pun_mes)
#endif
{
	DATA_SLOT *pbuf;
	OBJECT_POINT *pid;
#ifdef DEBUG_P
	PRINT_MSG("CAB unget %p", pun_mes);
#endif

	if ( checkcab(id) < 0 ) return -1;
	pid = &(object_points[id]);

	rtl_pthread_mutex_lock(&(pid->access_locker));


	pbuf = ((DATA_SLOT *)(pun_mes)) - 1;

#ifdef QNX_TAGGING
	if (ret_qnx_tag != NULL) *ret_qnx_tag = pbuf->tagging;
#endif
	
	/* If this buffer is the freshest slot (mrd) it shall remain in mrd
	 * If any other process is holding the slot I can't free it */
	if ((--(pbuf->n_uso) == 0) && (pbuf != pid->mrd)) {
		pbuf->next = pid->free_p;
		pid->free_p = pbuf;

    	}

	rtl_pthread_mutex_unlock(&(pid->access_locker));

    return 0;
}

void cab_print_all(void)
{
	int i;
	DATA_SLOT *temp_p;
	for (i=0; i<MAX_OBJS; i++){
		if (object_points[i].in_use == USE_AS_CAB){//if a CAB
			PRINTF("\n[%2d] %d*%d  next_f_p:%d  g_mem:%p mrd: (%p,%d) free_p: ",
					i,
					object_points[i].mem_slot_num,
					object_points[i].mem_slot_size,
					object_points[i].next_free_point,
					object_points[i].g_mem,
					object_points[i].mrd,
					object_points[i].mrd->n_uso
					);
			temp_p = object_points[i].free_p;
			while (temp_p != NULL){
				PRINTF("(%p,%d) -> ",temp_p->mem_p, temp_p->n_uso);
				temp_p = temp_p->next;
			}
			PRINTF(" NULL");

		}	
	}
	PRINTF("\n");
}

static signed char checkfifo(ALLOC_ID id)
{
    if ((id >= MAX_OBJS)||(id < 0))
		return -1;

    if (object_points[id].in_use == USE_AS_FIFO)	//Is a fifo?
		return 0;
    return -1;
}

ALLOC_ID fifo_create(unsigned int dim_mes, unsigned char num_mes, void *memory_point, unsigned int mem_size )
{
	ALLOC_ID id;
	OBJECT_POINT *pid;
	DATA_SLOT *tmp=NULL;
	unsigned char *mem;
	int i;

	if ( (id=next_free_point) != MAX_OBJS ){
		pid = &(object_points[id]);
		next_free_point = pid->next_free_point;
	}else{
		return -1;
	}

	if (memory_point != NULL){
		mem = (unsigned char *)memory_point;
		pid->memory_type = 1;//external memory point
		if (mem_size < Allocator_get_memory_size_needed( dim_mes, num_mes) ){
			ERROR_MSG("Not enough memory: Passed:%d but needed:%d", mem_size, Allocator_get_memory_size_needed( dim_mes, num_mes));
			return -3;
		}
	}else
	{
#ifdef USE_FTT_INTERNAL_POOL
		mem = vmalloc( Allocator_get_memory_size_needed( dim_mes, num_mes) );
		pid->memory_type = 0;//internal
#else
		return -2;
#endif
//		PRINT_MSG(" = a NULL");
	}

	/* descriptor */
	pid->g_mem = mem;
	pid->mem_slot_size = dim_mes;
	pid->mem_slot_num = num_mes;

	/* memory init */
	i=(int)num_mes;
	
	while(i--){
		tmp = (DATA_SLOT *)mem;
		mem += (dim_mes + sizeof( DATA_SLOT ));
		tmp->next = (DATA_SLOT *)mem;
		tmp->mem_p = (unsigned char *)(tmp+1);
		tmp->n_uso = 0;
	}
	tmp->next = NULL;

	pid->mrd = NULL; //(DATA_SLOT *)mem;
	pid->free_p = (DATA_SLOT *)pid->g_mem; //pid->mrd->next;


	/* Initiate the free items semaphore - queue size */
	/* It's the same as if I initiated with 0-locked and the posted (num_mes) times */
	if ( FTT_SEMAPHORE_init(&(pid->free_items_sem), num_mes) < 0 ){
		ERROR_MSG("Couldn't initialize semaphore");
		return -3;
	}
	

	rtl_pthread_mutex_init(&(pid->access_locker), 0);

	pid->in_use = USE_AS_FIFO;

	return (id);
}

void fifo_delete(ALLOC_ID id)
{
	OBJECT_POINT *pid;

	if ( checkfifo(id) < 0 ) return;

	pid = &(object_points[id]);

	/* Deallocate the semaphore */
	FTT_SEMAPHORE_destroy(&(pid->free_items_sem));
	
#ifdef USE_FTT_INTERNAL_POOL
	if 	(pid->memory_type == 0) //is internal
		vfree (pid->g_mem);
#endif
	pid->in_use = NOT_IN_USE;
	pid->next_free_point = next_free_point;
	next_free_point = id;


	rtl_pthread_mutex_destroy(&(pid->access_locker));


	return;
}

/* Get a slot for possible input */
unsigned char *fifo_reserve(ALLOC_ID id, unsigned char blocking)
{
	OBJECT_POINT *pid;
	unsigned char *buf;

	if ( checkfifo(id) < 0 ) return NULL;

	pid = &(object_points[id]);

	/* SEM */
	if (blocking){
		if (FTT_SEMAPHORE_wait(&(pid->free_items_sem)) < 0) return NULL; //The allocator sem was probably destroyed
	}
	else{
		if ( FTT_SEMAPHORE_trywait(&(pid->free_items_sem)) < 0 ) return NULL; //The allocator sem was probably full
	}

	rtl_pthread_mutex_lock(&(pid->access_locker));
	
	/* Double check - return if fifo is full */
	if (pid->free_p == NULL){
		/* After the semaphore checking this really shouldn't be happening */
		ERROR_MSG("FATAL error - FIFO full");

		rtl_pthread_mutex_unlock(&(pid->access_locker));

		return NULL;
	}

	buf = pid->free_p->mem_p;
	pid->free_p = pid->free_p->next;

	rtl_pthread_mutex_unlock(&(pid->access_locker));

	return (buf);
}

signed char fifo_reserve_abort(ALLOC_ID id, void *pbuf)
{
#ifdef QNX_TAGGING
	if ( fifo_unget( id, pbuf, NULL ) < 0 )
		return -1;
#else
	if ( fifo_unget( id, pbuf ) < 0 )
		return -1;
#endif
	return 0;
}

#ifdef QNX_TAGGING
signed char fifo_putmes(ALLOC_ID id, void *pbuf, unsigned int qnx_tag)
#else
signed char fifo_putmes(ALLOC_ID id, void *pbuf)
#endif
{
	DATA_SLOT **temp_p;
	OBJECT_POINT *pid;

	if ( checkfifo(id) < 0 ) return -1;
	
	pid = &(object_points[id]);

	rtl_pthread_mutex_lock(&(pid->access_locker));
	
	/* Put in the tail */
	temp_p = &(pid->mrd);
	while ( *temp_p != NULL )
		temp_p = &((*temp_p)->next);

	*temp_p = ((DATA_SLOT *)pbuf) -1;
	(*temp_p)->next = NULL;

#ifdef QNX_TAGGING
	(*temp_p)->tagging = qnx_tag;
#endif
	
	rtl_pthread_mutex_unlock(&(pid->access_locker));

	return 0;
}

unsigned char *fifo_getmes(ALLOC_ID id)
{
	OBJECT_POINT *pid;
	unsigned char *buf;

	if ( checkfifo(id) < 0 ) return NULL;

	pid = &(object_points[id]);

	rtl_pthread_mutex_lock(&(pid->access_locker));

	/* return if empty */
	if (pid->mrd == NULL){
		rtl_pthread_mutex_unlock(&(pid->access_locker));
		return NULL;
	}

	buf = pid->mrd->mem_p;
	pid->mrd = pid->mrd->next;

	rtl_pthread_mutex_unlock(&(pid->access_locker));

	return ( (unsigned char *)buf );
}

#ifdef QNX_TAGGING
signed char fifo_unget(ALLOC_ID id, void *pun_mes, unsigned int *ret_qnx_tag)
#else
signed char fifo_unget(ALLOC_ID id, void *pun_mes)
#endif
{
	DATA_SLOT *pbuf;
	OBJECT_POINT *pid;

	if ( checkfifo(id) < 0 ) return -1;

	pid = &(object_points[id]);

	rtl_pthread_mutex_lock(&(pid->access_locker));

	/* Let's get the pointer to the block pointer */
	/* Block: (DATA_SLOT) + data
	 * so, the block pointer comes with:
	 *   data_p - sizeof(DATA_SLOT)
	 */
	pbuf = ((DATA_SLOT *)(pun_mes)) - 1;
#ifdef QNX_TAGGING
	if (ret_qnx_tag != NULL) *ret_qnx_tag=pbuf->tagging;
#endif
	pbuf->next = pid->free_p;
	pid->free_p = pbuf;

	/* SEM */
	FTT_SEMAPHORE_post(&(pid->free_items_sem));

	rtl_pthread_mutex_unlock(&(pid->access_locker));

	/* return if the fifo got empty */
	if (pid->mrd == NULL)
		return 1;
	else
		return 0;
}

signed short fifo_size(ALLOC_ID id)
{
	DATA_SLOT **temp_p;
	OBJECT_POINT *pid;
	unsigned short ret_counter;

	if ( checkfifo(id) < 0 ) return -1;

	pid = &(object_points[id]);

	/* Put in the tail */
	temp_p = &(pid->mrd);
	ret_counter = 0;
	while ( *temp_p != NULL ){
		temp_p = &((*temp_p)->next);
		ret_counter ++;
	}

	return (ret_counter);
}

signed short fifo_is_loaded(ALLOC_ID id)
{
	OBJECT_POINT *pid;

	if ( checkfifo(id) < 0 ) return 0;

	pid = &(object_points[id]);

	/* return if empty */
	if (pid->mrd == NULL)
		return 0;
	return 1;
}

void fifo_print(ALLOC_ID id)
{
	int i=id;
	DATA_SLOT *temp_p;

	if ( checkfifo(i) < 0 ) return;


	if (object_points[i].in_use == USE_AS_FIFO){//if a FIFO
			PRINTF("\n[%2d] %d*%d  next_f_p:%d  g_mem:%p ",
					i,
					object_points[i].mem_slot_num,
					object_points[i].mem_slot_size,
					object_points[i].next_free_point,
					object_points[i].g_mem
					);
			
			PRINTF("free_d: ");
			temp_p = object_points[i].free_p;
			while (temp_p != NULL){
				PRINTF("(%p,%d)->",temp_p->mem_p, temp_p->n_uso);
				temp_p = temp_p->next;
			}
			PRINTF("NULL");
			
			PRINTF("  mrd: ");

			temp_p = object_points[i].mrd;
			while (temp_p != NULL){
				PRINTF("(%p,%d)->",temp_p->mem_p, temp_p->n_uso);
				temp_p = temp_p->next;
			}
			PRINTF("NULL");
		}	
}


void fifo_print_all(void)
{
	int i;
	DATA_SLOT *temp_p;
	for (i=0; i<MAX_OBJS; i++){
		if (object_points[i].in_use == USE_AS_FIFO){//if a FIFO
			PRINTF("\n[%2d] %d*%d  next_f_p:%d  g_mem:%p ",
					i,
					object_points[i].mem_slot_num,
					object_points[i].mem_slot_size,
					object_points[i].next_free_point,
					object_points[i].g_mem
					);
			
			PRINTF("free_d: ");
			temp_p = object_points[i].free_p;
			while (temp_p != NULL){
				PRINTF("(%p,%d)->",temp_p->mem_p, temp_p->n_uso);
				temp_p = temp_p->next;
			}
			PRINTF("NULL");
			
			PRINTF("  mrd: ");

			temp_p = object_points[i].mrd;
			while (temp_p != NULL){
				PRINTF("(%p,%d)->",temp_p->mem_p, temp_p->n_uso);
				temp_p = temp_p->next;
			}
			PRINTF("NULL");


		}	
	}
	PRINTF("\n");
}

