/*****************************************************************************
 * S_db.c:
 *****************************************************************************
 * Copyright (C) 2006-2012 the FTT-SE team.
 *
 * Author: Ricardo Marau <marau at fe.up.pt>
 *
 * FTT-SE is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  FTT-SE is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with FTT-SE.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

#include "Ports.h"

#include "S_db.h"
#include "ethernet.layer.h"

#include "S_nodeID.h"
#include "S_igmp_queue.h"


typedef enum{
	NDB_DIRECTION_PRODUCER = 0,
	NDB_DIRECTION_CONSUMER
}NDB_DIRECTION;

typedef enum{
	NDB_TX_NEW_DAT = 0x01,
	NDB_RX_NEW_DAT = 0x02
}NDB_STATUS;


typedef struct {
	unsigned short  m_id;	       ///< Message id
	unsigned int  m_size;        ///< n. of data bytes 
	unsigned int  m_max_size;        ///< n. of data bytes 
	unsigned short  m_MTU;
	unsigned int  m_period;
	NDB_DIRECTION m_direction; 
	unsigned char m_mac_var[6]; ///<Variable used to store a MAC
	unsigned char m_prod_max_no;
	unsigned char am_queue_notify; //-besides the possibility of enquiring directly the allocator
	                               //-I am tracking the AM vars that haven't yet been notified
                                       //-I can't put it global to all the reqs whithin an ID 'cause the
                                       //notification must be tag sensitive
} NDB_MESG_REQ_REQ;

typedef struct {
		unsigned short packets_no;      ///< n. of packets
		struct{
			unsigned long One_Bsize;      ///< One message data size in bytes
			unsigned long Last_Bsize;     ///< Last message data size in bytes
			unsigned long Last_Bsize_real;///< Last message data size in bytes - without inflation!
		}d_size;
} NDB_MESG_REQ_COMPUTATION;

static signed char NDB_mesg_req_req_validation(NDB_MESG_REQ_REQ *from)
{
	if(from->m_max_size > 2000000) return -1;
	
	if ( from->m_size > from->m_max_size )
		return -2;

	/* MTU validation */
	if( FTT_VAR_ID_IS_ID_SYNC(from->m_id) ){
		if(from->m_MTU > FTT_PKT_SDM_DATALEN_MAX)
			return -3;
	}else{
		if(from->m_MTU > FTT_PKT_ADM_DATALEN_MAX)
			return -3;
	}
	
	if (from->m_prod_max_no==0) return -4;
	
	return 0;
}

static void NDB_mesg_req_computation_do_it(NDB_MESG_REQ_COMPUTATION *to, NDB_MESG_REQ_REQ *from)
{
	int i;
#define REALSIZE_B(_i)	(( FTT_VAR_ID_IS_ID_SYNC(from->m_id) )? FTT_PKT_SDM_REALSIZE_B(_i): FTT_PKT_ADM_REALSIZE_B(_i))
#define X_DATALEN_MIN	(( FTT_VAR_ID_IS_ID_SYNC(from->m_id) )? FTT_PKT_SDM_DATALEN_MIN: FTT_PKT_ADM_DATALEN_MIN)
	to->packets_no = from->m_size / from->m_MTU;
	if ((i=(from->m_size %  from->m_MTU)))
		to->packets_no ++;
	if (i){
		/* Last packet not full */
		to->d_size.Last_Bsize = to->d_size.Last_Bsize_real = (i);
		if( i < X_DATALEN_MIN )
			to->d_size.Last_Bsize = (X_DATALEN_MIN);
	}else{
		/* Last packet full */\
		to->d_size.Last_Bsize = to->d_size.Last_Bsize_real = (from->m_MTU);
	}
	to->d_size.One_Bsize = (from->m_MTU);
	return;
}

typedef struct ndb_mesg_req{
	NDB_MESG_REQ_REQ req;
	NDB_MESG_REQ_COMPUTATION comp;
	struct ndb_mesg_req *link;
} NDB_MESG_REQ;

inline static void NDB_MESG_REQ_print(NDB_MESG_REQ *p, unsigned char cr){
	PRINTF(" (0x%x  %dB)", p->req.m_id, p->req.m_size );

	if (cr){ PRINTF("\n");}
}

#define NDB_MAX_VAR_REQ  (STA_MAX_VAR*3)

static NDB_MESG_REQ NDB_mesg_req_pool[NDB_MAX_VAR_REQ];

static NDB_MESG_REQ *NDB_mesg_req_free_items;

static void NDB_mesg_req_add_this_point(NDB_MESG_REQ **here, NDB_MESG_REQ *item)
{
	/* insert in the list top */
	item->link = *here;
	*here = item;
}

static NDB_MESG_REQ **NDB_mesg_req_goto_NULL_pp(NDB_MESG_REQ **here)
{
	/* Goto_last_pp - means get the address of the pointer that finishes the list with NULL */
	while ( (*here) != NULL )
		here = &((*here)->link);

	return here;
}

static NDB_MESG_REQ *NDB_mesg_req_goto_last_item_p(NDB_MESG_REQ *here)
{
	if ( here==NULL ) return NULL;
	
	while ( (here->link) != NULL )
		here = (here->link);

	return here;
}

static signed char NDB_mesg_req_get_from_top(NDB_MESG_REQ **queue_p, NDB_MESG_REQ **ret_p)
{
	if (*queue_p == NULL)
		return -1;

	*ret_p = *queue_p;
	*queue_p = (*queue_p)->link;

	return 0;
}

static signed char NDB_mesg_req_empty_from_here(NDB_MESG_REQ **queue_p)
{
	NDB_MESG_REQ *ret_p;

	if (*queue_p == NULL)
		return 0;

	while ( (*queue_p) != NULL ){
		NDB_mesg_req_get_from_top(queue_p, &ret_p);
		NDB_mesg_req_add_this_point(&NDB_mesg_req_free_items, ret_p);
	}
	return 0;
}

static signed char NDB_mesg_req_empty_until_tagid(NDB_MESG_REQ **queue_p, unsigned short tagid)
{
	NDB_MESG_REQ *ret_p;

	if (*queue_p == NULL)
		return -1;

	while ( (*queue_p) != NULL ){
		if ( FTT_VAR_ID_CMP_ID_TYPE_TAG( (*queue_p)->req.m_id, tagid) )
			return 0;
		NDB_mesg_req_get_from_top(queue_p, &ret_p);
		NDB_mesg_req_add_this_point(&NDB_mesg_req_free_items, ret_p);
	}
	return -2;
}

static signed char NDB_mesg_req_empty_until_tagid_inclusive(NDB_MESG_REQ **queue_p, unsigned short tagid)
{
	NDB_MESG_REQ *ret_p;
	
	if (*queue_p == NULL)
		return -1;
	
	while ( (*queue_p) != NULL ){
		if ( FTT_VAR_ID_CMP_ID_TYPE_TAG( (*queue_p)->req.m_id, tagid) ){
			NDB_mesg_req_get_from_top(queue_p, &ret_p);
			NDB_mesg_req_add_this_point(&NDB_mesg_req_free_items, ret_p);
			return 0;
		}
		NDB_mesg_req_get_from_top(queue_p, &ret_p);
		NDB_mesg_req_add_this_point(&NDB_mesg_req_free_items, ret_p);
	}
	return -1;
}

static signed char NDB_mesg_req_is_tagid_present(NDB_MESG_REQ *queue_p, unsigned short tagid)
{
	if (queue_p == NULL) return -1;

	while ( queue_p != NULL ){
		if ( FTT_VAR_ID_CMP_ID_TYPE_TAG( (queue_p)->req.m_id, tagid) )
			return 1;
		queue_p = queue_p->link;
	}

	return 0;
}

static signed char NDB_mesgtable_pool_Init(void)
{
	int i;

	/* NDB_mesgtable init */
	NDB_mesg_req_free_items = NULL;
	for( i=0; i < NDB_MAX_VAR_REQ; i++){
		NDB_mesg_req_add_this_point( &NDB_mesg_req_free_items, &(NDB_mesg_req_pool[i]) );
	}
	return 0;
}

static signed char NDB_mesgtable_pool_Close(void)
{
	return 0;
}

static NDB_MESG_REQ **NDB_mesg_req_track_tag_id ( NDB_MESG_REQ **t_RT, unsigned short vid ){
	/* Check for bad input pointer */
	if ( t_RT == NULL ) return ( NULL );

	while( *t_RT != NULL ){
		if ( FTT_VAR_ID_CMP_ID_TYPE_TAG( (*t_RT)->req.m_id, vid) )
			break;
		t_RT = &((*t_RT)->link);
	}

	return t_RT;
}

static unsigned char NDB_mesg_req_check_existence(NDB_MESG_REQ **temp, unsigned short id){
	return ( (*NDB_mesg_req_track_tag_id(temp, id)!=NULL) );
}

#define MAX_POSSIBLE_PRODUCERS 40 /* This is the same value hardcoded in the arrays in A_MST_ADD */

typedef struct {
	unsigned short tagless_id;
	unsigned int max_size;
	unsigned char max_prod_no;
	NDB_DIRECTION direction;
	unsigned char status;
	
	NDB_MESG_REQ *reqs;
	
	ALLOC_ID allocator;
	FTT_SIGNAL callback_signal;	///During the var lifetime I can't change its memory position.

	/* dynamic */
	void *allocator_upside_lock_p;
	
	struct
	{
		void *allocator_downside_lock_p;
		unsigned short fragmentation_seq_number;
		signed int source_tag; /* For producers: not used! Actualy only downside[0] is used.
								* For consumers: Identifies the source id that reserved this memory slot 
								* If < 0 - not used */
	}downside[MAX_POSSIBLE_PRODUCERS];
	
//	void *allocator_downside_lock_p;
//	unsigned short fragmentation_seq_number;
} NDB_mesgtype;

typedef struct ndb_tab_type{
	NDB_mesgtype mesg;
	struct ndb_tab_type *link;
}NDB_tabtype;


static signed char NDB_mesgtype_downside_get_idx_and_tag_it( NDB_mesgtype *mesg_p, unsigned char source_tag, unsigned char *ret_idx )
{
	int i;
	int first_free = -1;
	
	for ( i=0; i<MAX_POSSIBLE_PRODUCERS; i++ ){
		if ( mesg_p->downside[i].source_tag == source_tag ){
			*ret_idx = i;
			return 0;
		}
		if (first_free<0)
			if (mesg_p->downside[i].source_tag<0 )
				first_free = i;		
	}
	
	if (first_free<0) return -1;
	
	*ret_idx = first_free;
	mesg_p->downside[first_free].source_tag = source_tag;
	return 1;
}

static signed char NDB_mesgtype_downside_free_idx( NDB_mesgtype *mesg_p, unsigned char idx )
{

	if (idx >= MAX_POSSIBLE_PRODUCERS) return -1;
	
	mesg_p->downside[idx].source_tag = -1;

	return 0;
}

inline static void NDB_tabtype_print(NDB_tabtype *p, unsigned char cr)
{
	NDB_MESG_REQ *temp_p;
	if ( FTT_VAR_ID_IS_ID_SYNC( p->mesg.tagless_id ) ) {PRINT_MSG("'S'");}
		else {PRINT_MSG("'A'");}

	PRINTF(" ID: 0x%4x  MAX%6dB:", p->mesg.tagless_id, p->mesg.max_size );

	if ( (p->mesg.direction) == NDB_DIRECTION_PRODUCER){ PRINTF(" Producer"); }
	else{ PRINTF(" Consumer"); }

	if ( (p->mesg.allocator) == -1 ){ PRINTF(" NOBind");}
	else { PRINTF("   Bind"); }

	temp_p = p->mesg.reqs;
	if (temp_p == NULL) {PRINTF ("(NULL)");}
	while(temp_p != NULL){
		NDB_MESG_REQ_print(temp_p, 0);
		temp_p = temp_p->link;
	}

	if (cr){ PRINTF("\n");}
}


static void NDB_tabtype_add_this_point(NDB_tabtype **queue_p, NDB_tabtype *p)
{
	/* insert in the list top */
	p->link = *queue_p;
	*queue_p = p;
}

static signed char NDB_tabtype_get_from_top(NDB_tabtype **queue_p, NDB_tabtype **ret_p)
{
	if (*queue_p == NULL)
		return -1;

	*ret_p = *queue_p;
	*queue_p = (*queue_p)->link;

	return 0;
}

#define NDB_MAX_VARS  (STA_MAX_VAR)

static NDB_tabtype NDB_mesg_pool[NDB_MAX_VARS];

static NDB_tabtype *NDB_free_items;
static NDB_tabtype *NDB_inuse_items; //Variable and memory assigned

static signed char NDB_table_pool_Init(void)
{
	int i;

	/* NDB_mesgtable init */
	NDB_free_items = NDB_inuse_items = NULL;
	for( i=0; i < NDB_MAX_VARS; i++){
		NDB_tabtype_add_this_point( &NDB_free_items, &(NDB_mesg_pool[i]) );
	}
	return 0;
}
static signed char NDB_table_pool_Close(void)
{
	return 0;
}

static NDB_tabtype **NDB_table_track_tagless_id ( NDB_tabtype **t_RT, unsigned short vid ){
	/* Check for bad input pointer */
	if ( t_RT == NULL ) return ( NULL );

	while( *t_RT != NULL ){
		if ( FTT_VAR_ID_CMP_ID_TYPE( (*t_RT)->mesg.tagless_id, vid) )
			break;
		t_RT = &((*t_RT)->link);
	}

	return t_RT;
}

//static unsigned char NDB_table_check_existence(NDB_tabtype **temp, unsigned short id){
//	return ( (*NDB_table_track_tagless_id(temp, id)!=NULL) );
//}

static unsigned short NDB_table_Get_elements_no ( NDB_tabtype **t_RT ){

	unsigned short counter;
	/* Check for bad input pointer */
	if ( t_RT == NULL ) return 0;

	counter = 0;

	while( *t_RT != NULL ){
		counter ++;
		t_RT = &((*t_RT)->link);
	}

	return (counter);
}


signed char NDB_table_print( void )
{
//	int i;
	NDB_tabtype **temp_p = &NDB_inuse_items;

	/* Scan and print table */
	PRINT_MSG("NDB table contents (%u mesgs stored)", NDB_table_Get_elements_no( temp_p ));

	while( *temp_p != NULL ){
		NDB_tabtype_print( *temp_p, 0);
		temp_p = &((*temp_p)->link);
	}

	return 0;
}

static signed char NDB_table_mesg_append_req (NDB_tabtype *mesg_p, NDB_MESG_REQ *req_p )
{
	NDB_MESG_REQ **temp_pp;
	/* I am assuming that the id of t*/
	if ( !FTT_VAR_ID_CMP_ID_TYPE( mesg_p->mesg.tagless_id, req_p->req.m_id) ){
		ERROR_MSG("Programming error - this shouldn't happen");
		return -1;
	}

	if (NDB_mesg_req_check_existence( &(mesg_p->mesg.reqs), req_p->req.m_id)){
		/* There's a slot with the same requirements id and tag */
		return -2;
	}

	/* Let's append it in last */
	temp_pp = NDB_mesg_req_goto_NULL_pp( &(mesg_p->mesg.reqs) );

	NDB_mesg_req_add_this_point( temp_pp, req_p );

	return 0;// appended
}

static signed char NDB_table_mesg_get_new_message_slot (NDB_tabtype **ret_mesg_p, unsigned short id, unsigned int max_size, unsigned char max_prod_no, NDB_DIRECTION direction )
{
	int i;
	
	if ( NDB_tabtype_get_from_top( &NDB_free_items, ret_mesg_p ) < 0 ){
		ERROR_MSG("No available mesg items");
		return -1;
	}
	FTT_VAR_ID_SET_TAG( id, 0 );
	(*ret_mesg_p)->mesg.tagless_id = id;
	(*ret_mesg_p)->mesg.max_size = max_size;
	(*ret_mesg_p)->mesg.max_prod_no = max_prod_no;
	(*ret_mesg_p)->mesg.direction = direction;
	(*ret_mesg_p)->mesg.status = 0;
	(*ret_mesg_p)->mesg.reqs = NULL;
	(*ret_mesg_p)->mesg.allocator = -1;
	(*ret_mesg_p)->mesg.allocator_upside_lock_p = NULL;
	for (i=0; i<MAX_POSSIBLE_PRODUCERS; i++) {
		((*ret_mesg_p)->mesg.downside)[i].allocator_downside_lock_p = NULL;
		((*ret_mesg_p)->mesg.downside)[i].fragmentation_seq_number = 0;
		((*ret_mesg_p)->mesg.downside)[i].source_tag = -1; /* Not in use */
	}

	if ( FTT_signal_init( &((*ret_mesg_p)->mesg.callback_signal) )<0 ){
		ERROR_MSG("Couldn't initialize the signal");
		/* Let's undo the slot allocation */
		NDB_tabtype_add_this_point( &NDB_free_items, *ret_mesg_p);
		return -2;
	}
	
	NDB_tabtype_add_this_point( &NDB_inuse_items, *ret_mesg_p);

	return 0;
}

static signed char NDB_table_mesg_get_del_message_slot (unsigned short id)
{
	NDB_tabtype **mesg_pp;
	NDB_tabtype *mesg_p;
	/* track the pp */
	if ( (*(mesg_pp=NDB_table_track_tagless_id( &NDB_inuse_items, id)) == NULL) )
		return -1;

	/* Get it from the list where it was */
	NDB_tabtype_get_from_top( mesg_pp, &mesg_p);
	
	/* Handle the multicast anouncement */
	if (mesg_p->mesg.direction == NDB_DIRECTION_CONSUMER ){
		S_igmp_queue_send_leave_if_multicast( (mesg_p->mesg.reqs)->req.m_mac_var, S_Get_My_mac() );
	}

	/* Remove all the piggybacked requirements */	
	NDB_mesg_req_empty_from_here( &(mesg_p->mesg.reqs) );

	/* Remove the signal instantiation */
	FTT_signal_destroy( &(mesg_p->mesg.callback_signal) );
	
	/* place the mesg in the free items queue */
	NDB_tabtype_add_this_point(&NDB_free_items, mesg_p);

	return 0;
}

signed char NDB_add_msg(
		MESSAGE_TYPE m_type,
		unsigned short m_id,
		unsigned char m_tag,
		unsigned int m_size,
		unsigned int m_max_size,
		unsigned int m_period,
		unsigned short m_MTU,
		unsigned char *m_prod_mac,
		unsigned char m_prod_max_no)
		
{
	unsigned char new_flag;
	NDB_MESG_REQ *req_temp_p; // a requirement pointer
	NDB_tabtype *mesg_temp_p; // a message slot pointer


	/* Get new message requirements table slot */
	if ( NDB_mesg_req_get_from_top( &NDB_mesg_req_free_items, &req_temp_p) < 0 ){
		ERROR_MSG("Couldn't get any free requirements slot");
		return -1;
	}
	
	FTT_VAR_ID_SET_ID( m_id, m_id );
	if ( (m_type&0x0F) == 0 ){
		FTT_VAR_ID_SET_TYPE_SYNC( m_id );
	}else{
		FTT_VAR_ID_SET_TYPE_ASYNC( m_id );
	}
	FTT_VAR_ID_SET_TAG( m_id, m_tag );
//PRINT_MSG(">>>>>>>>>>>>>>>>>Preparing to add %x  size:%d  m_size:%d", m_id,m_size, m_max_size );
//NDB_table_print();

	/* Now that we have the req slot let's populate it */
	req_temp_p->req.m_id = m_id;
	req_temp_p->req.m_direction = ( (m_type>>4) == 0 ) ? NDB_DIRECTION_CONSUMER : NDB_DIRECTION_PRODUCER;
	
	ETH_L_CopyMAC_to_from(req_temp_p->req.m_mac_var, m_prod_mac);
	req_temp_p->req.m_prod_max_no = m_prod_max_no;

	req_temp_p->req.m_size = m_size;
	req_temp_p->req.m_max_size = m_max_size;
	req_temp_p->req.m_period = m_period;
	req_temp_p->req.m_MTU = m_MTU;

	req_temp_p->req.am_queue_notify = 0;

	/* Go with the req validation */
	if ( NDB_mesg_req_req_validation( &(req_temp_p->req) ) < 0 ){
		ERROR_MSG("Failed validation test");
		return -2;
	}

	/* Go with the offline optimization calculus */
	NDB_mesg_req_computation_do_it( &(req_temp_p->comp), &(req_temp_p->req) );
	
	/* NOW - the message slot */
	new_flag = 0;
	if( (mesg_temp_p = *NDB_table_track_tagless_id ( &NDB_inuse_items, m_id )) == NULL ){
//		PRINT_MSG("Going for new var table slot");
		new_flag = 1;
		/* This type+ID still doesn't exist in table - get one new */
		if ( NDB_table_mesg_get_new_message_slot ( &mesg_temp_p, m_id, m_max_size, m_prod_max_no, (( (m_type>>4) == 0 ) ? NDB_DIRECTION_CONSUMER : NDB_DIRECTION_PRODUCER ) ) < 0 ){
			ERROR_MSG("Couldn't get a new functional mesg item");
			/* release the req_item */
			NDB_mesg_req_add_this_point( &NDB_mesg_req_free_items, req_temp_p);
			return -3;
		}
	}
	
	/* Validating immutable parameters */
	if ( !new_flag ){
		/* Checking if adding with opposite directions */
		if ( mesg_temp_p->mesg.direction != req_temp_p->req.m_direction ) {
			ERROR_MSG("I'm registering a new requirments tag with opposite direction. This is not supported. Delete the var before!");
			NDB_mesg_req_add_this_point( &NDB_mesg_req_free_items, req_temp_p);
			return -4;
		}
	
		/* Mesg_max_prod_no */
		if ( mesg_temp_p->mesg.max_prod_no != req_temp_p->req.m_prod_max_no ){
			ERROR_MSG(" Message max_production no. not matching ");
			return -5;
		}
		
		/* Mesg_max_size */
		if ( mesg_temp_p->mesg.max_size != req_temp_p->req.m_max_size ){
			ERROR_MSG(" Message max_message_size not matching ");
			return -6;
		}
	}
	
	
	/* We have a message slot (either new or not) */
	

	/* Let's append the new requirement if the one(s) residing have a different tag is */
	if ( NDB_table_mesg_append_req ( mesg_temp_p, req_temp_p ) < 0 ){
		/* Failed - probably the tag was already assigned in the mesg slot */
		if ( new_flag ){
			NDB_mesg_req_add_this_point( &NDB_mesg_req_free_items, req_temp_p);
			NDB_table_mesg_get_del_message_slot (m_id);			
		}
	}

	/* Handle the multicast anouncement */
	if (req_temp_p->req.m_direction == NDB_DIRECTION_CONSUMER ){
		S_igmp_queue_send_report_if_multicast( req_temp_p->req.m_mac_var, S_Get_My_mac() );
	}
	
//NDB_table_print();
//PRINT_MSG(">>>>>>>>>>>>>>>>>>>>>>Finish Add 0x%04x", m_id);

	return 0;

}
static signed char NDB_bind_mesg_internal(
		unsigned short m_id,
		FTT_SIGNAL **ret_callback_signal,
		void *memory_point,
		unsigned int mem_size,
		unsigned char slots_no)
{
	NDB_tabtype *temp_mesg_p;
	unsigned char extra_slots = 0;

	if( (temp_mesg_p = *NDB_table_track_tagless_id( &NDB_inuse_items, m_id)) == NULL ){
		return -1;
	}

	if (ret_callback_signal != NULL)
		*ret_callback_signal = &(temp_mesg_p->mesg.callback_signal);

	/*Update the slots_no to include the possibility of having multiple producers locking onto the same FIFO/CAB */
	if (temp_mesg_p->mesg.direction == NDB_DIRECTION_CONSUMER )
		extra_slots = (temp_mesg_p->mesg.max_prod_no - 1);
	
	if( FTT_VAR_ID_IS_ID_SYNC( m_id ) ){
		/* Set-up a CAB for message data manipulation */
		/* Note that there must be room for n+1 megs if n tasks use the CAB*/
		if (slots_no != STA_SM_BUFF_SIZE){
			ERROR_MSG("Synchronous messages must be bound with a message slot = %d... forcing, continue 'n hope to have memory enough", STA_SM_BUFF_SIZE);
			slots_no = STA_SM_BUFF_SIZE;
		}
		temp_mesg_p->mesg.allocator = cab_create( temp_mesg_p->mesg.max_size, slots_no+extra_slots, memory_point, mem_size);
		
		if( temp_mesg_p->mesg.allocator < 0){
			ERROR_MSG("Creating CAB of var %x!",temp_mesg_p->mesg.tagless_id);
			return -2;
		}
	}else{
		if (slots_no < STA_AM_BUFF_SIZE){
			ERROR_MSG("Asynchronous messages must be bound with a message slot = %d... forcing, continue 'n hope to have memory enough", STA_AM_BUFF_SIZE);
			slots_no = STA_AM_BUFF_SIZE;
		}
		temp_mesg_p->mesg.allocator = fifo_create(temp_mesg_p->mesg.max_size, slots_no+extra_slots, memory_point, mem_size);
		
		if(temp_mesg_p->mesg.allocator < 0){
			ERROR_MSG("Creating FIFO of var %x!",temp_mesg_p->mesg.tagless_id);
			return -2;
		}

	}

	//	PRINT_MSG("Binded 0x%x",m_id );
	return 0;
}

signed char NDB_bind_mesg(
		MESSAGE_TYPE m_type,
		unsigned short m_id,
		FTT_SIGNAL **ret_callback_signal,
		void *memory_point,
		unsigned int mem_size,
		unsigned char slots_no )
{
	FTT_VAR_ID_SET_ID( m_id, m_id );
	if ( (m_type&0x0F) == 0 ){
		FTT_VAR_ID_SET_TYPE_SYNC( m_id );
	}else{
		FTT_VAR_ID_SET_TYPE_ASYNC( m_id );
	}

	return NDB_bind_mesg_internal( m_id, ret_callback_signal, memory_point, mem_size, slots_no);
}

static signed char NDB_m_pre_produce_abort(unsigned short MesgID);

static signed char NDB_unbind_mesg_internal( unsigned short m_id)
{
	NDB_tabtype *temp_mesg_p;

	if( (temp_mesg_p = *NDB_table_track_tagless_id( &NDB_inuse_items, m_id)) == NULL ){
		return -1;
	}

	/* Give back any reserved in the allocator - Checking only allocator_upside_lock_p.
	 * There should be no race condition with allocator_downside_lock_p */
	if (temp_mesg_p->mesg.allocator_upside_lock_p != NULL	){
		//ERROR_MSG("I should be doing sth here: allocator_upside_lock_p not NULL");
		NDB_m_pre_produce_abort(m_id);
	}

//	Removed when updating the "allocator_downside_lock_p" to downside[i].allocator_downside_lock_p
//  It seemed unnecessary given that by deleting the CAB/FIFO all memory is re-captured.
#warning Have to check this! It might be that a mesage is un-bind while the locker was in the middle of sth!!!!
//	if (temp_mesg_p->mesg.allocator_downside_lock_p != NULL	){
//		temp_mesg_p->mesg.allocator_downside_lock_p = NULL;
//	}
	
	if( FTT_VAR_ID_IS_ID_SYNC( m_id ) ){
		/* Remove the CAB associated with the synchronous message */
		cab_delete( temp_mesg_p->mesg.allocator );
		temp_mesg_p->mesg.allocator = -1;
	}else{
		/* Remove the PORT associated with the asynchronous message */
		fifo_delete( temp_mesg_p->mesg.allocator);
		temp_mesg_p->mesg.allocator = -1;
	}

	return 0;
}


signed char NDB_unbind_mesg( MESSAGE_TYPE m_type, unsigned short m_id )
{
	FTT_VAR_ID_SET_ID( m_id, m_id );
	if ( (m_type&0x0F) == 0 ){
		FTT_VAR_ID_SET_TYPE_SYNC( m_id );
	}else{
		FTT_VAR_ID_SET_TYPE_ASYNC( m_id );
	}
	
	
	return NDB_unbind_mesg_internal( m_id );
}

static signed char NDB_isbind_mesg_internal( unsigned short m_id)
{
	NDB_tabtype *temp_mesg_p;
	
	if( (temp_mesg_p = *NDB_table_track_tagless_id( &NDB_inuse_items, m_id)) == NULL )
		return 0;
	
	if (temp_mesg_p->mesg.allocator < 0) return 0;
	else return 1;
}

signed char NDB_isbind_mesg( MESSAGE_TYPE m_type, unsigned short m_id )
{
	FTT_VAR_ID_SET_ID( m_id, m_id );
	if ( (m_type&0x0F) == 0 ){
		FTT_VAR_ID_SET_TYPE_SYNC( m_id );
	}else{
		FTT_VAR_ID_SET_TYPE_ASYNC( m_id );
	}
	
	return NDB_isbind_mesg_internal( m_id );
}

static signed char NDB_get_current_internal(
		unsigned int m_id,
		unsigned char *tag,
		unsigned int *cur_size,
		unsigned int *max_size,
		unsigned short *period )
{
	NDB_tabtype *temp_mesg_p;
	NDB_MESG_REQ *mesg_req;

	/* Track this message ID - tagless */
	if( (temp_mesg_p = *NDB_table_track_tagless_id( &NDB_inuse_items, m_id)) == NULL ){
		return -1;
	}

	/* now - for this ID let's retrieve the parameters for the most recent tag */
	if ( ( mesg_req = NDB_mesg_req_goto_last_item_p( temp_mesg_p->mesg.reqs ) ) == NULL ){
		ERROR_MSG("No requirements available in this table slot: %x", m_id );
		return -2;
	}

	if (tag != NULL) *tag = FTT_VAR_ID_GET_TAG( mesg_req->req.m_id );
	if (cur_size != NULL) *cur_size = mesg_req->req.m_size;
	if (max_size != NULL) *max_size = mesg_req->req.m_max_size;
//	PRINT_MSG("in NDB_get_current_internal m_max_size: %d", mesg_req->req.m_max_size);
	if (period != NULL) *period = mesg_req->req.m_period;

	return 0;
}

signed char NDB_get_current(
		MESSAGE_TYPE m_type,
		unsigned int m_id,
		unsigned char *tag,
		unsigned int *cur_size,
		unsigned int *max_size,
		unsigned short *period )
{
	FTT_VAR_ID_SET_ID( m_id, m_id );
	if ( (m_type&0x0F) == 0 ){
		FTT_VAR_ID_SET_TYPE_SYNC( m_id );
	}else{
		FTT_VAR_ID_SET_TYPE_ASYNC( m_id );
	}
	return NDB_get_current_internal( m_id, tag, cur_size, max_size, period);
}

static signed char NDB_get_memory_pool_size_internal( unsigned int m_id, unsigned char slots_no, unsigned short *ret_mem_size )
{
	NDB_tabtype *temp_mesg_p;
	unsigned char extra_slots = 0;
	
	/* Track this message ID - tagless */
	if( (temp_mesg_p = *NDB_table_track_tagless_id( &NDB_inuse_items, m_id)) == NULL ){
		return -1;
	}
	
	if (temp_mesg_p->mesg.direction == NDB_DIRECTION_CONSUMER )
		extra_slots = (temp_mesg_p->mesg.max_prod_no - 1);
	
	if( FTT_VAR_ID_IS_ID_SYNC( m_id ) ){
		if (slots_no != STA_SM_BUFF_SIZE){
			ERROR_MSG("Synchronous messages must be bound with a message slot = %d... forcing, continue 'n hope to have memory enough", STA_SM_BUFF_SIZE);
			slots_no = STA_SM_BUFF_SIZE;
		}
	}else{
		if (slots_no < STA_AM_BUFF_SIZE){
			ERROR_MSG("Asynchronous messages must be bound with a message slot = %d... forcing, continue 'n hope to have memory enough", STA_AM_BUFF_SIZE);
			slots_no = STA_AM_BUFF_SIZE;
		}
	}
	
	if (ret_mem_size) *ret_mem_size = (unsigned short) Allocator_get_memory_size_needed( temp_mesg_p->mesg.max_size, slots_no+extra_slots);

	return 0;
}
signed char NDB_get_memory_pool_size(
							MESSAGE_TYPE m_type,
							unsigned int m_id,
							unsigned char slots_no,
							unsigned short *ret_mem_size  )
{
	FTT_VAR_ID_SET_ID( m_id, m_id );
	if ( (m_type&0x0F) == 0 ){
		FTT_VAR_ID_SET_TYPE_SYNC( m_id );
	}else{
		FTT_VAR_ID_SET_TYPE_ASYNC( m_id );
	}
	return NDB_get_memory_pool_size_internal( m_id, slots_no, ret_mem_size);
}


static signed char NDB_del_msg_internal( unsigned short m_id)
{

	/* Let's first unbind it */
	if ( NDB_unbind_mesg_internal( m_id ) < 0){
		return -1;
	}
	
	if (NDB_table_mesg_get_del_message_slot (m_id) < 0){
		return -2;
	}
	return 0;
}

static signed char NDB_clean_msg_tag_internal( unsigned short m_id)
{

	/* This function is called explecitely to clean old tags.
	 * It might be redundant with the cleaning that's in place, in which
	 * the old tags are removed upon a production/consumption on a newer tag
	 */
	NDB_tabtype *temp_mesg_p;
	NDB_MESG_REQ **mesg_req_pp;
	NDB_MESG_REQ *mesg_req_p;
	signed char is_tag_here;

	
	/* Track this message ID - tagless */
	if( (temp_mesg_p = *NDB_table_track_tagless_id( &NDB_inuse_items, m_id)) == NULL ){
		return -1;
	}
	
	is_tag_here = NDB_mesg_req_is_tagid_present( temp_mesg_p->mesg.reqs, m_id );
	if (is_tag_here < 0) {
		/* No requirements have been set here */
		NDB_del_msg_internal(m_id);
		//PRINT_MSG("Deleted 'cause there's no req");
		return -2;
	}

	if (is_tag_here == 0) {
		/*It might be that the tag was already deleted upon the reception of a new tag 
		 * or simply 'cause a new tag was added on this node for the first time and the one we're
		 * browsing is not present.*/
		//PRINT_MSG("This tag (0x%04x) was not here", m_id);
		return 0;
	}

	mesg_req_pp = &(temp_mesg_p->mesg.reqs);
	mesg_req_p = *mesg_req_pp;
	
	while (mesg_req_p != NULL) {
		if ( (mesg_req_p)->req.m_id == m_id ) {
			if ( (mesg_req_p)->link == NULL ) {
				/* I'm the last req tag */
				NDB_del_msg_internal(m_id);
				//PRINT_MSG("Deleted the whole var. It was the last req.");
				return 0;
			}
			
			/* Match the mac_prods to see if this tag has to be igmp dettached */
			if ( memcmp( (mesg_req_p)->req.m_mac_var , (mesg_req_p)->link->req.m_mac_var , 6 )  ) {
				/* If they are different, there might be the case that m_id is multicast */
				
				/* Handle the multicast anouncement */
				if ( (mesg_req_p)->req.m_direction == NDB_DIRECTION_CONSUMER ){
					S_igmp_queue_send_leave_if_multicast( (mesg_req_p)->req.m_mac_var, S_Get_My_mac() );
				}
			}
			
			/* remove m_id */
			NDB_mesg_req_empty_until_tagid_inclusive( mesg_req_pp, m_id );
			//PRINT_MSG("Deleted 'cause there's no req");

			return 0;
		}
		mesg_req_p = mesg_req_p->link;
	}
	
	ERROR_MSG("It should have found the m_id");
	return -3;
}

signed char NDB_del_msg( unsigned char m_type, unsigned short m_id)
{

	FTT_VAR_ID_SET_ID( m_id, m_id );
	if ( (m_type&0x0F) == 0 ){
		FTT_VAR_ID_SET_TYPE_SYNC( m_id );
	}else{
		FTT_VAR_ID_SET_TYPE_ASYNC( m_id );
	}
	
	return NDB_del_msg_internal(m_id);
}

signed char NDB_clean_msg_tag( unsigned char m_type, unsigned short m_id, unsigned char m_tag)
{
	
	FTT_VAR_ID_SET_ID( m_id, m_id );
	if ( (m_type&0x0F) == 0 ){
		FTT_VAR_ID_SET_TYPE_SYNC( m_id );
	}else{
		FTT_VAR_ID_SET_TYPE_ASYNC( m_id );
	}
	
	FTT_VAR_ID_SET_TAG( m_id, m_tag );
	
	return NDB_clean_msg_tag_internal(m_id);
}

signed char NDB_Init( void )
{

	/* NDB_mesgtable init */
	if ( NDB_mesgtable_pool_Init() < 0 ){
		ERROR_MSG("Failed initializing the requirements pool");
		return -1;
	}

	if ( NDB_table_pool_Init() < 0 ){
		ERROR_MSG("Failed initializing the ndb mesg slots pool");
		return -2;
	}
	
	/* Check if Allocator was previously initialized */
	if ( AllocatorInit_check() < 0){
		ERROR_MSG("Allocator was not previously initialized");
		return -3;
	}

#ifdef PRINT_MODULES_INIT
	PRINT_MSG(" -NDB_Init Ok");
#endif
	return 0;
}


signed char NDB_Close(void)
{
	while ( NDB_inuse_items != NULL ){
		NDB_del_msg_internal( NDB_inuse_items->mesg.tagless_id );
	}
	
	NDB_table_pool_Close();

	NDB_mesgtable_pool_Close();

#ifdef PRINT_MODULES_INIT
	PRINT_MSG(" -NDB_Close Ok");
#endif
	return 0;
}

////////////////////////////////////////////////////////////////////////////////////

signed char NDB_get_data2tx(
		unsigned short MesgID, 
		unsigned short frag_requested,
		unsigned char *mesg_value, 
		unsigned short *dm_len, 
		unsigned char *mesg_seqn,
		unsigned short *frag_seqn, 
#ifdef QNX_TAGGING
		QNX_TAGGING_TYPE *tagging,
#endif
		FTT_SIGNAL **callback_sig )
{
	NDB_tabtype *temp_p;
	
	NDB_mesgtype *aux_var; /* Pointer to particular message being produced */
	NDB_MESG_REQ *mesg_req;
	
	unsigned char is_sync = FTT_VAR_ID_IS_ID_SYNC(MesgID);
	int ret;

	signed char ret_val=0;	  /* Function exit code */

//	PRINT_MSG("NDB_get_data2tx called with MesgID: 0x%x    frag_requested:%d", MesgID, frag_requested );
	FTT_VAR_ID_SET_ID( MesgID, MesgID );
//PRINTF(" %d", frag_requested);
	/* Track it in the table */
	if( (temp_p = *NDB_table_track_tagless_id( &NDB_inuse_items, MesgID)) == NULL )
		return -1;

	aux_var = &(temp_p->mesg);
	
	/* Production test */
	if (!( aux_var->direction == NDB_DIRECTION_PRODUCER )){
		//ERROR_MSG("Sorry, not a producer of this var");
		return -2;
	}

	/* Avoid misconceptions */
	if ( aux_var->allocator < 0 ){
		ERROR_MSG("The message is being requested with no %s", (is_sync)?" CAB":"FIFO");
		return -3;
	}
	
#if 0
	/* I've received order to Tx a var with a given tag
	 *  - the older tags (same ID) have to be deleted
	 *  (remember that if the Master scheduled a new one, means the older
	 *  are no longer needed. Also remember that the procuder stopped accepting
	 *  data from the older tags) */

	if ( (ret=NDB_mesg_req_empty_until_tagid( &(aux_var->reqs), MesgID )) < 0){
		ERROR_MSG("There were no mesg_reqs before %x   ret:%d", MesgID, ret );
		return -4;
	}

#endif
	/* Track the requirements for this tag - it should be the fist one */
	if ( ( mesg_req = *NDB_mesg_req_track_tag_id ( &(aux_var->reqs), MesgID ) ) == NULL ){
		ERROR_MSG("Couldn't find the requirements for 0x%x", MesgID );
		return -5;
	}

	if ( frag_requested == 1 ){
		/* A first packet requested - start from the beggining */
		(aux_var->downside)[0].fragmentation_seq_number = 0;

		if ( (aux_var->downside)[0].allocator_downside_lock_p == NULL) {
			/* Get pointer for CAB/FIFO buffer and lock it */
			if (is_sync) (aux_var->downside)[0].allocator_downside_lock_p = cab_getmes( aux_var->allocator );
			else (aux_var->downside)[0].allocator_downside_lock_p = fifo_getmes( aux_var->allocator );
			
			if ( (aux_var->downside)[0].allocator_downside_lock_p == NULL) {
				ERROR_MSG("There are no messages to transmit");
				*dm_len = 0;
				return 0;
			}

			//PRINT_MSG("[NRDB.c] CAB/FIFO was oppened: %p", (aux_var->downside)[0].allocator_downside_lock_p);
		}else {
			PRINT_MSG("[NRDB.c] The CAB/FIFO was already oppened when received fragmentation trigger");
			/* Refresh to a new CAB/FIFO buffer */
			if (is_sync){
#ifdef QNX_TAGGING
				cab_unget( aux_var->allocator, (aux_var->downside)[0].allocator_downside_lock_p, tagging );
#else
				cab_unget( aux_var->allocator, (aux_var->downside)[0].allocator_downside_lock_p );
#endif
				(aux_var->downside)[0].allocator_downside_lock_p = cab_getmes( aux_var->allocator );
			}else{
#ifdef QNX_TAGGING
				fifo_unget( aux_var->allocator, (aux_var->downside)[0].allocator_downside_lock_p, tagging );
#else
				fifo_unget( aux_var->allocator, (aux_var->downside)[0].allocator_downside_lock_p );
#endif
				(aux_var->downside)[0].allocator_downside_lock_p = fifo_getmes( aux_var->allocator );
			}
		}
	}
	
	if ( (aux_var->downside)[0].allocator_downside_lock_p == NULL ) {
		PRINT_MSG("The %s was not properly requested", (is_sync)?" CAB":"FIFO");
		//if (!is_sync) PRINT_MSG("The FIFO queue: %d", fifo_size( aux_var->allocator) );
		//The previous was commented due to segfault coming from the fifo_size being called after a var is deleted
		return -6;
	}

	if( (frag_requested - 1) != (aux_var->downside)[0].fragmentation_seq_number ){
		ERROR_MSG("The req fragment (%d) not in the expected order(%d)", frag_requested - 1, (aux_var->downside)[0].fragmentation_seq_number );
		(aux_var->downside)[0].fragmentation_seq_number = frag_requested - 1;
	}

	/* Check if it's the last packet */
	if ((aux_var->downside)[0].fragmentation_seq_number == ( mesg_req->comp.packets_no-1)){
		/* This is the last paket - or first if single */

		/* Copy data from CAB/FIFO buffer */
		memcpy( mesg_value, 
				(aux_var->downside)[0].allocator_downside_lock_p + ((aux_var->downside)[0].fragmentation_seq_number * mesg_req->comp.d_size.One_Bsize), 
				mesg_req->comp.d_size.Last_Bsize_real);

		/* Release CAB/FIFO buffer */
		ret_val = (aux_var->status & NDB_TX_NEW_DAT)?1:0;

#ifdef QNX_TAGGING
		if (is_sync) ret = cab_unget( aux_var->allocator, (aux_var->downside)[0].allocator_downside_lock_p, tagging);
		else ret = fifo_unget( aux_var->allocator, (aux_var->downside)[0].allocator_downside_lock_p, tagging);
#else
		if (is_sync) ret = cab_unget( aux_var->allocator, (aux_var->downside)[0].allocator_downside_lock_p);
		else ret = fifo_unget( aux_var->allocator, (aux_var->downside)[0].allocator_downside_lock_p);
#endif
		
		/* Reset the NDB_TX_NEW_DAT flag - in case there's nothing left in the FIFO */
		if (ret == 0) aux_var->status &= ~NDB_TX_NEW_DAT;

		(aux_var->downside)[0].allocator_downside_lock_p = NULL;

		/* Fill function arguments */
		*dm_len = mesg_req->comp.d_size.Last_Bsize;
//		*mesg_seqn = (unsigned char)aux_var->m_ntx;
		*frag_seqn = (aux_var->downside)[0].fragmentation_seq_number + 1;
		*callback_sig = &(aux_var->callback_signal);

	}else {
		/* Just one more packet */
		
		/* Copy data from CAB buffer */
		memcpy( mesg_value, 
				(aux_var->downside)[0].allocator_downside_lock_p + (((aux_var->downside)[0].fragmentation_seq_number) * (mesg_req->comp.d_size.One_Bsize)),
				mesg_req->comp.d_size.One_Bsize);
		
		/* Fill function arguments */
		*dm_len = mesg_req->comp.d_size.One_Bsize;
//		*mesg_seqn = (unsigned char)aux_var->m_ntx;
		*frag_seqn = (aux_var->downside)[0].fragmentation_seq_number + 1;
		*callback_sig = NULL;

		ret_val = 0;
	}

	(aux_var->downside)[0].fragmentation_seq_number ++;

	return (ret_val);
}


signed char NDB_set_data2rx(
		unsigned short MesgID,
		unsigned char  source_nodeID,
		unsigned short frag_requested,
		unsigned char *mesg_value, 
		unsigned char mesg_tx_new_flag, 
#ifdef QNX_TAGGING
		QNX_TAGGING_TYPE tagging,
#endif 
		struct rtl_timespec rxtimestamp)
{
	NDB_tabtype *temp_p;
	
	NDB_mesgtype *aux_var; /* Pointer to particular message being received */
	NDB_MESG_REQ *mesg_req;

	unsigned char is_sync = FTT_VAR_ID_IS_ID_SYNC(MesgID);
	
	unsigned char idx;

	//PRINT_MSG("NDB_set_data2rx MesgID: 0x%04x  frag:%d src:%d", MesgID, frag_requested, source_nodeID );
	
	FTT_VAR_ID_SET_ID( MesgID, MesgID );

	if( (temp_p = *NDB_table_track_tagless_id( &NDB_inuse_items, MesgID)) == NULL )
		return -1;

	aux_var = &(temp_p->mesg);
	
	/* Consumption test */
	if (!( aux_var->direction == NDB_DIRECTION_CONSUMER )){
		//ERROR_MSG("Sorry, not a consumer of this var");
		return -2;
	}
	
	/* Avoid misconceptions */
	if ( aux_var->allocator < 0 ){
		ERROR_MSG("The message is being requested with no %s", (is_sync)?" CAB":"FIFO");
		return -3;
	}

#if 0 /*Remove when adding multiple sources. Problems when receiving on a tag that had been deleted.*/
	
	/* I've received order to Tx a var with a given tag
	 *  - the older tags (same ID) have to be deleted
	 *  (remember that if the Master scheduled a new one, means the older
	 *  are no longer needed. Also remember that the procuder stopped accepting
	 *  data from the older tags) */
	{	
	int ret;
	if ( (ret=NDB_mesg_req_empty_until_tagid( &(aux_var->reqs), MesgID )) < 0){
		ERROR_MSG("There were no mesg_reqs before %x   ret:%d", MesgID, ret );
		return -4;
	}
	}

#endif

	/* Track the requirements for this tag - it should be the fist one */
	if ( ( mesg_req = *NDB_mesg_req_track_tag_id ( &(aux_var->reqs), MesgID ) ) == NULL ){
		ERROR_MSG("Couldn't find the requirements for 0x%x", MesgID );
		return -5;
	}
	
	/* Get the idx refering to the proper source allocating buffer */
	if ( NDB_mesgtype_downside_get_idx_and_tag_it( aux_var, source_nodeID, &idx ) < 0){
		ERROR_MSG("All sources slots were occupied... check if you have more than %d producers for this var", MAX_POSSIBLE_PRODUCERS);
		return -7;
	}
	
	if ( frag_requested == 1 ){
		/* A first packet requested - start from the beggining */

		(aux_var->downside)[idx].fragmentation_seq_number = 0;

		if ((aux_var->downside)[idx].allocator_downside_lock_p == NULL) {

			/* Get pointer for CAB/FIFO buffer and lock it */
			if (is_sync) (aux_var->downside)[idx].allocator_downside_lock_p = cab_reserve( aux_var->allocator );
			else (aux_var->downside)[idx].allocator_downside_lock_p = fifo_reserve( aux_var->allocator, 0 /*no-block*/ );
			//PRINT_MSG("%s was oppened: %p",(is_sync)?" CAB":"FIFO", (aux_var->downside)[idx].allocator_downside_lock_p);
			if ( (aux_var->downside)[idx].allocator_downside_lock_p == NULL ) {
				PRINT_MSG("The %s was FULL - Message lost!", (is_sync)?" CAB":"FIFO");
				return -8;
			}
		}else {
			PRINT_MSG("Received Frag:1 but CAB/FIFO was already open");
		}
	}
	
	if ( (aux_var->downside)[idx].allocator_downside_lock_p == NULL ) {
		PRINT_MSG("Rx: The %s was not properly requested. Probably lost frag no. 1.", (is_sync)?" CAB":"FIFO");
		return -6;
	}

	if( (frag_requested - 1) != (aux_var->downside)[idx].fragmentation_seq_number ){
		ERROR_MSG("The requested fragment is not in the expected order %d", (aux_var->downside)[idx].fragmentation_seq_number);
		(aux_var->downside)[idx].fragmentation_seq_number = frag_requested - 1;
	}

	/* Check if it's the last packet */
	if ((aux_var->downside)[idx].fragmentation_seq_number == (mesg_req->comp.packets_no-1)){
		/* This is the last paket - or first if single */

		/* Copy data to the buffer */
		/* Luis Silva detected the short message bug - using Last_Bsize_real*/
		memcpy(	(aux_var->downside)[idx].allocator_downside_lock_p + ((aux_var->downside)[idx].fragmentation_seq_number * mesg_req->comp.d_size.One_Bsize),
				mesg_value, 
				mesg_req->comp.d_size.Last_Bsize_real);

		/* Release CAB/FIFO buffer */
#ifdef QNX_TAGGING
		if (is_sync) cab_putmes( aux_var->allocator, (aux_var->downside)[idx].allocator_downside_lock_p, tagging);
		else fifo_putmes( aux_var->allocator, (aux_var->downside)[idx].allocator_downside_lock_p, tagging);
#else
		if (is_sync) cab_putmes( aux_var->allocator, (aux_var->downside)[idx].allocator_downside_lock_p);
		else fifo_putmes( aux_var->allocator, (aux_var->downside)[idx].allocator_downside_lock_p);
#endif
		(aux_var->downside)[idx].allocator_downside_lock_p = NULL;
		NDB_mesgtype_downside_free_idx( aux_var, idx );
		
		(aux_var->status) |= NDB_RX_NEW_DAT;
		if ( mesg_tx_new_flag  )
			(aux_var->status) |= NDB_TX_NEW_DAT ;
		else
			(aux_var->status) &= ~NDB_TX_NEW_DAT;

		FTT_signal( &(aux_var->callback_signal) );

	}else {
		/* Just one more packet */

		/* Copy data from CAB buffer */
		memcpy(	(aux_var->downside)[idx].allocator_downside_lock_p + (((aux_var->downside)[idx].fragmentation_seq_number) * (mesg_req->comp.d_size.One_Bsize)),
				mesg_value, 
				mesg_req->comp.d_size.One_Bsize);
	}

	(aux_var->downside)[idx].fragmentation_seq_number ++;

	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////

signed char NDB_test_produce_and_bind(unsigned short MesgID, unsigned char **dest_mac)
{
	NDB_tabtype *temp_p;
	NDB_MESG_REQ *temp_mesg_p;
	
	/* Track it in the table */
	if( (temp_p = *NDB_table_track_tagless_id( &NDB_inuse_items, MesgID )) == NULL )
		return -1;

	/* Track the requirements for this tag */
	if ( ( temp_mesg_p = *NDB_mesg_req_track_tag_id ( &(temp_p->mesg.reqs), MesgID ) ) == NULL ){
		//ERROR_MSG("Couldn't find the requirements for 0x%x", MesgID );
		//The error message is commented because this really could happan as a feature!
		//E.g. a node being added as an additional consumer gets a new tag. While the Master still schedules the old tag, the 
		//dispatcher doesn't find the requirements for this variable after passing the track_tagless_id.
		return -2;
	}

	/* Production test */
	if ( !( temp_p->mesg.direction == NDB_DIRECTION_PRODUCER ) || (temp_p->mesg.allocator < 0) )
		return -3;
	
	if (dest_mac != NULL) 
		*dest_mac = temp_mesg_p->req.m_mac_var;
	return 0;
}

signed char NDB_test_consume_and_bind(unsigned short MesgID)
{
	NDB_tabtype *temp_p;
	
	/* Track it in the table */
	if( (temp_p = *NDB_table_track_tagless_id( &NDB_inuse_items, MesgID)) == NULL )
		return -1;

	/* Consumption test */
	if ( !( temp_p->mesg.direction == NDB_DIRECTION_CONSUMER ) || (temp_p->mesg.allocator < 0) )
		return -2;

	return 0;
}

signed char NDB_test_consumer_get_mac( unsigned char m_type, unsigned short m_id, unsigned char *mac)
{
	NDB_tabtype *temp_p;
	NDB_MESG_REQ *temp_mesg_p;

	FTT_VAR_ID_SET_ID( m_id, m_id );
	if ( (m_type&0x0F) == 0 ){
		FTT_VAR_ID_SET_TYPE_SYNC( m_id );
	}else{
		FTT_VAR_ID_SET_TYPE_ASYNC( m_id );
	}

	/* Track message in table */
	if( (temp_p = *NDB_table_track_tagless_id( &NDB_inuse_items, m_id)) == NULL )
		return -1;

	/* Track the first requirements for this */
	if ( (temp_mesg_p = temp_p->mesg.reqs) == NULL ){
		ERROR_MSG("Couldn't find the requirements - It's empty :<(" );
		return -2;
	}

	/* Check for valid status */
	if ( !( temp_p->mesg.direction == NDB_DIRECTION_CONSUMER ) )
		return -3;

	if ( mac != NULL )
		ETH_L_CopyMAC_to_from( mac, temp_mesg_p->req.m_mac_var );
	
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////

static signed char NDB_m_pre_produce (unsigned short MesgID, unsigned char alloc_block, unsigned char **return_p)
{
	NDB_tabtype *temp_p;
	NDB_mesgtype *aux_var;    /* Pointer to particular message being produced */

	unsigned char is_sync = FTT_VAR_ID_IS_ID_SYNC(MesgID);
	
	*return_p = NULL;

	/* Track message in table */
	if( (temp_p = *NDB_table_track_tagless_id( &NDB_inuse_items, MesgID)) == NULL ){
		ERROR_MSG("Couldn't find the Message ID");
		return -1;
	}

	aux_var = &(temp_p->mesg);

	/* Checks if station is a sender of it */
	if (!( aux_var->direction == NDB_DIRECTION_PRODUCER)){
		/* Station is not a producer of this var */
		ERROR_MSG("Message not producer of this var!");
		return -2;
	}

	/* Check if message was bound to the application - memory&signal */
	if (aux_var->allocator<0){
		return -3;
	}

	if (aux_var->allocator_upside_lock_p != NULL){
		ERROR_MSG(" Careful!!!! A pre-send/produce was previously issued");
		*return_p = aux_var->allocator_upside_lock_p;
		return 1;
	}

	if (is_sync) aux_var->allocator_upside_lock_p = cab_reserve(aux_var->allocator);
	else (aux_var->allocator_upside_lock_p = fifo_reserve(aux_var->allocator, alloc_block));

	if (aux_var->allocator_upside_lock_p == NULL){
		if (!alloc_block) {
			//ERROR_MSG(" %s is full - no available slots", (is_sync)?" CAB":"FIFO" );
			return -7;
		}
		return -10; //semaphore destroyed while waiting
	}

	*return_p = aux_var->allocator_upside_lock_p;
	return 0;
}

static signed char NDB_m_pre_produce_abort(unsigned short MesgID)
{
	int ret;
	NDB_tabtype *temp_p;
	NDB_mesgtype *aux_var;

	unsigned char is_sync = FTT_VAR_ID_IS_ID_SYNC(MesgID);

	/* Check if message exists in table */
	if( (temp_p = *NDB_table_track_tagless_id( &NDB_inuse_items, MesgID)) == NULL )
		return -1;

	aux_var = &(temp_p->mesg);

	/* Checks if station is a sender of it */
	if (!( aux_var->direction == NDB_DIRECTION_PRODUCER)){
		/* Station is not a producer of this var */
		ERROR_MSG("Message not producer of this var!");
		return -2;
	}
	/* Check if message was bound to the application - memory&signal */
	if (aux_var->allocator<0)
		return -3;

	if (aux_var->allocator_upside_lock_p == NULL) /* Nothing to abort */
		return 0;
	
	if (is_sync) ret = cab_reserve_abort(aux_var->allocator, aux_var->allocator_upside_lock_p );
	else ret = fifo_reserve_abort(aux_var->allocator, aux_var->allocator_upside_lock_p );
	if (ret < 0){
		ERROR_MSG("(cab/fifo)_reserve_abort returned: %d", ret);
		return -4;
	}
	/* Free the locked pointer */
	aux_var->allocator_upside_lock_p = NULL;

	return 0;
}

static signed char NDB_m_produce(unsigned short MesgID, unsigned char *mesg_value, unsigned char alloc_block)
{
	int ret;
	NDB_tabtype *temp_p;
	NDB_mesgtype *aux_var;    /* Pointer to particular message being produced */
	NDB_MESG_REQ *mesg_req;

	unsigned char is_sync = FTT_VAR_ID_IS_ID_SYNC(MesgID);

//PRINT_MSG("Sending order for %x<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<", MesgID);
	/* Check if message exists in table */
	if( (temp_p = *NDB_table_track_tagless_id( &NDB_inuse_items, MesgID)) == NULL ){
		return -1;
	}

	aux_var = &(temp_p->mesg);

	/* Checks if station is a sender of it */
	if (!((aux_var->direction) == NDB_DIRECTION_PRODUCER)){
		/* Station is not a producer of this var */
		ERROR_MSG("Message not producer of this var!");
		return -2;
	}
	/* Check if message was bound to the application - memory&signal */
	if (aux_var->allocator<0){
		return -3;
	}

//NDB_table_print();
	/* -We shall produce only in the most recent tag - If a tag older than the most recent
	 * is in production we shall return an error 
	 * -Let's then get the most recent requirements and compare its tag with the one supplied */

	if ( ( mesg_req = NDB_mesg_req_goto_last_item_p( aux_var->reqs ) ) == NULL ){
		ERROR_MSG("No requirements available in this table slot: %x", MesgID );
		return -4;
	}

	if ( !FTT_VAR_ID_CMP_ID_TYPE_TAG( mesg_req->req.m_id, MesgID ) ){
#ifdef SLAVE_SEND_TAG_MISMATCH_WARN
		ERROR_MSG("Trying to produce/send to an old tagged id: the new 0x%x  trying 0x%x",mesg_req->req.m_id, MesgID );
#endif
		return -5; //Do not change this value so easilly, it's being used in the app
	}

	/* At this point mesg_req points to the requirements of the wanted tag */

	if (aux_var->allocator_upside_lock_p == NULL){
		/* Let's reserve the buffer */
		if (mesg_value == NULL){
			ERROR_MSG(" Bad input pointer ");
			return -6;
		}

		if (is_sync) aux_var->allocator_upside_lock_p = cab_reserve(aux_var->allocator);
		else (aux_var->allocator_upside_lock_p = fifo_reserve(aux_var->allocator, alloc_block));

		if (aux_var->allocator_upside_lock_p == NULL){
			if (!alloc_block){
				ERROR_MSG(" %s is full", (is_sync)?" CAB":"FIFO" );
				return -7;
			}
			return -10; //semaphore destroyed willink waiting
		}

		memcpy(
				(unsigned char *)(aux_var->allocator_upside_lock_p),
				mesg_value,
				mesg_req->req.m_size );
	}

#ifdef QNX_TAGGING
#warning Place tag here
	if (is_sync) ret = cab_putmes(aux_var->allocator, aux_var->allocator_upside_lock_p, 0x5555/*place_tag_here*/ );
	else ret = fifo_putmes(aux_var->allocator, aux_var->allocator_upside_lock_p, 0x5555/*place_tag_here*/);
#else
	if (is_sync) ret = cab_putmes(aux_var->allocator, aux_var->allocator_upside_lock_p);
	else ret = fifo_putmes(aux_var->allocator, aux_var->allocator_upside_lock_p);
#endif
	if ( ret < 0){
		ERROR_MSG ("%s returned: %d", (is_sync)?"cab_putmes":"fifo_putmes",ret);
		return(-8);
	}

	if (!is_sync){ 
		(mesg_req->req.am_queue_notify) ++;
//		PRINT_MSG("UUUUUUUUUUUUUUUUUUUUUUUUUUUPdating production queue with tag %x",mesg_req->req.m_id );
	}

	/* Free the locked pointer */
	aux_var->allocator_upside_lock_p = NULL;

	/* Set the NDB_TX_NEW_DAT bit */
	(aux_var->status) |= NDB_TX_NEW_DAT;

	return 0;
}

static signed char NDB_m_consume_status (unsigned short MesgID )
{
	NDB_tabtype *temp_p;
	NDB_mesgtype *aux_var;    /* Pointer to particular message being produced */
	unsigned char ret_val = 0;
	unsigned char is_sync = FTT_VAR_ID_IS_ID_SYNC(MesgID);

	/* Check if message exists in table */
	if( (temp_p = *NDB_table_track_tagless_id( &NDB_inuse_items, MesgID)) == NULL )
		return -1;
	
	aux_var = &(temp_p->mesg);


	/* Checks if station is a consumer of it */
	if (!( aux_var->direction == NDB_DIRECTION_CONSUMER)){
		/* Station is not a consumer of this var */
		ERROR_MSG("Message not consumer of this var!");
		return -2;
	}

	if (is_sync){
		/* Save previous NDB_RX_NEW_DAT and NDB_TX_NEW_DAT bits */
		if ( aux_var->status & NDB_RX_NEW_DAT ){
			if (aux_var->status & NDB_TX_NEW_DAT) ret_val = 2;
			else ret_val = 1;
		}
	}else{
		ret_val = fifo_is_loaded( aux_var->allocator );
	}

	return ret_val;
}

static unsigned char *NDB_m_pre_consume (unsigned short MesgID, unsigned char *ret_tag, unsigned char *ret_status )
{

	NDB_tabtype *temp_p;
	NDB_mesgtype *aux_var;    /* Pointer to particular message being produced */

	unsigned char is_sync = FTT_VAR_ID_IS_ID_SYNC(MesgID);
	unsigned char ret_val = 0;

	/* Check if message exists in table */
	/* Track message in table */
	if( (temp_p = *NDB_table_track_tagless_id( &NDB_inuse_items, MesgID)) == NULL )
		return NULL;
	

	aux_var = &(temp_p->mesg);

	/* Checks if station is a consumer of it */
	if (!( aux_var->direction == NDB_DIRECTION_CONSUMER)){
		/* Station is not a consumer of this var */
		ERROR_MSG("Message not consumer of this var!");
		return NULL;
	}

	/* Check if message was bound to the application - memory&signal */
	if (aux_var->allocator<0)
		return NULL;

	if (aux_var->allocator_upside_lock_p != NULL){
		ERROR_MSG(" Careful!!!! A pointer was already locking");
		return aux_var->allocator_upside_lock_p;
	}

	if ( aux_var->status & NDB_RX_NEW_DAT ){
		if (aux_var->status & NDB_TX_NEW_DAT) ret_val = 2;
		else ret_val = 1;
	}
	
	if (ret_status != NULL) *ret_status = ret_val;

	if (ret_tag != NULL)
		*ret_tag = FTT_VAR_ID_GET_TAG( aux_var->reqs->req.m_id );

	if (is_sync) return (aux_var->allocator_upside_lock_p = cab_getmes(aux_var->allocator));
	else return (aux_var->allocator_upside_lock_p = fifo_getmes(aux_var->allocator));

}

static signed char NDB_m_consume(unsigned short MesgID, unsigned char *ret_tag, unsigned char *mesg_value, unsigned char *ret_status )
{
	int ret;
	NDB_tabtype *temp_p;
	NDB_mesgtype *aux_var;    /* Pointer to particular message being produced */
	NDB_MESG_REQ *mesg_req;

	unsigned char is_sync = FTT_VAR_ID_IS_ID_SYNC(MesgID);
	unsigned char ret_val = 0;

	/* Check if message exists in table */
	if( (temp_p = *NDB_table_track_tagless_id( &NDB_inuse_items, MesgID)) == NULL )
		return -1;
	
	aux_var = &(temp_p->mesg);


	/* Checks if station is a consumer of it */
	if (!( aux_var->direction == NDB_DIRECTION_CONSUMER)){
		/* Station is not a consumer of this var */
		ERROR_MSG("Message not consumer of this var!");
		return -2;
	}
	
	/* Check if message was bound to the application - memory&signal */
	if (aux_var->allocator<0)
		return -3;

//	/* Track the requirements for this tag */
//	if ( ( mesg_req = *NDB_mesg_req_track_tag_id ( &(aux_var->reqs), MesgID ) ) == NULL ){
//		ERROR_MSG("Couldn't find the requirements for 0x%x", MesgID );
//		return -4;
//	}
	mesg_req = aux_var->reqs;

	if (aux_var->allocator_upside_lock_p == NULL){
		/* Let's reserve the buffer */
		if (is_sync) aux_var->allocator_upside_lock_p = cab_getmes(aux_var->allocator);
		else aux_var->allocator_upside_lock_p = fifo_getmes(aux_var->allocator);

		if (aux_var->allocator_upside_lock_p == NULL){
			//ERROR_MSG(" %s is empty", (is_sync)?" CAB":"FIFO");
			return -5;
		}
		if (mesg_value == NULL){
			ERROR_MSG(" Bad input pointer ");
			return -6;
		}
		memcpy(
				mesg_value,
				(unsigned char *)(aux_var->allocator_upside_lock_p),
				mesg_req->req.m_size );
	}

	/* Save previous NDB_RX_NEW_DAT and NDB_TX_NEW_DAT bits */
	if ( aux_var->status & NDB_RX_NEW_DAT ){
		if (aux_var->status & NDB_TX_NEW_DAT) ret_val = 2;
		else ret_val = 1;
	}

#ifdef QNX_TAGGING
	{
		QNX_TAGGING_TYPE tagging;
		if (is_sync) ret = cab_unget(aux_var->allocator, aux_var->allocator_upside_lock_p, &tagging);
		else ret = fifo_unget(aux_var->allocator, aux_var->allocator_upside_lock_p, &tagging);
#warning Do sth with tag
	}
#else
	if (is_sync) ret = cab_unget(aux_var->allocator, aux_var->allocator_upside_lock_p);
	else ret = fifo_unget(aux_var->allocator, aux_var->allocator_upside_lock_p);
#endif

	if ( ret  < 0){
		ERROR_MSG ("%s returned: %d", (is_sync)?"cab_unget":"fifo_unget", ret);
		return(-7);
	}
	
	/* Reset the NDB_RX_NEW_DAT bit if the FIFO got empty */
	if (ret == 0) (aux_var->status) &= ~NDB_RX_NEW_DAT;

	/* Free the locked pointer */
	aux_var->allocator_upside_lock_p = NULL;
	
	if (ret_status != NULL) *ret_status = ret_val;
	return (0);
}

signed char NDB_pre_produce(unsigned short MesgID, unsigned char **return_p)
{
	FTT_VAR_ID_SET_ID( MesgID, MesgID );
	FTT_VAR_ID_SET_TYPE_SYNC( MesgID );
	return ( NDB_m_pre_produce( MesgID, 0 /*no alloc_block*/, return_p ) );
}

signed char NDB_pre_produce_abort(unsigned short MesgID)
{
	FTT_VAR_ID_SET_ID( MesgID, MesgID );
	FTT_VAR_ID_SET_TYPE_SYNC( MesgID );
	return ( NDB_m_pre_produce_abort( MesgID ) );
}

signed char NDB_produce(unsigned short MesgID, unsigned char tag, void *mesg_value)
{
	FTT_VAR_ID_SET_ID( MesgID, MesgID );
	FTT_VAR_ID_SET_TYPE_SYNC( MesgID );
	FTT_VAR_ID_SET_TAG( MesgID, tag );
	return ( NDB_m_produce( MesgID, (unsigned char *)mesg_value, 0 /*no alloc_block*/ ) );
}

signed char NDB_consume_status( unsigned short MesgID )
{
	FTT_VAR_ID_SET_ID( MesgID, MesgID );
	FTT_VAR_ID_SET_TYPE_SYNC( MesgID );
	return ( NDB_m_consume_status( MesgID ) );
}

unsigned char *NDB_pre_consume(unsigned short MesgID, unsigned char *ret_tag, unsigned char *ret_status )
{
	FTT_VAR_ID_SET_ID( MesgID, MesgID );
	FTT_VAR_ID_SET_TYPE_SYNC( MesgID );
	return ( NDB_m_pre_consume( MesgID, ret_tag, ret_status ) );
}

signed char NDB_consume(unsigned short MesgID, unsigned char *ret_tag, unsigned char *mesg_value, unsigned char *ret_status )
{
	FTT_VAR_ID_SET_ID( MesgID, MesgID );
	FTT_VAR_ID_SET_TYPE_SYNC( MesgID );
	return ( NDB_m_consume( MesgID, ret_tag, mesg_value, ret_status ) );
}

signed char NDB_pre_send(unsigned short MesgID, unsigned char alloc_block, unsigned char **return_p)
{
	FTT_VAR_ID_SET_ID( MesgID, MesgID );
	FTT_VAR_ID_SET_TYPE_ASYNC( MesgID );
	return ( NDB_m_pre_produce( MesgID, alloc_block, return_p ) );
}

signed char NDB_pre_send_abort(unsigned short MesgID)
{
	FTT_VAR_ID_SET_ID( MesgID, MesgID );
	FTT_VAR_ID_SET_TYPE_ASYNC( MesgID );
	return ( NDB_m_pre_produce_abort( MesgID ) );
}

signed char NDB_send(unsigned short MesgID, unsigned char tag, void *mesg_value, unsigned char alloc_block)
{
	FTT_VAR_ID_SET_ID( MesgID, MesgID );
	FTT_VAR_ID_SET_TYPE_ASYNC( MesgID );
	FTT_VAR_ID_SET_TAG( MesgID, tag );
	return ( NDB_m_produce( MesgID, (unsigned char *)mesg_value, alloc_block ) );
}

signed char NDB_receive_status(unsigned short MesgID )
{
	FTT_VAR_ID_SET_ID( MesgID, MesgID );
	FTT_VAR_ID_SET_TYPE_ASYNC( MesgID );
	return ( NDB_m_consume_status( MesgID ) );
}

unsigned char *NDB_pre_receive(unsigned short MesgID, unsigned char *ret_tag, unsigned char *ret_status )
{
	FTT_VAR_ID_SET_ID( MesgID, MesgID );
	FTT_VAR_ID_SET_TYPE_ASYNC( MesgID );
	return ( NDB_m_pre_consume( MesgID, ret_tag, ret_status ) );
}

signed char NDB_receive(unsigned short MesgID, unsigned char *ret_tag, unsigned char *mesg_value, unsigned char *ret_status )
{
	FTT_VAR_ID_SET_ID( MesgID, MesgID );
	FTT_VAR_ID_SET_TYPE_ASYNC( MesgID );
	return ( NDB_m_consume( MesgID, ret_tag, mesg_value, ret_status ) );
}

#ifdef NODE_ASYNC_REPLY


void NDB_AM_Get_sent(NDB_AM_Get_sent_RETURN_TYPE *p, unsigned short *size)
{
	unsigned short max_size = *size;
	NDB_tabtype *temp_table_p;
	NDB_MESG_REQ *temp_mesg_p;

	temp_table_p = NDB_inuse_items;
	*size = 0;

	while(temp_table_p != NULL){
		temp_mesg_p = temp_table_p->mesg.reqs;
		while(temp_mesg_p != NULL){
			if ( FTT_VAR_ID_IS_ID_ASYNC(temp_mesg_p->req.m_id) )
				if ( temp_mesg_p->req.m_direction == NDB_DIRECTION_PRODUCER ){
					/* append this item */
					if ( temp_mesg_p->req.am_queue_notify )
					if (*size == max_size) return;
					p[*size].tagged_id = temp_mesg_p->req.m_id;
					if ( (p[*size].num = temp_mesg_p->req.am_queue_notify) ){
						temp_mesg_p->req.am_queue_notify = 0;
						(*size) ++;
					}
					//p[*size].num = temp_mesg_p->req.am_queue_notify;
				}
			temp_mesg_p = temp_mesg_p->link;
		}

		temp_table_p = temp_table_p->link;
	}
	return ;
}

#endif




