/*****************************************************************************
 * S_igmp_queue.c:
 *****************************************************************************
 * Copyright (C) 2006-2012 the FTT-SE team.
 *
 * Author: Ricardo Marau <marau at fe.up.pt>
 *
 * FTT-SE is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  FTT-SE is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with FTT-SE.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

#include "Ports.h"

#include "S_allocator.h"
#include "S_igmp_queue.h"
#include <string.h>

//For the last functions
#include "igmp.h"
#include "ethernet.layer.h"


static ALLOC_ID igmp_queue;

signed char S_igmp_queue_init(void)
{

	/* Check if Allocator was previously initialized */
	if ( AllocatorInit_check() < 0){
		ERROR_MSG("Allocator was not previously initialized");
		return -1;
	}

	igmp_queue = fifo_create( 2+1500, 10, NULL, 0 /*mem_size - pointless since memory is FTT created*/);

	if (igmp_queue<0){
		ERROR_MSG("fifo_create ret: %d",igmp_queue);
		return -1;
	}

#ifdef PRINT_MODULES_INIT
	PRINT_MSG(" -S_igmp_queue_init Ok");
#endif
	return 0;
}

void S_igmp_queue_close(void)
{
	S_igmp_queue_flush_queue();
	
	fifo_delete(igmp_queue);
	
#ifdef PRINT_MODULES_INIT
	PRINT_MSG(" -S_igmp_queue_close Ok");
#endif
	return;
}

signed char S_igmp_queue_put_message( unsigned char *msg, unsigned short len )
{
	void *temp_p;

//#warning check if I should make this blocking	
	
	if ( (temp_p = fifo_reserve( igmp_queue, 0 /*non-block*/ )) == NULL) {
		ERROR_MSG("Couldn't allocate room to senf IGMP message");
		return -1;
	}

	*((unsigned short *)temp_p) = len;
	memcpy(	temp_p + 2 , msg, len);

#ifdef QNX_TAGGING
	fifo_putmes( igmp_queue, temp_p, 0);
#else
	fifo_putmes( igmp_queue, temp_p);
#endif
	return 0;
	
}

signed char S_igmp_queue_get_message(unsigned char *mesg_rcv, unsigned short *len)
{
	void *temp_p;
	if ((temp_p = fifo_getmes( igmp_queue )) == NULL)
		return -1;

	*len = *((unsigned short *)temp_p);
	memcpy(	mesg_rcv, temp_p + 2, *len);

#ifdef QNX_TAGGING
	fifo_unget( igmp_queue, temp_p, NULL);
#else
	fifo_unget( igmp_queue, temp_p);
#endif
	return 0;
}


//////////////////////////////////////////////
//The following functions are macros to form the messages using the above. 
signed char S_igmp_queue_send_report_if_multicast( unsigned char *mesg_mac, unsigned char *my_mac )
{
	unsigned char igmp_compose[1514];
	IP_ADDR ip_add;
	
	/* Check if the message is multicated */
	if (IGMP_isMulticast(mesg_mac)){
		ip_add.ad[0] = 224;
		ip_add.ad[1] = mesg_mac[3];
		ip_add.ad[2] = mesg_mac[4];
		ip_add.ad[3] = mesg_mac[5];
//		PRINT_MSG("Sending add IP %d.%d.%d.%d",ip_add.ad[0],ip_add.ad[1],ip_add.ad[2],ip_add.ad[3] );
		
		IGMP_set_ip( igmp_compose, IGMP_sizeof(), ip_add, IGMP_PROTOCOL );
		IGMP_assign_igmp_cmd(igmp_compose, &ip_add);
		
		ETH_L_CopyMAC_to_from( ETH_L_Dest_addr(igmp_compose), mesg_mac );
		ETH_L_CopyMAC_to_from( ETH_L_Source_addr(igmp_compose), my_mac );
		*((unsigned short *)(ETH_L_Type_addr(igmp_compose))) = htons ( 0x0800 );
		
		S_igmp_queue_put_message( igmp_compose, IGMP_frameSize() );

//		PRINT_MSG("IGMP Join for:");
//		PRINT_MSG(" %2x %2x %2x %2x %2x %2x \n", (mesg_mac)[0], (mesg_mac)[1], (mesg_mac)[2], (mesg_mac)[3], (mesg_mac)[4], (mesg_mac)[5] );
//		PRINT_MSG("IP %d.%d.%d.%d",ip_add.ad[0],ip_add.ad[1],ip_add.ad[2],ip_add.ad[3] );
	}

	return 0;
}

signed char S_igmp_queue_send_leave_if_multicast( unsigned char *mesg_mac, unsigned char *my_mac )
{
	unsigned char igmp_compose[1514];
	IP_ADDR ip_add;

	if (IGMP_isMulticast(mesg_mac)){
		ip_add.ad[0] = 224;
		ip_add.ad[1] = mesg_mac[3];
		ip_add.ad[2] = mesg_mac[4];
		ip_add.ad[3] = mesg_mac[5];
		IGMP_set_ip( igmp_compose, IGMP_sizeof(), ip_add, IGMP_PROTOCOL );
		IGMP_deassign_igmp_cmd(igmp_compose, &ip_add);
		
		ETH_L_CopyMAC_to_from( ETH_L_Dest_addr(igmp_compose), mesg_mac );
		ETH_L_CopyMAC_to_from( ETH_L_Source_addr(igmp_compose), my_mac );

		*((unsigned short *)(ETH_L_Type_addr(igmp_compose))) = htons ( 0x0800 );
		
		if ( S_igmp_queue_put_message( igmp_compose, IGMP_frameSize() ) < 0 )
			return -1;

//		PRINT_MSG("IGMP Leave for:");
//		PRINT_MSG(" %2x %2x %2x %2x %2x %2x \n", (mesg_mac)[0], (mesg_mac)[1], (mesg_mac)[2], (mesg_mac)[3], (mesg_mac)[4], (mesg_mac)[5] );
//		PRINT_MSG("IP %d.%d.%d.%d",ip_add.ad[0],ip_add.ad[1],ip_add.ad[2],ip_add.ad[3] );
	
	}
		
	return 0;
}


void S_igmp_queue_flush_queue(void)
{
	unsigned char *tmsg;
	unsigned short mesg_efflen;    /* Real data length of message (in bytes) */
	int ret;
	
	while (1)
	{
		if ( ETH_L_tx_reserve_buffer( &tmsg ) < 0 ){
			ERROR_MSG("Problems getting a new Tx buffer - Progamming error ");
		}
		
		if (S_igmp_queue_get_message(tmsg, &mesg_efflen)<0){
			ETH_L_tx_un_reserve_buffer( );
			//PRINT_MSG(" unreserv & break");
			break;
		}
		
		//PRINT_MSG("Sending IGMP to ethernet");
		
		if ( (ret = ETH_L_tx_send_messagebuffer(NULL, mesg_efflen)) < 0 ){
			ERROR_MSG("Message not sent. ETH_L_tx_send_messagebuffer returned: %d", ret);
			/* Perhaps I should un-reserv the buffer tmsg */
		}
	}
	
	return;
}

