/*****************************************************************************
 * M_int_req_queue.c:
 *****************************************************************************
 * Copyright (C) 2006-2012 the FTT-SE team.
 *
 * Author: Ricardo Marau <marau at fe.up.pt>
 *
 * FTT-SE is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  FTT-SE is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with FTT-SE.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

#include "Ports.h"
#include "request_cmds.h"

#define MAXIMUM_COMMAND_CONTAINER_SIZE 300
#define COMMANDS_ARRAY_MAX_SIZE 20 //defines the number of simultaneous groups of changes admissible


typedef struct command_idx{
	struct command_idx *link;
	unsigned char command_type;
	unsigned short command_size;
	unsigned char command_container[MAXIMUM_COMMAND_CONTAINER_SIZE];
}COMMAND_IDX;

static COMMAND_IDX commands_idx_heap[ COMMANDS_ARRAY_MAX_SIZE ];

static COMMAND_IDX *free_slots;

typedef struct{
	unsigned int req_id;
	COMMAND_IDX *queue_base;  // NULL   -not in used	
}COMMAND_GROUPS_IDX;

static COMMAND_GROUPS_IDX command_groups[COMMANDS_ARRAY_MAX_SIZE];


signed char M_int_req_queue_Init(void)
{
	unsigned short i;
	/* Test for each command structures size if they fit */

	/* Initialize all commands container as free */
	for (i=0; i<(COMMANDS_ARRAY_MAX_SIZE-1); i++)
		commands_idx_heap[i].link = &(commands_idx_heap[i+1]);
	commands_idx_heap[(COMMANDS_ARRAY_MAX_SIZE-1)].link = NULL;
	free_slots = &(commands_idx_heap[0]);

	/* Initialize the command groups as empty */
	for (i=0; i<COMMANDS_ARRAY_MAX_SIZE; i++)
		command_groups[i].queue_base = NULL;

#ifdef PRINT_MODULES_INIT
	PRINT_MSG(" -M_int_req_queue_Init Ok");
#endif
	
	return 0;
}

signed char M_int_req_queue_Close(void)
{
#ifdef PRINT_MODULES_INIT
	PRINT_MSG(" -M_int_req_queue_Close Ok");
#endif
	return 0;
}

static unsigned char is_empty(COMMAND_IDX **queue_base_p)
{
	if ( *queue_base_p == NULL ) /* EMPTY */
		return 1;
	return 0;
}

static signed char de_queue_from_top( COMMAND_IDX **queue_base_p, COMMAND_IDX **item )
{
	if ( *queue_base_p == NULL ) /* EMPTY */
		return -1;

	*item = *queue_base_p;
	*queue_base_p = (*queue_base_p)->link;

	return 0;
}

static signed char enqueue_in_top( COMMAND_IDX **queue_base_p, COMMAND_IDX *item )
{
	item->link = *queue_base_p;
	*queue_base_p = item;
	return 0;
}

static signed char enqueue_in_bottom( COMMAND_IDX **queue_base_p, COMMAND_IDX *item )
{
	COMMAND_IDX **temp_p = queue_base_p;

	/* Go to the end */
	while( *temp_p != NULL ) temp_p = &((*temp_p)->link);

	/* Append it */
	item->link = NULL;
	*temp_p = item;

	return 0;

}

signed char M_int_req_queue_Append_cmd(unsigned int req_uid, M_INT_MSG_TYPE command_type,void *command_data, unsigned short command_size)
{
	unsigned short i;
	signed short available;

	COMMAND_IDX *free_slot_temp;

	if ( command_size>MAXIMUM_COMMAND_CONTAINER_SIZE ){
		ERROR_MSG("Design error - Command size exceeded!");
		return -3;
	}

	/* Search for available group */
	available = -1;
	for ( i=0; i<COMMANDS_ARRAY_MAX_SIZE; i++ ){
		/* check for equal */
		if (command_groups[i].queue_base != NULL)
			if ( command_groups[i].req_id == req_uid )
				break;
		
		if ( available<0 ){
			/* we still haven't found a free slot */
			if (command_groups[i].queue_base == NULL)
				available = i;
		}
	
	}

	if (i == COMMANDS_ARRAY_MAX_SIZE){
		/* none of them is equal */
		if (available < 0){
		       	/* FULL */
			return -1;
		}
		else{
			/* let's create */
			i = available;
			command_groups[available].req_id = req_uid;
		}
	}

	/* Get free slot in group */
	if (de_queue_from_top( &free_slots, &free_slot_temp ) < 0 ) /* no available slots */
		return -2;

	/* populate it*/
	free_slot_temp->command_type = command_type;
	free_slot_temp->command_size = command_size;
	memcpy(free_slot_temp->command_container, command_data, command_size);

	/* enqueue */
	if (enqueue_in_bottom( &(command_groups[i].queue_base), free_slot_temp  ) < 0 )
		return -4;

	return 0;
}

signed char M_int_req_queue_Get_cmd( unsigned int req_uid, M_INT_MSG_TYPE *command_type, void *command_data)
{
	unsigned short i;
	COMMAND_IDX *temp_item;

	/* Search for the group */
	for ( i=0; i<COMMANDS_ARRAY_MAX_SIZE; i++ ){
		/* check for equal */
		if (command_groups[i].queue_base != NULL)
			if ( command_groups[i].req_id == req_uid )
				break;
	}

	if (i == COMMANDS_ARRAY_MAX_SIZE){
		/* none of them is equal */
		return -1;
	}
	
	if ( de_queue_from_top( &(command_groups[i].queue_base), &temp_item ) < 0 ){
		/* It was already empty */
		return -2;
	}

	/* Copy */
	*command_type = temp_item->command_type;
	memcpy(command_data, temp_item->command_container, temp_item->command_size);

	/* put it in the heap */
	enqueue_in_top(&free_slots, temp_item);

	/* check if it was the last command */
	if (is_empty( &(command_groups[i].queue_base) )){
//		command_groups[i].queue_base = NULL; //which already was
		return 0;
	}

	return 1;
}

signed char M_int_req_queue_Abort_cmds ( unsigned int req_uid )
{
	unsigned short i;
	COMMAND_IDX *temp_item;

	/* Search for the group */
	for ( i=0; i<COMMANDS_ARRAY_MAX_SIZE; i++ ){
		/* check for equal */
		if (command_groups[i].queue_base != NULL)
			if ( command_groups[i].req_id == req_uid )
				break;
	}

	if (i == COMMANDS_ARRAY_MAX_SIZE){
		/* none of them is equal */
		return -1;
	}

	while (1){
		if ( de_queue_from_top( &(command_groups[i].queue_base), &temp_item ) < 0 ){
			/* It was already empty */
			return 0;
		}

		/* put it in the heap */
		enqueue_in_top(&free_slots, temp_item);
	}

}



