/*****************************************************************************
 * S_int_async_ch.c:
 *****************************************************************************
 * Copyright (C) 2006-2012 the FTT-SE team.
 *
 * Author: Ricardo Marau <marau at fe.up.pt>
 *
 * FTT-SE is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  FTT-SE is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with FTT-SE.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/


#include "Ports.h"
#include "S_ftt-core.layer.h"
#include "request_cmds.h"
#include "ftt-interface.layer_types.h"


static unsigned char int_status=0; // 0 -not initialized
                                   // 1 -waiting for aynchronous channels ids
				   // 2 -initialized and ready to run
				   // 3 -initialization failed
static unsigned short g_id_rx, g_id_tx;
static unsigned char g_id_rx_tag, g_id_tx_tag;
static unsigned char g_app_id;
static unsigned int g_ec_len_us;

static rtl_pthread_t int_initialization_th;

static FTT_SIGNAL *g_asynch_tx_sig;
static FTT_SIGNAL *g_asynch_rx_sig;

static FTT_SEMAPHORE init_wake_semaph;
static unsigned char init_flag=0;

unsigned char S_int_Block_on_initialization(unsigned int *ret_ec_len_us)
{
	if (init_flag){
		ERROR_MSG("You should not call S_int_Block_on_initialization twice in the same FTT scope!!");
		return -1;
	}
	init_flag = 1;

	if ( FTT_SEMAPHORE_wait(&init_wake_semaph) < 0 )
		return -2;
	
	if (ret_ec_len_us) *ret_ec_len_us = g_ec_len_us;
	return g_app_id;
}

void *int_initialization(void *t)
{
	int ret;

	int_status = 1;

	PRINT_MSG("\ni: Finishing initialization: getting async buffers...");
	/* Request and block for asynchronous communication channels */
	if( (ret = S_FTT_CORE_L_Get_asynchronous_buffers( &g_app_id, &g_id_rx, &g_id_tx, &g_ec_len_us)) < 0 ){
		ERROR_MSG("S_FTT_CORE_L_Get_asynchronous_buffers returned %d", ret);
		int_status = 3;
		return NULL;
	}
	
	PRINT_MSG("i: Master detected: ");
    S_FTT_CORE_L_Wait4var( 1 /*AM*/, g_id_tx );
	//PRINTF(" TX");
	S_FTT_CORE_L_Wait4var( 1 /*AM*/, g_id_rx );
	//PRINTF(" RX");

	/* Now, bind the newly created - let the FTT provide the memory */
	if ( (ret=S_FTT_CORE_L_bind_mesg( 1 /*AM*/, g_id_rx, &g_asynch_rx_sig, NULL,
									 0, /*mem_size - 0 'cause mem is controlled by the FTT */
									 10 /* FIFO size */
									 )) < 0){
		ERROR_MSG("S_FTT_CORE_L_bind_mesg rx returned error: ret %d", ret);
		return NULL;
	}
	
	if ( S_FTT_CORE_L_get_current( 1 /*AM*/, g_id_rx, &g_id_rx_tag, NULL, NULL ) < 0)
		ERROR_MSG("Error getting var tag");

	if ( (ret=S_FTT_CORE_L_bind_mesg( 1 /*AM*/, g_id_tx, &g_asynch_tx_sig, NULL,
									 0, /*mem_size - 0 'cause mem is controlled by the FTT */
									 10 /* FIFO size */
									 )) < 0){
		ERROR_MSG("S_FTT_CORE_L_bind_mesg tx returned error: ret %d ", ret);
		return NULL;
	}
	
	if ( S_FTT_CORE_L_get_current( 1 /*AM*/, g_id_tx, &g_id_tx_tag, NULL, NULL ) < 0)
		ERROR_MSG("Error getting var tag");

	PRINTF(" AMs binded tx 0x%04x and rx 0x%04x",g_id_tx,g_id_rx);

	int_status = 2;
	FTT_SEMAPHORE_post(&init_wake_semaph);

	return NULL;
}

signed char S_int_async_ch_Init(unsigned char app_id)
{
	rtl_pthread_attr_t attr;
	struct rtl_sched_param sched_param;
	int ret;

	/* Check app_id range */
	if ( (app_id) >= M_INT_MAX_APP_NODES ){
		ERROR_MSG("Application ID out of range, it must be between 1-%d", (M_INT_MAX_APP_NODES-1));
		return -5;
	}

	if ( FTT_SEMAPHORE_init(&init_wake_semaph, 0) < 0 ){       //Initialization locked
		ERROR_MSG("Couldn't initialize semaphore");
		return -6;
	}
	
	int_status = 0;
	g_app_id = app_id;

	/* Initialize the thread that will wait for the asynch channels */
	rtl_pthread_attr_init( &attr );
	sched_param.sched_priority = rtl_sched_get_priority_min(RTL_SCHED_FIFO);
	rtl_pthread_attr_setschedparam( &attr, &sched_param );
#ifdef RTL_PRO	
	rtl_pthread_attr_setfp_np(&attr, 1);
#endif
	if ( (ret=rtl_pthread_create( &int_initialization_th, &attr, int_initialization, (void *)0 )) ){
		ERROR_MSG("Init Fatal Error: pthread_create returned %d", ret);
		return -1;
	}
#ifdef PRINT_THREAD_ID
#ifdef RTL_PRO
	ERROR_MSG("New thread id: %d", int_initialization_th);
#else
	ERROR_MSG("New thread id: %lu", int_initialization_th);
#endif
#endif

	return 0;

}

signed char S_int_async_ch_Close(void)
{
	/* cancel the int_initialization thread */
	rtl_pthread_cancel( int_initialization_th );
	rtl_pthread_join( int_initialization_th, NULL );

	FTT_SEMAPHORE_destroy(&init_wake_semaph);
	return 0;
}

signed char S_int_async_ch_SendCMD( unsigned char *buf, unsigned short size )
{
	int ret;
	if ( int_status!=2 ){
		ERROR_MSG("Possibly the the asynch channels are still not set");
		return -1;
	}

	if (size > REQ_CMDS_MAX_SIZE){
		//size exccess
		ERROR_MSG("Command size exceeded");
		return -2;
	}

	if ( (ret=S_FTT_CORE_send(g_id_tx, g_id_tx_tag, buf, 0)) < 0){
		ERROR_MSG("Failed sending the command. ret=%d", ret);
		return -3;
	}

	return 0;
}

signed char S_int_async_ch_GetCMD_blocking ( unsigned char *buf, unsigned short buf_max_size )
{
	if ( int_status!=2 )
		return -1;
//#warning I should be checking the size directly
	if (buf_max_size < REQ_REPLY_MAX_SIZE){
		//size exccess
		ERROR_MSG("Command size not enough");
		return -2;
	}

	FTT_wait4signal( g_asynch_rx_sig );

	if ( S_FTT_CORE_receive( g_id_rx, NULL, buf, NULL/*status*/) < 0 )
		return -3;

	return 0;
}
